@@ -3,7 +3,7 @@ use futures::future::Future;
 use futures::stream::Stream;
 extern crate clap;
 extern crate sequoia_ipc as ipc;
-use ipc::assuan::Client;
+use crate::ipc::assuan::Client;
 
 fn main() {
     let matches = clap::App::new("assuan-client")
