@@ -3,22 +3,21 @@
 #![warn(missing_docs)]
 
 use std::collections::BTreeMap;
+use std::convert::TryFrom;
 use std::ops::{Deref, DerefMut};
 use std::path::{Path, PathBuf};
 use std::process::Command;
 
 use futures::{Async, Future, Stream};
 
-extern crate libc;
-extern crate tempfile;
-
-use crate::openpgp::types::HashAlgorithm;
-use crate::openpgp::fmt::hex;
-use crate::openpgp::crypto;
-use crate::openpgp::crypto::sexp::Sexp;
-use crate::openpgp::packet::prelude::*;
-use crate::openpgp::parse::Parse;
-use crate::openpgp::serialize::Serialize;
+use sequoia_openpgp as openpgp;
+use openpgp::types::HashAlgorithm;
+use openpgp::fmt::hex;
+use openpgp::crypto;
+use openpgp::crypto::sexp::Sexp;
+use openpgp::packet::prelude::*;
+use openpgp::parse::Parse;
+use openpgp::serialize::Serialize;
 
 use crate::Result;
 use crate::assuan;
@@ -111,7 +110,7 @@ impl Context {
         for argument in arguments {
             gpgconf.arg(argument);
         }
-        let output = gpgconf.output().map_err(|e| -> failure::Error {
+        let output = gpgconf.output().map_err(|e| -> anyhow::Error {
             Error::GPGConf(e.to_string()).into()
         })?;
 
@@ -242,7 +241,7 @@ impl DerefMut for Agent {
 
 impl Stream for Agent {
     type Item = assuan::Response;
-    type Error = failure::Error;
+    type Error = anyhow::Error;
 
     /// Attempt to pull out the next value of this stream, returning
     /// None if the stream is finished.
@@ -263,7 +262,7 @@ impl Agent {
     /// server is running for the given context, this operation will
     /// fail.
     pub fn connect<'c>(ctx: &'c Context)
-                   -> impl Future<Item = Self, Error = failure::Error> + 'c
+                   -> impl Future<Item = Self, Error = anyhow::Error> + 'c
     {
         futures::lazy(move || ctx.socket("agent"))
             .and_then(Self::connect_to)
@@ -275,7 +274,7 @@ impl Agent {
     /// server is running for the given context, this operation will
     /// fail.
     pub fn connect_to<P>(path: P)
-                         -> impl Future<Item = Self, Error = failure::Error>
+                         -> impl Future<Item = Self, Error = anyhow::Error>
         where P: AsRef<Path>
     {
         assuan::Client::connect(path)
@@ -287,8 +286,8 @@ impl Agent {
     pub fn sign<'a, R>(&'a mut self,
                        key: &'a Key<key::PublicParts, R>,
                        algo: HashAlgorithm, digest: &'a [u8])
-        -> impl Future<Item = crypto::mpis::Signature,
-                       Error = failure::Error> + 'a
+        -> impl Future<Item = crypto::mpi::Signature,
+                       Error = anyhow::Error> + 'a
         where R: key::KeyRole
     {
         SigningRequest::new(&mut self.c, key, algo, digest)
@@ -298,9 +297,9 @@ impl Agent {
     /// by the agent.
     pub fn decrypt<'a, R>(&'a mut self,
                           key: &'a Key<key::PublicParts, R>,
-                          ciphertext: &'a crypto::mpis::Ciphertext)
+                          ciphertext: &'a crypto::mpi::Ciphertext)
         -> impl Future<Item = crypto::SessionKey,
-                       Error = failure::Error> + 'a
+                       Error = anyhow::Error> + 'a
         where R: key::KeyRole
     {
         DecryptionRequest::new(&mut self.c, key, ciphertext)
@@ -309,14 +308,15 @@ impl Agent {
     /// Computes options that we want to communicate.
     fn options() -> Vec<String> {
         use std::env::var;
-        use std::ffi::CStr;
 
         let mut r = Vec::new();
 
         if let Ok(tty) = var("GPG_TTY") {
             r.push(format!("OPTION ttyname={}", tty));
         } else {
+            #[cfg(unix)]
             unsafe {
+                use std::ffi::CStr;
                 let tty = libc::ttyname(0);
                 if ! tty.is_null() {
                     if let Ok(tty) = CStr::from_ptr(tty).to_str() {
@@ -406,8 +406,8 @@ fn protocol_error<T>(response: &assuan::Response) -> Result<T> {
 impl<'a, 'b, 'c, R> Future for SigningRequest<'a, 'b, 'c, R>
     where R: key::KeyRole
 {
-    type Item = crypto::mpis::Signature;
-    type Error = failure::Error;
+    type Item = crypto::mpi::Signature;
+    type Error = anyhow::Error;
 
     fn poll(&mut self) -> std::result::Result<Async<Self::Item>, Self::Error> {
         use self::SigningRequestState::*;
@@ -520,7 +520,7 @@ struct DecryptionRequest<'a, 'b, 'c, R>
 {
     c: &'a mut assuan::Client,
     key: &'b Key<key::PublicParts, R>,
-    ciphertext: &'c crypto::mpis::Ciphertext,
+    ciphertext: &'c crypto::mpi::Ciphertext,
     options: Vec<String>,
     state: DecryptionRequestState,
 }
@@ -530,7 +530,7 @@ impl<'a, 'b, 'c, R> DecryptionRequest<'a, 'b, 'c, R>
 {
     fn new(c: &'a mut assuan::Client,
            key: &'b Key<key::PublicParts, R>,
-           ciphertext: &'c crypto::mpis::Ciphertext)
+           ciphertext: &'c crypto::mpi::Ciphertext)
            -> Self {
         Self {
             c,
@@ -555,7 +555,7 @@ impl<'a, 'b, 'c, R> Future for DecryptionRequest<'a, 'b, 'c, R>
     where R: key::KeyRole
 {
     type Item = crypto::SessionKey;
-    type Error = failure::Error;
+    type Error = anyhow::Error;
 
     fn poll(&mut self) -> std::result::Result<Async<Self::Item>, Self::Error> {
         use self::DecryptionRequestState::*;
@@ -631,7 +631,7 @@ impl<'a, 'b, 'c, R> Future for DecryptionRequest<'a, 'b, 'c, R>
                     },
                     Async::Ready(None) => {
                         let mut buf = Vec::new();
-                        Sexp::from_ciphertext(&self.ciphertext)?
+                        Sexp::try_from(self.ciphertext)?
                             .serialize(&mut buf)?;
                         self.c.data(&buf)?;
                         self.state = Inquire(Vec::new(), true);
@@ -701,7 +701,7 @@ impl<'a> KeyPair<'a> {
         where R: key::KeyRole
     {
         Ok(KeyPair {
-            public: key.mark_role_unspecified_ref(),
+            public: key.role_as_unspecified(),
             agent_socket: ctx.socket("agent")?.into(),
         })
     }
@@ -713,10 +713,10 @@ impl<'a> crypto::Signer for KeyPair<'a> {
     }
 
     fn sign(&mut self, hash_algo: HashAlgorithm, digest: &[u8])
-            -> Result<openpgp::crypto::mpis::Signature>
+            -> openpgp::Result<openpgp::crypto::mpi::Signature>
     {
         use crate::openpgp::types::PublicKeyAlgorithm::*;
-        use crate::openpgp::crypto::mpis::PublicKey;
+        use crate::openpgp::crypto::mpi::PublicKey;
 
         #[allow(deprecated)]
         match (self.public.pk_algo(), self.public.mpis())
@@ -743,14 +743,15 @@ impl<'a> crypto::Decryptor for KeyPair<'a> {
         self.public
     }
 
-    fn decrypt(&mut self, ciphertext: &crypto::mpis::Ciphertext)
-               -> Result<crypto::SessionKey>
+    fn decrypt(&mut self, ciphertext: &crypto::mpi::Ciphertext,
+               _plaintext_len: Option<usize>)
+               -> openpgp::Result<crypto::SessionKey>
     {
-        use crate::openpgp::crypto::mpis::{PublicKey, Ciphertext};
+        use crate::openpgp::crypto::mpi::{PublicKey, Ciphertext};
 
         match (self.public.mpis(), ciphertext) {
             (PublicKey::RSA { .. }, Ciphertext::RSA { .. })
-                | (PublicKey::Elgamal { .. }, Ciphertext::Elgamal { .. })
+                | (PublicKey::ElGamal { .. }, Ciphertext::ElGamal { .. })
                 | (PublicKey::ECDH { .. }, Ciphertext::ECDH { .. }) => {
                     let mut a = Agent::connect_to(&self.agent_socket).wait()?;
                     let sk = a.decrypt(self.public, ciphertext).wait()?;
@@ -767,17 +768,17 @@ impl<'a> crypto::Decryptor for KeyPair<'a> {
 }
 
 
-#[derive(Fail, Debug)]
+#[derive(thiserror::Error, Debug)]
 /// Errors used in this module.
 pub enum Error {
     /// Errors related to `gpgconf`.
-    #[fail(display = "gpgconf: {}", _0)]
+    #[error("gpgconf: {0}")]
     GPGConf(String),
     /// The remote operation failed.
-    #[fail(display = "Operation failed: {}", _0)]
+    #[error("Operation failed: {0}")]
     OperationFailed(String),
     /// The remote party violated the protocol.
-    #[fail(display = "Protocol violation: {}", _0)]
+    #[error("Protocol violation: {0}")]
     ProtocolError(String),
 
 }
