@@ -60,7 +60,7 @@ char *pgp_keyid_to_hex (const pgp_keyid_t keyid);
 /*/
 /// Compares KeyIDs.
 /*/
-int pgp_keyid_equal (const pgp_keyid_t a, const pgp_keyid_t b);
+bool pgp_keyid_equal (const pgp_keyid_t a, const pgp_keyid_t b);
 
 
 /* sequoia::openpgp::Fingerprint.  */
@@ -122,7 +122,7 @@ pgp_keyid_t pgp_fingerprint_to_keyid (const pgp_fingerprint_t fp);
 /*/
 /// Compares Fingerprints.
 /*/
-int pgp_fingerprint_equal (const pgp_fingerprint_t a, const pgp_fingerprint_t b);
+bool pgp_fingerprint_equal (const pgp_fingerprint_t a, const pgp_fingerprint_t b);
 
 /* sequoia::openpgp::RevocationStatus.  */
 
@@ -316,13 +316,6 @@ pgp_signature_t pgp_signature_from_file (pgp_error_t *errp,
 pgp_signature_t pgp_signature_from_bytes (pgp_error_t *errp,
                                           const uint8_t *b, size_t len);
 
-/*/
-/// Serializes this object.
-/*/
-pgp_status_t pgp_signature_serialize (pgp_error_t *errp,
-				      const pgp_signature_t signature,
-				      pgp_writer_t writer);
-
 /*/
 /// Converts the signature to a packet.
 /*/
@@ -567,13 +560,6 @@ pgp_literal_t pgp_literal_from_file (pgp_error_t *errp,
 pgp_literal_t pgp_literal_from_bytes (pgp_error_t *errp,
                                       const uint8_t *b, size_t len);
 
-/*/
-/// Serializes this object.
-/*/
-pgp_status_t pgp_literal_serialize (pgp_error_t *errp,
-                                    const pgp_literal_t literal,
-                                    pgp_writer_t writer);
-
 /*/
 /// Converts the literal to a packet.
 /*/
@@ -584,7 +570,7 @@ pgp_packet_t pgp_literal_into_packet (pgp_literal_t literal);
 /*/
 void pgp_literal_free (pgp_literal_t literal);
 
-/* openpgp::cert::UserIDBundle.  */
+/* openpgp::amalgamation::UserIDAmalgamation.  */
 
 /*/
 /// Returns the user id.
@@ -595,27 +581,107 @@ void pgp_literal_free (pgp_literal_t literal);
 ///
 /// The caller must free the returned value.
 /*/
-char *pgp_user_id_bundle_user_id (pgp_user_id_bundle_t binding);
+char *pgp_user_id_amalgamation_user_id (pgp_user_id_amalgamation_t ua);
 
 /*/
-/// Returns a reference to the self-signature, if any.
+/// Frees the User ID Amalgamation.
 /*/
-pgp_signature_t pgp_user_id_bundle_selfsig(pgp_user_id_bundle_t binding,
-                                            pgp_policy_t policy);
+void pgp_user_id_amalgamation_free (pgp_user_id_amalgamation_t ua);
 
-/* openpgp::cert::UserIDBundleIter.  */
+/*/
+/// Clones the UserID Amalgamation.
+/*/
+pgp_user_id_amalgamation_t pgp_user_id_amalgamation_clone (pgp_user_id_amalgamation_t ua);
+
+/*/
+/// Returns a human readable description of this object suitable for
+/// debugging.
+/*/
+char *pgp_user_id_amalgamation_debug (const pgp_user_id_amalgamation_t ua);
+
+/*/
+/// Returns the user id.
+///
+/// This function may fail and return NULL if the user id contains an
+/// interior NUL byte.  We do this rather than complicate the API, as
+/// there is no valid use for such user ids; they must be malicious.
+///
+/// The caller must free the returned value.
+/*/
+pgp_user_id_t pgp_valid_user_id_amalgamation_user_id
+    (pgp_valid_user_id_amalgamation_t ua);
+
+/*/
+/// Returns the Valid UserID Amalgamation's revocation status.
+/*/
+pgp_revocation_status_t pgp_valid_user_id_amalgamation_revocation_status
+    (pgp_valid_user_id_amalgamation_t ua);
+
+/*/
+/// Returns a reference to the self-signature.
+/*/
+pgp_signature_t pgp_valid_user_id_amalgamation_binding_signature
+    (pgp_valid_user_id_amalgamation_t ua);
+
+/*/
+/// Changes the policy applied to the `ValidUserIDAmalgamation`.
+///
+/// This consumes the UserID amalgamation.
+/*/
+pgp_valid_user_id_amalgamation_t pgp_valid_user_id_amalgamation_with_policy
+    (pgp_error_t *errp,
+     pgp_valid_user_id_amalgamation_t ua,
+     pgp_policy_t policy,
+     time_t time);
+
+/*/
+/// Frees the User ID Amalgamation.
+/*/
+void pgp_valid_user_id_amalgamation_free (pgp_valid_user_id_amalgamation_t ua);
+
+/*/
+/// Clones the UserID Amalgamation.
+/*/
+pgp_valid_user_id_amalgamation_t pgp_valid_user_id_amalgamation_clone (pgp_valid_user_id_amalgamation_t ua);
+
+/*/
+/// Returns a human readable description of this object suitable for
+/// debugging.
+/*/
+char *pgp_valid_user_id_amalgamation_debug (const pgp_valid_user_id_amalgamation_t ua);
+
+/* openpgp::cert::UserIDIter.  */
 
 /*/
 /// Returns the next element in the iterator.
 /*/
-pgp_user_id_bundle_t pgp_user_id_bundle_iter_next (pgp_user_id_bundle_iter_t iter);
+pgp_user_id_amalgamation_t pgp_cert_user_id_iter_next (pgp_cert_user_id_iter_t iter);
 
 /*/
-/// Frees an pgp_user_id_bundle_iter_t.
+/// Sets a policy.
+///
+/// Only User IDs that are valid according to the policy at the
+/// specified time are returned.
 /*/
-void pgp_user_id_bundle_iter_free (pgp_user_id_bundle_iter_t iter);
+pgp_cert_valid_user_id_iter_t pgp_cert_user_id_iter_policy
+    (pgp_cert_user_id_iter_t iter, pgp_policy_t policy, time_t when);
 
-/* openpgp::cert::KeyIter.  */
+/*/
+/// Frees a pgp_cert_user_id_iter_t.
+/*/
+void pgp_cert_user_id_iter_free (pgp_cert_user_id_iter_t iter);
+
+/*/
+/// Returns the next element in the iterator.
+/*/
+pgp_valid_user_id_amalgamation_t pgp_cert_valid_user_id_iter_next (pgp_cert_valid_user_id_iter_t iter);
+
+/*/
+/// Frees a pgp_cert_valid_user_id_iter_t.
+/*/
+void pgp_cert_valid_user_id_iter_free (pgp_cert_valid_user_id_iter_t iter);
+
+/* openpgp::cert::KeyAmalgamationIter.  */
 
 /*/
 /// Changes the iterator to only return keys that have secret keys.
@@ -633,19 +699,19 @@ void pgp_cert_key_iter_secret (pgp_cert_key_iter_t iter);
 void pgp_cert_key_iter_unencrypted_secret (pgp_cert_key_iter_t iter);
 
 /*/
-/// Changes the iterator to only return keys that have unencrypted
-/// secret keys.
+/// Sets a policy.
 ///
-/// Note: you may not call this function after starting to iterate.
+/// Only keys that are valid according to the policy at the specified
+/// time are returned.
 /*/
 pgp_cert_valid_key_iter_t pgp_cert_key_iter_policy
     (pgp_cert_key_iter_t iter, pgp_policy_t policy, time_t when);
 
 /*/
-/// Returns a reference to the next key.  Returns NULL if there are no
+/// Returns a key amalgamation for the next key.  Returns NULL if there are no
 /// more elements.
 /*/
-pgp_key_t pgp_cert_key_iter_next (pgp_cert_key_iter_t iter);
+pgp_key_amalgamation_t pgp_cert_key_iter_next (pgp_cert_key_iter_t iter);
 
 /*/
 /// Frees an pgp_cert_key_iter_t.
@@ -736,7 +802,7 @@ void pgp_cert_valid_key_iter_secret (pgp_cert_valid_key_iter_t iter);
 void pgp_cert_valid_key_iter_unencrypted_secret (pgp_cert_valid_key_iter_t iter);
 
 /*/
-/// Returns a reference to the next key.  Returns NULL if there are no
+/// Returns a key amalgamation to the next key.  Returns NULL if there are no
 /// more elements.
 ///
 /// If signature is not NULL, stores the current self-signature (if
@@ -747,9 +813,10 @@ void pgp_cert_valid_key_iter_unencrypted_secret (pgp_cert_valid_key_iter_t iter)
 /// If rev is not NULL, this stores the key's revocation status in
 /// *rev.
 /*/
-pgp_key_t pgp_cert_valid_key_iter_next (pgp_cert_valid_key_iter_t iter,
-                                        pgp_signature_t *signature,
-                                        pgp_revocation_status_t *rev);
+pgp_valid_key_amalgamation_t pgp_cert_valid_key_iter_next
+    (pgp_cert_valid_key_iter_t iter,
+     pgp_signature_t *signature,
+     pgp_revocation_status_t *rev);
 
 /*/
 /// Frees an pgp_cert_valid_key_iter_t.
@@ -807,7 +874,7 @@ pgp_cert_t pgp_cert_clone (pgp_cert_t cert);
 /*/
 /// Compares Certs.
 /*/
-int pgp_cert_equal (const pgp_cert_t a, const pgp_cert_t b);
+bool pgp_cert_equal (const pgp_cert_t a, const pgp_cert_t b);
 
 /*/
 /// Returns a human readable description of this object intended for
@@ -887,26 +954,15 @@ pgp_key_t pgp_cert_primary_key (pgp_cert_t cert);
 /// If `when` is 0, then returns the Cert's revocation status as of the
 /// time of the call.
 /*/
-pgp_revocation_status_t pgp_cert_revoked (pgp_cert_t cert,
-                                          pgp_policy_t policy, time_t when);
+pgp_revocation_status_t pgp_cert_revocation_status (pgp_cert_t cert,
+                                                    pgp_policy_t policy, time_t when);
 
 /*/
-/// Writes a revocation certificate to the writer.
+/// Returns a new revocation certificate for the Cert.
 ///
-/// This function consumes the writer.  It does *not* consume cert.
+/// This function does *not* consume `cert`.
 /*/
 pgp_signature_t pgp_cert_revoke (pgp_error_t *errp,
-                              pgp_cert_t cert,
-                              pgp_signer_t primary_signer,
-                              pgp_reason_for_revocation_t code,
-                              const char *reason);
-
-/*/
-/// Adds a revocation certificate to the cert.
-///
-/// This function consumes the cert.
-/*/
-pgp_cert_t pgp_cert_revoke_in_place (pgp_error_t *errp,
                                  pgp_cert_t cert,
                                  pgp_signer_t primary_signer,
                                  pgp_reason_for_revocation_t code,
@@ -921,10 +977,7 @@ pgp_status_t pgp_cert_alive(pgp_error_t *errp, pgp_cert_t cert,
                             pgp_policy_t policy, time_t when);
 
 /*/
-/// Changes the Cert's expiration.
-///
-/// Expiry is when the key should expire in seconds relative to the
-/// key's creation (not the current time).
+/// Sets the key to expire at the given time.
 ///
 /// This function consumes `cert` and returns a new `Cert`.
 /*/
@@ -932,7 +985,7 @@ pgp_cert_t pgp_cert_set_expiration_time(pgp_error_t *errp,
                                pgp_cert_t cert,
                                pgp_policy_t policy,
                                pgp_signer_t signer,
-                               uint32_t expiry);
+                               time_t expiry);
 
 /*/
 /// Returns whether the Cert includes any secret key material.
@@ -940,9 +993,16 @@ pgp_cert_t pgp_cert_set_expiration_time(pgp_error_t *errp,
 int pgp_cert_is_tsk(pgp_cert_t cert);
 
 /*/
-/// Returns an iterator over the `UserIDBundle`s.
+/// Returns an iterator over the `UserID`s.
 /*/
-pgp_user_id_bundle_iter_t pgp_cert_user_id_binding_iter (pgp_cert_t cert);
+pgp_cert_user_id_iter_t pgp_cert_user_id_iter (pgp_cert_t cert);
+
+/*/
+/// Returns an iterator over the valid `UserID`s.
+/*/
+pgp_cert_valid_user_id_iter_t pgp_cert_valid_user_id_iter (pgp_cert_t cert,
+                                                           pgp_policy_t policy,
+                                                           time_t when);
 
 /*/
 /// Returns an iterator over all `Key`s in a Cert.
@@ -965,7 +1025,8 @@ pgp_cert_valid_key_iter_t pgp_cert_valid_key_iter
 /*/
 /// Returns the Cert's primary user id (if any).
 /*/
-char *pgp_cert_primary_user_id(pgp_cert_t cert, pgp_policy_t policy);
+char *pgp_cert_primary_user_id(pgp_cert_t cert, pgp_policy_t policy,
+                               time_t when);
 
 /*/
 /// Returns a CertParser.
@@ -995,7 +1056,7 @@ pgp_cert_parser_t pgp_cert_parser_from_packet_parser(pgp_packet_parser_result_t
 pgp_cert_t pgp_cert_parser_next(pgp_error_t *errp, pgp_cert_parser_t parser);
 
 /*/
-/// Frees an pgp_cert_key_iter_t.
+/// Frees an pgp_cert_parser_t.
 /*/
 void pgp_cert_parser_free (pgp_cert_parser_t parser);
 
@@ -1052,6 +1113,12 @@ void pgp_cert_builder_add_transport_encryption_subkey(pgp_cert_builder_t *certb)
 /*/
 void pgp_cert_builder_add_certification_subkey(pgp_cert_builder_t *certb);
 
+/*/
+/// Sets the creation time.
+/*/
+void pgp_cert_builder_set_creation_time(pgp_cert_builder_t *certb,
+                                        time_t when);
+
 /*/
 /// Generates the actual Cert.
 ///
@@ -1063,6 +1130,89 @@ pgp_status_t pgp_cert_builder_generate(pgp_error_t *errp,
                                       pgp_signature_t *revocation);
 
 
+/* KeyAmalgamation */
+
+/*/
+/// Returns a reference to the Key Amalgamation's `Key`.
+/*/
+pgp_key_t pgp_key_amalgamation_key (pgp_key_amalgamation_t ka);
+
+/*/
+/// Frees the Key Amalgamation.
+/*/
+void pgp_key_amalgamation_free (pgp_key_amalgamation_t ka);
+
+/*/
+/// Clones the Key Amalgamation.
+/*/
+pgp_key_amalgamation_t pgp_key_amalgamation_clone (pgp_key_amalgamation_t ka);
+
+/*/
+/// Returns a human readable description of this object suitable for
+/// debugging.
+/*/
+char *pgp_key_amalgamation_debug (const pgp_key_amalgamation_t ka);
+
+
+/* ValidKeyAmalgamation */
+
+/*/
+/// Returns a reference to the Valid Key Amalgamation's `Key`.
+/*/
+pgp_key_t pgp_valid_key_amalgamation_key (pgp_valid_key_amalgamation_t ka);
+
+/*/
+/// Returns the Valid Key Amalgamation's revocation status.
+/*/
+pgp_revocation_status_t pgp_valid_key_amalgamation_revocation_status (pgp_valid_key_amalgamation_t ka);
+
+/*/
+/// Returns a reference to the Valid Key Amalgamation's binding signature.
+/*/
+pgp_signature_t pgp_valid_key_amalgamation_binding_signature (pgp_valid_key_amalgamation_t ka);
+
+/*/
+/// Creates one or more self-signatures that when merged with the
+/// certificate cause the key to expire at the specified time.
+///
+/// The returned buffer must be freed using libc's allocator.
+/*/
+pgp_status_t pgp_valid_key_amalgamation_set_expiration_time
+  (pgp_error_t *errp,
+   pgp_valid_key_amalgamation_t ka,
+   pgp_signer_t signer,
+   time_t time,
+   pgp_packet_t **packets,
+   size_t *packet_count);
+
+/*/
+/// Changes the policy applied to the `ValidKeyAmalgamation`.
+///
+/// This consumes the key amalgamation.
+/*/
+pgp_valid_key_amalgamation_t pgp_valid_key_amalgamation_with_policy
+    (pgp_error_t *errp,
+     pgp_valid_key_amalgamation_t ka,
+     pgp_policy_t policy,
+     time_t time);
+
+/*/
+/// Frees the Valid Key Amalgamation.
+/*/
+void pgp_valid_key_amalgamation_free (pgp_valid_key_amalgamation_t ka);
+
+/*/
+/// Clones the Valid Key Amalgamation.
+/*/
+pgp_valid_key_amalgamation_t pgp_valid_key_amalgamation_clone (pgp_valid_key_amalgamation_t ka);
+
+/*/
+/// Returns a human readable description of this object suitable for
+/// debugging.
+/*/
+char *pgp_valid_key_amalgamation_debug (const pgp_valid_key_amalgamation_t ka);
+
+
 /* TSK */
 
 /*/
@@ -1285,6 +1435,27 @@ pgp_status_t pgp_user_id_uri(pgp_error_t *errp, pgp_packet_t uid,
 const uint8_t *pgp_user_attribute_value (pgp_packet_t ua,
 					size_t *value_len);
 
+/*/
+/// Frees a pgp_user_id_t.
+/*/
+void pgp_user_id_free (pgp_user_id_t userid);
+
+/*/
+/// Clones the UserID.
+/*/
+pgp_user_id_t pgp_user_id_clone (pgp_user_id_t userid);
+
+/*/
+/// Returns a human readable description of this object suitable for
+/// debugging.
+/*/
+char *pgp_user_id_debug (const pgp_user_id_t userid);
+
+/*/
+/// Compares passwords.
+/*/
+bool pgp_user_id_equal (const pgp_user_id_t a, const pgp_user_id_t b);
+
 /*/
 /// Returns the session key.
 ///
@@ -1652,7 +1823,7 @@ void pgp_message_structure_free (pgp_message_structure_t);
 char *pgp_message_structure_debug (const pgp_message_structure_t);
 
 pgp_message_structure_iter_t
-pgp_message_structure_iter (pgp_message_structure_t);
+pgp_message_structure_into_iter (pgp_message_structure_t);
 
 /*/
 /// Frees this object.
@@ -1721,17 +1892,29 @@ bool pgp_verification_result_good_checksum (pgp_verification_result_t,
 					    pgp_key_t *,
 					    pgp_signature_t *,
 					    pgp_revocation_status_t *);
-bool pgp_verification_result_not_alive (pgp_verification_result_t,
-                                        pgp_signature_t *,
-					pgp_cert_t *,
-					pgp_key_t *,
-					pgp_signature_t *,
-					pgp_revocation_status_t *);
+bool pgp_verification_result_malformed_signature (pgp_verification_result_t,
+                                                  pgp_signature_t *,
+                                                  pgp_error_t *);
 bool pgp_verification_result_missing_key (pgp_verification_result_t,
-					  pgp_signature_t *);
-bool pgp_verification_result_error (pgp_verification_result_t,
-                                    pgp_signature_t *,
-                                    pgp_error_t *);
+                                          pgp_signature_t *);
+bool pgp_verification_result_unbound_key (pgp_verification_result_t,
+                                          pgp_signature_t *,
+                                          pgp_cert_t *,
+                                          pgp_error_t *);
+bool pgp_verification_result_bad_key (pgp_verification_result_t,
+                                      pgp_signature_t *,
+                                      pgp_cert_t *,
+                                      pgp_key_t *,
+                                      pgp_signature_t *,
+                                      pgp_revocation_status_t *,
+                                      pgp_error_t *);
+bool pgp_verification_result_bad_signature (pgp_verification_result_t,
+                                            pgp_signature_t *,
+                                            pgp_cert_t *,
+                                            pgp_key_t *,
+                                            pgp_signature_t *,
+                                            pgp_revocation_status_t *,
+                                            pgp_error_t *);
 
 /*/
 /// Decrypts an OpenPGP message.
@@ -1750,7 +1933,7 @@ bool pgp_verification_result_error (pgp_verification_result_t,
 pgp_reader_t pgp_decryptor_new (pgp_error_t *errp,
     pgp_policy_t policy,
     pgp_reader_t input,
-    pgp_decryptor_get_public_keys_cb_t get_public_keys,
+    pgp_decryptor_get_certs_cb_t get_certs,
     pgp_decryptor_decrypt_cb_t decrypt,
     pgp_decryptor_check_cb_t check,
     pgp_decryptor_inspect_cb_t inspect,
@@ -1765,26 +1948,34 @@ pgp_reader_t pgp_decryptor_new (pgp_error_t *errp,
 pgp_reader_t pgp_verifier_new (pgp_error_t *errp,
     pgp_policy_t policy,
     pgp_reader_t input,
-    pgp_decryptor_get_public_keys_cb_t get_public_keys,
+    pgp_decryptor_get_certs_cb_t get_certs,
     pgp_decryptor_check_cb_t check,
+    pgp_decryptor_inspect_cb_t inspect,
     void *cookie, time_t time);
 
 /*/
-/// Verifies a detached OpenPGP signature.///
-/// A Certificate (see [RFC 4880, section 11.1]) can be used to verify
-/// signatures and encrypt data.  It can be stored in a keystore and
-/// uploaded to keyservers.
-///
-/// [RFC 4880, section 11.1]: https://tools.ietf.org/html/rfc4880#section-11.1
-
+/// Verifies a detached OpenPGP signature.
 /*/
-pgp_reader_t pgp_detached_verifier_new (pgp_error_t *errp,
+pgp_detached_verifier_t pgp_detached_verifier_new (pgp_error_t *errp,
     pgp_policy_t policy,
-    pgp_reader_t signature_input, pgp_reader_t input,
-    pgp_decryptor_get_public_keys_cb_t get_public_keys,
+    pgp_reader_t signature_input,
+    pgp_decryptor_get_certs_cb_t get_certs,
     pgp_decryptor_check_cb_t check,
+    pgp_decryptor_inspect_cb_t inspect,
     void *cookie, time_t time);
 
+/*/
+/// Frees this object.
+/*/
+void pgp_detached_verifier_free (pgp_detached_verifier_t);
+
+/*/
+/// Verifies `data` using `verifier`.
+/*/
+pgp_status_t pgp_detached_verifier_verify (pgp_error_t *errp,
+                                           pgp_detached_verifier_t verifier,
+                                           pgp_reader_t data);
+
 /*/
 /// Returns a new standard policy.
 /*/
@@ -1806,6 +1997,27 @@ void pgp_standard_policy_free (pgp_standard_policy_t);
 /*/
 char *pgp_standard_policy_debug (const pgp_standard_policy_t);
 
+/*/
+/// Returns a new null policy.
+/*/
+pgp_policy_t pgp_null_policy ();
+
+/*/
+/// Clones the object.
+/*/
+pgp_policy_t pgp_null_policy_clone (pgp_null_policy_t);
+
+/*/
+/// Frees this object.
+/*/
+void pgp_null_policy_free (pgp_null_policy_t);
+
+/*/
+/// Returns a human readable description of this object suitable for
+/// debugging.
+/*/
+char *pgp_null_policy_debug (const pgp_null_policy_t);
+
 /*/
 /// Clones the object.
 /*/
