@@ -5,6 +5,7 @@
 //! [`sequoia-openpgp::Cert`]: ../../../sequoia_openpgp/cert/struct.Cert.html
 //! [related functionality]: ../../../sequoia_openpgp/cert/index.html
 
+use std::convert::TryFrom;
 use std::ptr;
 use std::slice;
 use libc::{c_char, c_int, size_t, time_t};
@@ -17,19 +18,7 @@ use self::openpgp::{
         PacketParserResult,
         Parse,
     },
-    cert::{
-        CipherSuite,
-        CertBuilder,
-        CertParser,
-        CertRevocationBuilder,
-        components::{
-            Amalgamation,
-            KeyIter,
-            UserIDBundle,
-            UserIDBundleIter,
-            ValidKeyIter,
-        },
-    },
+    cert::prelude::*,
 };
 
 use crate::error::Status;
@@ -41,6 +30,8 @@ use super::packet_pile::PacketPile;
 use super::tsk::TSK;
 use super::revocation_status::RevocationStatus;
 use super::policy::Policy;
+use super::amalgamation::{UserIDAmalgamation, ValidUserIDAmalgamation};
+use super::key_amalgamation::{KeyAmalgamation, ValidKeyAmalgamation};
 
 use crate::Maybe;
 use crate::RefRaw;
@@ -77,7 +68,7 @@ pub struct Cert(openpgp::Cert);
 fn pgp_cert_from_packet_pile(errp: Option<&mut *mut crate::error::Error>,
                             m: *mut PacketPile)
                             -> Maybe<Cert> {
-    openpgp::Cert::from_packet_pile(m.move_from_raw()).move_into_raw(errp)
+    openpgp::Cert::try_from(m.move_from_raw()).move_into_raw(errp)
 }
 
 /// Returns the first Cert found in the packet parser.
@@ -90,7 +81,7 @@ fn pgp_cert_from_packet_parser(errp: Option<&mut *mut crate::error::Error>,
 {
     let ppr = ffi_param_move!(ppr);
 
-    openpgp::Cert::from_packet_parser(*ppr).move_into_raw(errp)
+    openpgp::Cert::try_from(*ppr).move_into_raw(errp)
 }
 
 /// Merges `other` into `cert`.
@@ -127,7 +118,7 @@ fn pgp_cert_merge_packets(errp: Option<&mut *mut crate::error::Error>,
         slice::from_raw_parts_mut(packets, packets_len)
     };
     let packets =
-        packets.iter_mut().map(|&mut p| p.move_from_raw()).collect();
+        packets.iter_mut().map(|&mut p| p.move_from_raw());
     cert.merge_packets(packets).move_into_raw(errp)
 }
 
@@ -155,7 +146,7 @@ fn pgp_cert_as_tsk(cert: *const Cert) -> *mut TSK<'static> {
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
 fn pgp_cert_primary_key(cert: *const Cert) -> *const Key {
     let key = cert.ref_raw().primary_key().key()
-        .mark_parts_unspecified_ref().mark_role_unspecified_ref();
+        .parts_as_unspecified().role_as_unspecified();
     key.move_into_raw()
 }
 
@@ -168,12 +159,12 @@ fn pgp_cert_primary_key(cert: *const Cert) -> *const Key {
 /// If `when` is 0, then returns the Cert's revocation status as of the
 /// time of the call.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
-fn pgp_cert_revoked(cert: *const Cert, policy: *const Policy, when: time_t)
+fn pgp_cert_revocation_status(cert: *const Cert, policy: *const Policy, when: time_t)
     -> *mut RevocationStatus<'static>
 {
     let policy = &**policy.ref_raw();
     cert.ref_raw()
-        .revoked(policy, maybe_time(when))
+        .revocation_status(policy, maybe_time(when))
         .move_into_raw()
 }
 
@@ -231,7 +222,7 @@ fn int_to_reason_for_revocation(code: c_int) -> ReasonForRevocation {
 /// cert = pgp_cert_merge_packets (NULL, cert, &packet, 1);
 /// assert (cert);
 ///
-/// pgp_revocation_status_t rs = pgp_cert_revoked (cert, policy, 0);
+/// pgp_revocation_status_t rs = pgp_cert_revocation_status (cert, policy, 0);
 /// assert (pgp_revocation_status_variant (rs) == PGP_REVOCATION_STATUS_REVOKED);
 /// pgp_revocation_status_free (rs);
 ///
@@ -262,70 +253,6 @@ fn pgp_cert_revoke(errp: Option<&mut *mut crate::error::Error>,
     sig.move_into_raw(errp)
 }
 
-/// Adds a revocation certificate to the cert.
-///
-/// This function consumes the cert.
-///
-/// # Example
-///
-/// ```c
-/// #include <assert.h>
-/// #include <sequoia/openpgp.h>
-///
-/// pgp_cert_builder_t builder;
-/// pgp_cert_t cert;
-/// pgp_signature_t revocation;
-/// pgp_key_t primary_key;
-/// pgp_key_pair_t primary_keypair;
-/// pgp_signer_t primary_signer;
-/// pgp_policy_t policy = pgp_standard_policy ();
-///
-/// builder = pgp_cert_builder_new ();
-/// pgp_cert_builder_set_cipher_suite (&builder, PGP_CERT_CIPHER_SUITE_CV25519);
-/// pgp_cert_builder_generate (NULL, builder, &cert, &revocation);
-/// assert (cert);
-/// assert (revocation);
-/// pgp_signature_free (revocation);    /* Free the generated one.  */
-///
-/// primary_key = pgp_cert_primary_key (cert);
-/// primary_keypair = pgp_key_into_key_pair (NULL, pgp_key_clone (primary_key));
-/// pgp_key_free (primary_key);
-/// assert (primary_keypair);
-/// primary_signer = pgp_key_pair_as_signer (primary_keypair);
-/// cert = pgp_cert_revoke_in_place (NULL, cert, primary_signer,
-///                               PGP_REASON_FOR_REVOCATION_KEY_COMPROMISED,
-///                               "It was the maid :/");
-/// assert (cert);
-/// pgp_signer_free (primary_signer);
-/// pgp_key_pair_free (primary_keypair);
-///
-/// pgp_revocation_status_t rs = pgp_cert_revoked (cert, policy, 0);
-/// assert (pgp_revocation_status_variant (rs) == PGP_REVOCATION_STATUS_REVOKED);
-/// pgp_revocation_status_free (rs);
-///
-/// pgp_cert_free (cert);
-/// pgp_policy_free (policy);
-/// ```
-#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
-fn pgp_cert_revoke_in_place(errp: Option<&mut *mut crate::error::Error>,
-                           cert: *mut Cert,
-                           primary_signer: *mut Box<dyn crypto::Signer>,
-                           code: c_int,
-                           reason: Option<&c_char>)
-                           -> Maybe<Cert>
-{
-    let cert = cert.move_from_raw();
-    let signer = ffi_param_ref_mut!(primary_signer);
-    let code = int_to_reason_for_revocation(code);
-    let reason = if let Some(reason) = reason {
-        ffi_param_cstr!(reason as *const c_char).to_bytes()
-    } else {
-        b""
-    };
-
-    cert.revoke_in_place(signer.as_mut(), code, reason).move_into_raw(errp)
-}
-
 /// Returns whether the Cert is alive at the specified time.
 ///
 /// If `when` is 0, then the current time is used.
@@ -336,13 +263,12 @@ fn pgp_cert_alive(errp: Option<&mut *mut crate::error::Error>,
 {
     let policy = &**policy.ref_raw();
     ffi_make_fry_from_errp!(errp);
-    ffi_try_status!(cert.ref_raw().alive(policy, maybe_time(when)))
+    let valid_cert = ffi_try_or_status!(
+        cert.ref_raw().with_policy(policy, maybe_time(when)));
+    ffi_try_status!(valid_cert.alive())
 }
 
-/// Changes the Cert's expiration.
-///
-/// Expiry is when the key should expire in seconds relative to the
-/// key's creation (not the current time).
+/// Sets the key to expire at the given time.
 ///
 /// This function consumes `cert` and returns a new `Cert`.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
@@ -350,16 +276,17 @@ fn pgp_cert_set_expiration_time(errp: Option<&mut *mut crate::error::Error>,
                        cert: *mut Cert,
                        policy: *const Policy,
                        primary_signer: *mut Box<dyn crypto::Signer>,
-                       expiry: u32)
+                       expiry: time_t)
     -> Maybe<Cert>
 {
+    ffi_make_fry_from_errp!(errp);
     let policy = &**policy.ref_raw();
     let cert = cert.move_from_raw();
     let signer = ffi_param_ref_mut!(primary_signer);
 
-    cert.set_expiration_time(policy, signer.as_mut(),
-                    Some(std::time::Duration::new(expiry as u64, 0)))
-        .move_into_raw(errp)
+    let sigs = ffi_try_or!(cert.set_expiration_time(policy, None, signer.as_mut(),
+                                                    maybe_time(expiry)), None);
+    cert.merge_packets(sigs).move_into_raw(errp)
 }
 
 /// Returns whether the Cert includes any secret key material.
@@ -372,85 +299,150 @@ fn pgp_cert_is_tsk(cert: *const Cert)
 
 /// Returns an iterator over the Cert's user id bindings.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
-fn pgp_cert_primary_user_id(cert: *const Cert, policy: *const Policy)
+fn pgp_cert_primary_user_id(cert: *const Cert, policy: *const Policy,
+                            when: time_t)
                            -> *mut c_char
 {
     let cert = cert.ref_raw();
     let policy = &**policy.ref_raw();
-    if let Some(binding) = cert.primary_userid(policy, None) {
+    if let Ok(binding) = cert.with_policy(policy, maybe_time(when))
+        .and_then(|valid_cert| valid_cert.primary_userid())
+    {
         ffi_return_string!(binding.userid().value())
     } else {
         ptr::null_mut()
     }
 }
 
-/* UserIDBundle */
+/* UserIDIter */
 
-/// Returns the user id.
-///
-/// This function may fail and return NULL if the user id contains an
-/// interior NUL byte.  We do this rather than complicate the API, as
-/// there is no valid use for such user ids; they must be malicious.
-///
-/// The caller must free the returned value.
-#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
-pub extern "C" fn pgp_user_id_bundle_user_id(
-    binding: *const UserIDBundle)
-    -> *mut c_char
-{
-    let binding = ffi_param_ref!(binding);
-
-    ffi_return_maybe_string!(binding.userid().value())
+/// Wraps a UserIDIter for export via the FFI.
+pub struct UserIDIterWrapper<'a> {
+    pub(crate) // For serialize.rs.
+    iter: Option<ComponentAmalgamationIter<'a, openpgp::packet::UserID>>,
+    // Whether next has been called.
+    next_called: bool,
 }
 
-/// Returns a reference to the self-signature, if any.
+/// Returns an iterator over the Cert's user ids.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
-pub extern "C" fn pgp_user_id_bundle_selfsig(
-    binding: *const UserIDBundle,
-    policy: *const Policy)
-    -> Maybe<Signature>
-{
-    let binding = ffi_param_ref!(binding);
-    let policy = &**policy.ref_raw();
-    binding.binding_signature(policy, None).move_into_raw()
-}
-
-
-/* UserIDBundleIter */
-
-/// Returns an iterator over the Cert's user id bindings.
-#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
-pub extern "C" fn pgp_cert_user_id_binding_iter(cert: *const Cert)
-    -> *mut UserIDBundleIter<'static>
+pub extern "C" fn pgp_cert_user_id_iter(cert: *const Cert)
+    -> *mut UserIDIterWrapper<'static>
 {
     let cert = cert.ref_raw();
-    box_raw!(cert.userids().bundles())
+    box_raw!(UserIDIterWrapper {
+        iter: Some(cert.userids()),
+        next_called: false,
+    })
 }
 
-/// Frees a pgp_user_id_bundle_iter_t.
+/// Changes the iterator to only return keys that are valid at time
+/// `t`.
+///
+/// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
-pub extern "C" fn pgp_user_id_bundle_iter_free(
-    iter: Option<&mut UserIDBundleIter>)
+pub extern "C" fn pgp_cert_user_id_iter_policy<'a>(
+    iter_wrapper: *mut UserIDIterWrapper<'a>,
+    policy: *const Policy,
+    when: time_t)
+    -> *mut ValidUserIDIterWrapper<'static>
+{
+    let policy = policy.ref_raw();
+    let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
+    if iter_wrapper.next_called {
+        panic!("Can't change UserIDIter filter after iterating.");
+    }
+
+    use std::mem::transmute;
+    box_raw!(ValidUserIDIterWrapper {
+        iter: Some(unsafe {
+            transmute(iter_wrapper.iter.take().unwrap()
+                      .with_policy(&**policy, maybe_time(when)))
+        }),
+        next_called: false,
+    })
+}
+
+
+/// Frees a pgp_user_id_iter_t.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
+pub extern "C" fn pgp_cert_user_id_iter_free(
+    iter: Option<&mut UserIDIterWrapper>)
 {
     ffi_free!(iter)
 }
 
-/// Returns the next `UserIDBundle`.
+/// Returns the next `UserIDAmalgamation`.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
-pub extern "C" fn pgp_user_id_bundle_iter_next<'a>(
-    iter: *mut UserIDBundleIter<'a>)
-    -> Option<&'a UserIDBundle>
+pub extern "C" fn pgp_cert_user_id_iter_next<'a>(
+    iter_wrapper: *mut UserIDIterWrapper<'a>)
+    -> Maybe<UserIDAmalgamation<'a>>
 {
-    let iter = ffi_param_ref_mut!(iter);
-    iter.next()
+    let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
+    iter_wrapper.next_called = true;
+
+    if let Some(ua) = iter_wrapper.iter.as_mut().unwrap().next() {
+        Some(ua).move_into_raw()
+    } else {
+        None
+    }
 }
 
-/* cert::KeyIter. */
-
-/// Wraps a KeyIter for export via the FFI.
-pub struct KeyIterWrapper<'a> {
+/// Wraps a ValidKeyAmalgamationIter for export via the FFI.
+pub struct ValidUserIDIterWrapper<'a> {
     pub(crate) // For serialize.rs.
-    iter: KeyIter<'a, openpgp::packet::key::PublicParts>,
+    iter: Option<ValidComponentAmalgamationIter<'a, openpgp::packet::UserID>>,
+    // Whether next has been called.
+    next_called: bool,
+}
+
+/// Returns an iterator over the Cert's user id bundles.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
+pub extern "C" fn pgp_cert_valid_user_id_iter(cert: *const Cert,
+                                          policy: *const Policy, when: time_t)
+    -> *mut ValidUserIDIterWrapper<'static>
+{
+    let cert = cert.ref_raw();
+    let iter = box_raw!(UserIDIterWrapper {
+        iter: Some(cert.userids()),
+        next_called: false,
+    });
+
+    pgp_cert_user_id_iter_policy(iter, policy, when)
+}
+
+/// Frees a pgp_user_id_iter_t.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
+pub extern "C" fn pgp_cert_valid_user_id_iter_free(
+    iter: Option<&mut ValidUserIDIterWrapper>)
+{
+    ffi_free!(iter)
+}
+
+/// Returns the next `UserIDAmalgamation`.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
+pub extern "C" fn pgp_cert_valid_user_id_iter_next<'a>(
+    iter_wrapper: *mut ValidUserIDIterWrapper<'a>)
+    -> Maybe<ValidUserIDAmalgamation<'a>>
+{
+    let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
+    iter_wrapper.next_called = true;
+
+    if let Some(ua) = iter_wrapper.iter.as_mut().unwrap().next() {
+        Some(ua).move_into_raw()
+    } else {
+        None
+    }
+}
+
+
+/* cert::KeyAmalgamationIter. */
+
+/// Wraps a KeyAmalgamationIter for export via the FFI.
+pub struct KeyAmalgamationIterWrapper<'a> {
+    pub(crate) // For serialize.rs.
+    iter: Option<KeyAmalgamationIter<'a, openpgp::packet::key::PublicParts,
+                         openpgp::packet::key::UnspecifiedRole>>,
     // Whether next has been called.
     next_called: bool,
 }
@@ -461,11 +453,11 @@ pub struct KeyIterWrapper<'a> {
 /// subkeys.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_key_iter(cert: *const Cert)
-    -> *mut KeyIterWrapper<'static>
+    -> *mut KeyAmalgamationIterWrapper<'static>
 {
     let cert = cert.ref_raw();
-    box_raw!(KeyIterWrapper {
-        iter: cert.keys(),
+    box_raw!(KeyAmalgamationIterWrapper {
+        iter: Some(cert.keys()),
         next_called: false,
     })
 }
@@ -473,7 +465,7 @@ pub extern "C" fn pgp_cert_key_iter(cert: *const Cert)
 /// Frees a pgp_cert_key_iter_t.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_key_iter_free(
-    iter: Option<&mut KeyIterWrapper>)
+    iter: Option<&mut KeyAmalgamationIterWrapper>)
 {
     ffi_free!(iter)
 }
@@ -483,16 +475,17 @@ pub extern "C" fn pgp_cert_key_iter_free(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_key_iter_secret<'a>(
-    iter_wrapper: *mut KeyIterWrapper<'a>)
+    iter_wrapper: *mut KeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = unsafe { std::mem::transmute(tmp.secret()) };
+    use std::mem::transmute;
+    iter_wrapper.iter = Some(unsafe {
+        transmute(iter_wrapper.iter.take().unwrap().secret())
+    });
 }
 
 /// Changes the iterator to only return keys that have unencrypted
@@ -501,17 +494,17 @@ pub extern "C" fn pgp_cert_key_iter_secret<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_key_iter_unencrypted_secret<'a>(
-    iter_wrapper: *mut KeyIterWrapper<'a>)
+    iter_wrapper: *mut KeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter =
-        unsafe { std::mem::transmute(tmp.unencrypted_secret()) };
+    use std::mem::transmute;
+    iter_wrapper.iter = Some(unsafe {
+        transmute(iter_wrapper.iter.take().unwrap().unencrypted_secret())
+    });
 }
 
 /// Changes the iterator to only return keys that are valid at time
@@ -520,24 +513,23 @@ pub extern "C" fn pgp_cert_key_iter_unencrypted_secret<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_key_iter_policy<'a>(
-    iter_wrapper: *mut KeyIterWrapper<'a>,
+    iter_wrapper: *mut KeyAmalgamationIterWrapper<'a>,
     policy: *const Policy,
     when: time_t)
-    -> *mut ValidKeyIterWrapper<'static>
+    -> *mut ValidKeyAmalgamationIterWrapper<'static>
 {
     let policy = policy.ref_raw();
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-
-    box_raw!(ValidKeyIterWrapper {
-        iter: unsafe {
-            std::mem::transmute(tmp.with_policy(&**policy, maybe_time(when)))
-        },
+    use std::mem::transmute;
+    box_raw!(ValidKeyAmalgamationIterWrapper {
+        iter: Some(unsafe {
+            transmute(iter_wrapper.iter.take().unwrap()
+                      .with_policy(&**policy, maybe_time(when)))
+        }),
         next_called: false,
     })
 }
@@ -553,24 +545,24 @@ pub extern "C" fn pgp_cert_key_iter_policy<'a>(
 /// *rso.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_key_iter_next<'a>(
-    iter_wrapper: *mut KeyIterWrapper<'a>)
-    -> Maybe<Key>
+    iter_wrapper: *mut KeyAmalgamationIterWrapper<'a>)
+    -> Maybe<KeyAmalgamation<'a>>
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     iter_wrapper.next_called = true;
 
-    if let Some(ka) = iter_wrapper.iter.next() {
-        Some(ka.key().mark_parts_unspecified_ref().mark_role_unspecified_ref())
-            .move_into_raw()
+    if let Some(ka) = iter_wrapper.iter.as_mut().unwrap().next() {
+        Some(ka.parts_into_unspecified()).move_into_raw()
     } else {
         None
     }
 }
 
-/// Wraps a ValidKeyIter for export via the FFI.
-pub struct ValidKeyIterWrapper<'a> {
+/// Wraps a ValidKeyAmalgamationIter for export via the FFI.
+pub struct ValidKeyAmalgamationIterWrapper<'a> {
     pub(crate) // For serialize.rs.
-    iter: ValidKeyIter<'a, openpgp::packet::key::PublicParts>,
+    iter: Option<ValidKeyAmalgamationIter<'a, openpgp::packet::key::PublicParts,
+                              openpgp::packet::key::UnspecifiedRole>>,
     // Whether next has been called.
     next_called: bool,
 }
@@ -583,11 +575,11 @@ pub struct ValidKeyIterWrapper<'a> {
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter(cert: *const Cert,
                                           policy: *const Policy, when: time_t)
-    -> *mut ValidKeyIterWrapper<'static>
+    -> *mut ValidKeyAmalgamationIterWrapper<'static>
 {
     let cert = cert.ref_raw();
-    let iter = box_raw!(KeyIterWrapper {
-        iter: cert.keys(),
+    let iter = box_raw!(KeyAmalgamationIterWrapper {
+        iter: Some(cert.keys()),
         next_called: false,
     });
 
@@ -597,7 +589,7 @@ pub extern "C" fn pgp_cert_valid_key_iter(cert: *const Cert,
 /// Frees a pgp_cert_key_iter_t.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_free(
-    iter: Option<&mut ValidKeyIterWrapper>)
+    iter: Option<&mut ValidKeyAmalgamationIterWrapper>)
 {
     ffi_free!(iter)
 }
@@ -607,16 +599,17 @@ pub extern "C" fn pgp_cert_valid_key_iter_free(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_secret<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>)
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change ValidKeyIter filter after iterating.");
+        panic!("Can't change ValidKeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = unsafe { std::mem::transmute(tmp.secret()) };
+    use std::mem::transmute;
+    iter_wrapper.iter = Some(unsafe {
+        transmute(iter_wrapper.iter.take().unwrap().secret())
+    });
 }
 
 /// Changes the iterator to only return keys that have unencrypted
@@ -625,17 +618,17 @@ pub extern "C" fn pgp_cert_valid_key_iter_secret<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_unencrypted_secret<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>)
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change ValidKeyIter filter after iterating.");
+        panic!("Can't change ValidKeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter =
-        unsafe { std::mem::transmute(tmp.unencrypted_secret()) };
+    use std::mem::transmute;
+    iter_wrapper.iter = Some(unsafe {
+        transmute(iter_wrapper.iter.take().unwrap().unencrypted_secret())
+    });
 }
 
 /// Changes the iterator to only return keys that are certification
@@ -649,16 +642,15 @@ pub extern "C" fn pgp_cert_valid_key_iter_unencrypted_secret<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_for_certification<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>)
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = tmp.for_certification();
+    iter_wrapper.iter =
+        Some(iter_wrapper.iter.take().unwrap().for_certification());
 }
 
 /// Changes the iterator to only return keys that are certification
@@ -672,16 +664,15 @@ pub extern "C" fn pgp_cert_valid_key_iter_for_certification<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_for_signing<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>)
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = tmp.for_signing();
+    iter_wrapper.iter =
+        Some(iter_wrapper.iter.take().unwrap().for_signing());
 }
 
 /// Changes the iterator to only return keys that are capable of
@@ -695,16 +686,15 @@ pub extern "C" fn pgp_cert_valid_key_iter_for_signing<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_for_storage_encryption<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>)
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = tmp.for_storage_encryption();
+    iter_wrapper.iter =
+        Some(iter_wrapper.iter.take().unwrap().for_storage_encryption());
 }
 
 /// Changes the iterator to only return keys that are capable of
@@ -718,16 +708,15 @@ pub extern "C" fn pgp_cert_valid_key_iter_for_storage_encryption<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_for_transport_encryption<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>)
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = tmp.for_transport_encryption();
+    iter_wrapper.iter =
+        Some(iter_wrapper.iter.take().unwrap().for_transport_encryption());
 }
 
 /// Changes the iterator to only return keys that are alive.
@@ -738,16 +727,14 @@ pub extern "C" fn pgp_cert_valid_key_iter_for_transport_encryption<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_alive<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>)
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = tmp.alive();
+    iter_wrapper.iter = Some(iter_wrapper.iter.take().unwrap().alive());
 }
 
 /// Changes the iterator to only return keys whose revocation status
@@ -756,17 +743,16 @@ pub extern "C" fn pgp_cert_valid_key_iter_alive<'a>(
 /// Note: you may not call this function after starting to iterate.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_revoked<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>,
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>,
     revoked: bool)
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     if iter_wrapper.next_called {
-        panic!("Can't change KeyIter filter after iterating.");
+        panic!("Can't change KeyAmalgamationIter filter after iterating.");
     }
 
-    use std::mem;
-    let tmp = mem::replace(&mut iter_wrapper.iter, unsafe { mem::zeroed() });
-    iter_wrapper.iter = tmp.revoked(Some(revoked));
+    iter_wrapper.iter =
+        Some(iter_wrapper.iter.take().unwrap().revoked(revoked));
 }
 
 /// Returns the next valid key.  Returns NULL if there are no more
@@ -778,18 +764,17 @@ pub extern "C" fn pgp_cert_valid_key_iter_revoked<'a>(
 /// *rso.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
 pub extern "C" fn pgp_cert_valid_key_iter_next<'a>(
-    iter_wrapper: *mut ValidKeyIterWrapper<'a>,
+    iter_wrapper: *mut ValidKeyAmalgamationIterWrapper<'a>,
     sigo: Option<&mut *mut Signature>,
     rso: Option<&mut *mut RevocationStatus<'a>>)
-    -> Maybe<Key>
+    -> Maybe<ValidKeyAmalgamation<'a>>
 {
     let iter_wrapper = ffi_param_ref_mut!(iter_wrapper);
     iter_wrapper.next_called = true;
 
-    if let Some(ka) = iter_wrapper.iter.next() {
+    if let Some(ka) = iter_wrapper.iter.as_mut().unwrap().next() {
         let sig = ka.binding_signature();
-        let rs = ka.revoked();
-        let key = ka.key();
+        let rs = ka.revocation_status();
 
         if let Some(ptr) = sigo {
             *ptr = sig.move_into_raw();
@@ -799,10 +784,7 @@ pub extern "C" fn pgp_cert_valid_key_iter_next<'a>(
             *ptr = rs.move_into_raw();
         }
 
-        let key
-            = key.mark_parts_unspecified_ref().mark_role_unspecified_ref();
-
-        Some(key).move_into_raw()
+        Some(ka.parts_into_unspecified()).move_into_raw()
     } else {
         None
     }
@@ -810,7 +792,7 @@ pub extern "C" fn pgp_cert_valid_key_iter_next<'a>(
 
 /// Wraps a CertParser for export via the FFI.
 pub struct CertParserWrapper<'a> {
-    parser: CertParser<'a, std::vec::IntoIter<self::openpgp::Packet>>,
+    parser: CertParser<'a>,
 }
 
 /// Returns a CertParser.
@@ -819,7 +801,7 @@ pub struct CertParserWrapper<'a> {
 /// concatenated together.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
 fn pgp_cert_parser_from_bytes(errp: Option<&mut *mut crate::error::Error>,
-                             buf: *mut u8, len: size_t)
+                             buf: *const u8, len: size_t)
     -> *mut CertParserWrapper<'static>
 {
     ffi_make_fry_from_errp!(errp);
@@ -837,7 +819,7 @@ fn pgp_cert_parser_from_packet_parser(ppr: *mut PacketParserResult<'static>)
     -> *mut CertParserWrapper<'static>
 {
     let ppr = ffi_param_move!(ppr);
-    let parser = CertParser::from_packet_parser(*ppr);
+    let parser = CertParser::from(*ppr);
     box_raw!(CertParserWrapper { parser: parser })
 }
 
@@ -1004,6 +986,17 @@ pub extern "C" fn pgp_cert_builder_add_certification_subkey
     *certb = box_raw!(certb_);
 }
 
+/// Sets the creation time.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
+pub extern "C" fn pgp_cert_builder_set_creation_time
+    (certb: *mut *mut CertBuilder, when: time_t)
+{
+    let certb = ffi_param_ref_mut!(certb);
+    let certb_ = ffi_param_move!(*certb);
+    let certb_ = certb_.set_creation_time(maybe_time(when));
+    *certb = box_raw!(certb_);
+}
+
 /// Generates the actual Cert.
 ///
 /// Consumes `certb`.
@@ -1025,7 +1018,7 @@ pub extern "C" fn pgp_cert_builder_generate
         },
         Err(e) => {
             *cert_out = None;
-            Err::<(), failure::Error>(e).move_into_raw(errp)
+            Err::<(), anyhow::Error>(e).move_into_raw(errp)
         },
     }
 }
