@@ -16,21 +16,21 @@ use libc::{c_int, c_void, time_t};
 extern crate sequoia_openpgp as openpgp;
 
 use self::openpgp::{
-    cert::components::Amalgamation,
+    cert::prelude::*,
     crypto::SessionKey,
     types::SymmetricAlgorithm,
     packet::{
         PKESK,
         SKESK,
     },
+    parse::Parse,
 };
 use self::openpgp::parse::stream::{
     self,
     DecryptionHelper,
-    Decryptor,
+    DecryptorBuilder,
     VerificationHelper,
-    Verifier,
-    DetachedVerifier,
+    VerifierBuilder,
 };
 
 use crate::Maybe;
@@ -42,7 +42,7 @@ use crate::RefMutRaw;
 use crate::maybe_time;
 
 use super::super::{
-    error::Status,
+    error::{Error, Status},
     crypto,
     io,
     keyid,
@@ -60,14 +60,15 @@ pub struct MessageStructure<'a>(stream::MessageStructure<'a>);
 
 /// Iterates over the message structure.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
-fn pgp_message_structure_iter(structure: *const MessageStructure)
-                              -> *mut MessageStructureIter {
-    structure.ref_raw().iter().move_into_raw()
+fn pgp_message_structure_into_iter(structure: *mut MessageStructure)
+                                   -> *mut MessageStructureIter {
+    structure.move_from_raw().into_iter().move_into_raw()
 }
 
 /// Iterates over the message structure.
 #[crate::ffi_wrapper_type(prefix = "pgp_", derive = "Iterator(MessageLayer)")]
-pub struct MessageStructureIter<'a>(stream::MessageStructureIter<'a>);
+pub struct MessageStructureIter<'a>(
+    std::vec::IntoIter<stream::MessageLayer<'a>>);
 
 /// Represents a layer of the message structure.
 #[crate::ffi_wrapper_type(prefix = "pgp_", derive = "Debug")]
@@ -169,12 +170,144 @@ pub struct VerificationResult<'a>(stream::VerificationResult<'a>);
 fn pgp_verification_result_variant(result: *const VerificationResult)
     -> c_int
 {
-    use self::stream::VerificationResult::*;
+    use self::stream::VerificationError::*;
     match result.ref_raw() {
-        GoodChecksum { .. } => 1,
-        MissingKey { .. } => 2,
-        NotAlive { .. } => 3,
-        Error { .. } => 4,
+        Ok(_) => 0,
+        Err(MalformedSignature { .. }) => 1,
+        Err(MissingKey { .. }) => 2,
+        Err(UnboundKey { .. }) => 3,
+        Err(BadKey { .. }) => 4,
+        Err(BadSignature { .. }) => 5,
+    }
+}
+
+/// Decomposes a `VerificationResult::Ok(GoodChecksum)`.
+///
+/// Returns `true` iff the given value is a
+/// `VerificationResult::Ok(GoodChecksum)`, and returns the variants
+/// members in `sig_r` and the like iff `sig_r != NULL`.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
+fn pgp_verification_result_good_checksum<'a>(
+    result: *const VerificationResult<'a>,
+    sig_r: Maybe<*mut Signature>,
+    cert_r: Maybe<*mut Cert>,
+    key_r: Maybe<*mut Key>,
+    binding_r: Maybe<*mut Signature>,
+    revocation_status_r:
+    Maybe<*mut RevocationStatus<'a>>)
+    -> bool
+{
+    use self::stream::GoodChecksum;
+    if let Ok(GoodChecksum { sig, ka }) = result.ref_raw() {
+        if let Some(mut p) = sig_r {
+            *unsafe { p.as_mut() } = sig.move_into_raw();
+        }
+        if let Some(mut p) = cert_r {
+            *unsafe { p.as_mut() } = ka.cert().cert().move_into_raw();
+        }
+        if let Some(mut p) = key_r {
+            *unsafe { p.as_mut() } = {
+                let key = ka.key()
+                    .parts_as_unspecified()
+                    .role_as_unspecified();
+                key.move_into_raw()
+            };
+        }
+        if let Some(mut p) = binding_r {
+            *unsafe { p.as_mut() } =
+                ka.binding_signature().move_into_raw();
+        }
+        if let Some(mut p) = revocation_status_r {
+            *unsafe { p.as_mut() } = ka.revocation_status().move_into_raw();
+        }
+        true
+    } else {
+        false
+    }
+}
+
+/// Decomposes a
+/// `VerificationResult::Err(VerificationError::MalformedSignature {
+/// .. })`.
+///
+/// Returns `true` iff the given value is a
+/// `VerificationResult::Err(VerificationError::MalformedSignature {
+/// .. })`, and returns the variants members in `sig_r` and the like
+/// iff `sig_r != NULL`.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
+fn pgp_verification_result_malformed_signature<'a>(
+    result: *const VerificationResult<'a>,
+    sig_r: Maybe<*mut Signature>,
+    error_r: Maybe<*mut Error>)
+    -> bool
+{
+    use self::stream::VerificationError::*;
+    if let Err(MalformedSignature { sig, error }) = result.ref_raw() {
+        if let Some(mut p) = sig_r {
+            *unsafe { p.as_mut() } = sig.move_into_raw();
+        }
+        if let Some(mut p) = error_r {
+            *unsafe { p.as_mut() } = error.move_into_raw();
+        }
+        true
+    } else {
+        false
+    }
+}
+
+/// Decomposes a
+/// `VerificationResult::Err(VerificationError::MissingKey { .. })`.
+///
+/// Returns `true` iff the given value is a
+/// `VerificationResult::Err(VerificationError::MissingKey { .. })`,
+/// and returns the variants members in `sig_r` and the like iff
+/// `sig_r != NULL`.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
+fn pgp_verification_result_missing_key<'a>(
+    result: *const VerificationResult<'a>,
+    sig_r: Maybe<*mut Signature>)
+    -> bool
+{
+    use self::stream::VerificationError::*;
+    if let Err(MissingKey { sig }) = result.ref_raw() {
+        if let Some(mut p) = sig_r {
+            *unsafe { p.as_mut() } = sig.move_into_raw();
+        }
+        true
+    } else {
+        false
+    }
+}
+
+/// Decomposes a
+/// `VerificationResult::Err(VerificationError::UnboundKey { .. })`.
+///
+/// Returns `true` iff the given value is a
+/// `VerificationResult::Err(VerificationError::UnboundKey { .. })`,
+/// and returns the variants members in `sig_r` and the like iff
+/// `sig_r != NULL`.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
+fn pgp_verification_result_unbound_key<'a>(
+    result: *const VerificationResult<'a>,
+    sig_r: Maybe<*mut Signature>,
+    cert_r: Maybe<*mut Cert>,
+    error_r: Maybe<*mut Error>)
+    -> bool
+{
+    use self::stream::VerificationError::*;
+    if let Err(UnboundKey { sig, cert, error }) = result.ref_raw() {
+        if let Some(mut p) = sig_r {
+            *unsafe { p.as_mut() } = sig.move_into_raw();
+        }
+        if let Some(mut p) = cert_r {
+            *unsafe { p.as_mut() } = cert.move_into_raw();
+        }
+        if let Some(mut p) = error_r {
+            *unsafe { p.as_mut() } = error.move_into_raw();
+        }
+        true
+    } else {
+        false
     }
 }
 
@@ -188,22 +321,23 @@ fn $fn_name<'a>(
     key_r: Maybe<*mut Key>,
     binding_r: Maybe<*mut Signature>,
     revocation_status_r:
-    Maybe<*mut RevocationStatus<'a>>)
+    Maybe<*mut RevocationStatus<'a>>,
+    error_r: Maybe<*mut Error>)
     -> bool
 {
-    use self::stream::VerificationResult::*;
-    if let $variant { sig, cert, ka } = result.ref_raw() {
+    use self::stream::VerificationError::*;
+    if let Err($variant { sig, ka, error }) = result.ref_raw() {
         if let Some(mut p) = sig_r {
             *unsafe { p.as_mut() } = sig.move_into_raw();
         }
         if let Some(mut p) = cert_r {
-            *unsafe { p.as_mut() } = cert.move_into_raw();
+            *unsafe { p.as_mut() } = ka.cert().cert().move_into_raw();
         }
         if let Some(mut p) = key_r {
             *unsafe { p.as_mut() } = {
                 let key = ka.key()
-                    .mark_parts_unspecified_ref()
-                    .mark_role_unspecified_ref();
+                    .parts_as_unspecified()
+                    .role_as_unspecified();
                 key.move_into_raw()
             };
         }
@@ -212,85 +346,9 @@ fn $fn_name<'a>(
                 ka.binding_signature().move_into_raw();
         }
         if let Some(mut p) = revocation_status_r {
-            *unsafe { p.as_mut() } = ka.revoked().move_into_raw();
+            *unsafe { p.as_mut() } = ka.revocation_status().move_into_raw();
         }
-        true
-    } else {
-        false
-    }
-}
-    }
-}
-
-/// Decomposes a `VerificationResult::GoodChecksum`.
-///
-/// Returns `true` iff the given value is a
-/// `VerificationResult::GoodChecksum`, and returns the variants members
-/// in `sig_r` and the like iff `sig_r != NULL`.
-make_decomposition_fn!(pgp_verification_result_good_checksum, GoodChecksum);
-
-/// Decomposes a `VerificationResult::NotAlive`.
-///
-/// Returns `true` iff the given value is a
-/// `VerificationResult::NotAlive`, and returns the variant's members
-/// in `sig_r` and the like iff `sig_r != NULL`.
-#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
-fn pgp_verification_result_not_alive<'a>(
-    result: *const VerificationResult<'a>,
-    sig_r: Maybe<*mut Signature>)
-    -> bool
-{
-    use self::stream::VerificationResult::*;
-    if let NotAlive { sig, .. } = result.ref_raw() {
-        if let Some(mut p) = sig_r {
-            *unsafe { p.as_mut() } = sig.move_into_raw();
-        }
-        true
-    } else {
-        false
-    }
-}
-
-/// Decomposes a `VerificationResult::MissingKey`.
-///
-/// Returns `true` iff the given value is a
-/// `VerificationResult::MissingKey`, and returns the variants members
-/// in `sig_r` and the like iff `sig_r != NULL`.
-#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
-fn pgp_verification_result_missing_key<'a>(
-    result: *const VerificationResult<'a>,
-    sig_r: Maybe<*mut Signature>)
-    -> bool
-{
-    use self::stream::VerificationResult::*;
-    if let MissingKey { sig, .. } = result.ref_raw() {
-        if let Some(mut p) = sig_r {
-            *unsafe { p.as_mut() } = sig.move_into_raw();
-        }
-        true
-    } else {
-        false
-    }
-}
-
-/// Decomposes a `VerificationResult::Error`.
-///
-/// Returns `true` iff the given value is a
-/// `VerificationResult::Error`, and returns the variants members
-/// in `sig_r` and the like iff `sig_r != NULL`.
-#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
-fn pgp_verification_result_error<'a>(
-    result: *const VerificationResult<'a>,
-    sig_r: Maybe<*mut Signature>,
-    err_r: Maybe<*mut crate::error::Error>)
-    -> bool
-{
-    use self::stream::VerificationResult::*;
-    if let Error { sig, error, .. } = result.ref_raw() {
-        if let Some(mut p) = sig_r {
-            *unsafe { p.as_mut() } = sig.move_into_raw();
-        }
-        if let Some(mut p) = err_r {
+        if let Some(mut p) = error_r {
             *unsafe { p.as_mut() } = error.move_into_raw();
         }
         true
@@ -298,6 +356,26 @@ fn pgp_verification_result_error<'a>(
         false
     }
 }
+    }
+}
+
+/// Decomposes a
+/// `VerificationResult::Err(VerificationError::BadKey { .. })`.
+///
+/// Returns `true` iff the given value is a
+/// `VerificationResult::Err(VerificationError::BadKey { .. })`,
+/// and returns the variants members in `sig_r` and the like iff
+/// `sig_r != NULL`.
+make_decomposition_fn!(pgp_verification_result_bad_key, BadKey);
+
+/// Decomposes a
+/// `VerificationResult::Err(VerificationError::BadSignature { .. })`.
+///
+/// Returns `true` iff the given value is a
+/// `VerificationResult::Err(VerificationError::BadSignature { .. })`,
+/// and returns the variants members in `sig_r` and the like iff
+/// `sig_r != NULL`.
+make_decomposition_fn!(pgp_verification_result_bad_signature, BadSignature);
 
 /// Passed as the first argument to the callbacks used by pgp_verify
 /// and pgp_decrypt.
@@ -335,9 +413,10 @@ type InspectCallback = fn(*mut HelperCookie, *const PacketParser) -> Status;
 type DecryptCallback = fn(*mut HelperCookie,
                           *const *const PKESK, usize,
                           *const *const SKESK, usize,
+                          u8, // XXX SymmetricAlgorithm
                           extern "C" fn (*mut c_void, u8,
                                               *const crypto::SessionKey)
-                                              -> Status,
+                                              -> bool,
                           *mut c_void,
                           *mut Maybe<super::super::fingerprint::Fingerprint>)
                           -> Status;
@@ -352,28 +431,44 @@ type CheckCallback = fn(*mut HelperCookie,
 
 // This fetches keys and computes the validity of the verification.
 struct VHelper {
-    get_public_keys_cb: GetPublicKeysCallback,
+    inspect_cb: Option<InspectCallback>,
+    get_certs_cb: GetPublicKeysCallback,
     check_signatures_cb: CheckCallback,
     cookie: *mut HelperCookie,
 }
 
 impl VHelper {
-    fn new(get_public_keys: GetPublicKeysCallback,
+    fn new(inspect_cb: Option<InspectCallback>,
+           get_certs: GetPublicKeysCallback,
            check_signatures: CheckCallback,
            cookie: *mut HelperCookie)
        -> Self
     {
         VHelper {
-            get_public_keys_cb: get_public_keys,
+            inspect_cb,
+            get_certs_cb: get_certs,
             check_signatures_cb: check_signatures,
-            cookie: cookie,
+            cookie,
         }
     }
 }
 
 impl VerificationHelper for VHelper {
-    fn get_public_keys(&mut self, ids: &[openpgp::KeyHandle])
-        -> Result<Vec<openpgp::Cert>, failure::Error>
+    fn inspect(&mut self, pp: &PacketParser) -> openpgp::Result<()> {
+        if let Some(cb) = self.inspect_cb {
+            match cb(self.cookie, pp) {
+                Status::Success => Ok(()),
+                // XXX: Convert the status to an error better.
+                status => Err(anyhow::anyhow!(
+                    "Inspect Callback returned an error: {:?}", status).into()),
+            }
+        } else {
+            Ok(())
+        }
+    }
+
+    fn get_certs(&mut self, ids: &[openpgp::KeyHandle])
+        -> Result<Vec<openpgp::Cert>, anyhow::Error>
     {
         // The size of ID is not known in C.  Convert to KeyID, and
         // move it to C.
@@ -386,7 +481,7 @@ impl VerificationHelper for VHelper {
 
         let mut free : FreeCallback = |_| {};
 
-        let result = (self.get_public_keys_cb)(
+        let result = (self.get_certs_cb)(
             self.cookie,
             ids.as_ptr(), ids.len(),
             &mut cert_refs_raw, &mut cert_refs_raw_len as *mut usize,
@@ -418,7 +513,7 @@ impl VerificationHelper for VHelper {
     }
 
     fn check(&mut self, structure: stream::MessageStructure)
-        -> Result<(), failure::Error>
+        -> Result<(), anyhow::Error>
     {
         let result = (self.check_signatures_cb)(self.cookie,
                                                 structure.move_into_raw());
@@ -458,7 +553,7 @@ impl VerificationHelper for VHelper {
 /// };
 ///
 /// static pgp_status_t
-/// get_public_keys_cb (void *cookie_opaque,
+/// get_certs_cb (void *cookie_opaque,
 ///                     pgp_keyid_t *keyids, size_t keyids_len,
 ///                     pgp_cert_t **certs, size_t *certs_len,
 ///                     void (**our_free)(void *))
@@ -477,7 +572,7 @@ impl VerificationHelper for VHelper {
 /// check_cb (void *cookie_opaque, pgp_message_structure_t structure)
 /// {
 ///   pgp_message_structure_iter_t iter =
-///     pgp_message_structure_iter (structure);
+///     pgp_message_structure_into_iter (structure);
 ///   pgp_message_layer_t layer = pgp_message_structure_iter_next (iter);
 ///   assert (layer);
 ///   assert (pgp_message_layer_compression (layer, NULL));
@@ -498,7 +593,6 @@ impl VerificationHelper for VHelper {
 ///   pgp_verification_result_iter_free (results);
 ///   pgp_message_layer_free (layer);
 ///   pgp_message_structure_iter_free (iter);
-///   pgp_message_structure_free (structure);
 ///   return PGP_STATUS_SUCCESS;
 /// }
 ///
@@ -523,7 +617,7 @@ impl VerificationHelper for VHelper {
 ///     .key = cert,  /* Move.  */
 ///   };
 ///   plaintext = pgp_verifier_new (NULL, policy, source,
-///                                 get_public_keys_cb, check_cb,
+///                                 get_certs_cb, check_cb, NULL,
 ///                                 &cookie, 1554542219);
 ///   assert (source);
 ///
@@ -542,20 +636,31 @@ impl VerificationHelper for VHelper {
 fn pgp_verifier_new<'a>(errp: Option<&mut *mut crate::error::Error>,
                         policy: *const Policy,
                         input: *mut io::Reader,
-                        get_public_keys: GetPublicKeysCallback,
+                        get_certs: GetPublicKeysCallback,
                         check: CheckCallback,
+                        inspect: Option<InspectCallback>,
                         cookie: *mut HelperCookie,
                         time: time_t)
                         -> Maybe<io::Reader>
 {
+    ffi_make_fry_from_errp!(errp);
     let policy = policy.ref_raw().as_ref();
-    let helper = VHelper::new(get_public_keys, check, cookie);
+    let helper = VHelper::new(inspect, get_certs, check, cookie);
 
-    Verifier::from_reader(policy, input.ref_mut_raw(), helper, maybe_time(time))
+    ffi_try_or!(VerifierBuilder::from_reader(input.ref_mut_raw()), None)
+        .with_policy(policy, maybe_time(time), helper)
         .map(|r| io::ReaderKind::Generic(Box::new(r)))
         .move_into_raw(errp)
 }
 
+/// Verifies a detached signature.
+///
+/// Wraps [`sequoia-openpgp::parse::stream::DetachedVerifier`].
+///
+///   [`sequoia-openpgp::parse::stream::DetachedVerifier`]: ../../../../sequoia_openpgp/parse/stream/struct.DetachedVerifier.html
+#[crate::ffi_wrapper_type(prefix = "pgp_")]
+pub struct DetachedVerifier(openpgp::parse::stream::DetachedVerifier<'static, VHelper>);
+
 /// Verifies a detached OpenPGP signature.
 ///
 /// # Example
@@ -576,7 +681,7 @@ fn pgp_verifier_new<'a>(errp: Option<&mut *mut crate::error::Error>,
 /// };
 ///
 /// static pgp_status_t
-/// get_public_keys_cb (void *cookie_opaque,
+/// get_certs_cb (void *cookie_opaque,
 ///                     pgp_keyid_t *keyids, size_t keyids_len,
 ///                     pgp_cert_t **certs, size_t *certs_len,
 ///                     void (**our_free)(void *))
@@ -595,7 +700,7 @@ fn pgp_verifier_new<'a>(errp: Option<&mut *mut crate::error::Error>,
 /// check_cb (void *cookie_opaque, pgp_message_structure_t structure)
 /// {
 ///   pgp_message_structure_iter_t iter =
-///     pgp_message_structure_iter (structure);
+///     pgp_message_structure_into_iter (structure);
 ///   pgp_message_layer_t layer = pgp_message_structure_iter_next (iter);
 ///   assert (layer);
 ///   pgp_verification_result_iter_t results;
@@ -612,19 +717,17 @@ fn pgp_verifier_new<'a>(errp: Option<&mut *mut crate::error::Error>,
 ///   pgp_verification_result_iter_free (results);
 ///   pgp_message_layer_free (layer);
 ///   pgp_message_structure_iter_free (iter);
-///   pgp_message_structure_free (structure);
 ///   return PGP_STATUS_SUCCESS;
 /// }
 ///
 /// int
 /// main (int argc, char **argv)
 /// {
+///   pgp_status_t rc;
 ///   pgp_cert_t cert;
 ///   pgp_reader_t signature;
 ///   pgp_reader_t source;
-///   pgp_reader_t plaintext;
-///   uint8_t buf[128];
-///   ssize_t nread;
+///   pgp_detached_verifier_t verifier;
 ///   pgp_policy_t policy = pgp_standard_policy ();
 ///
 ///   cert = pgp_cert_from_file (NULL,
@@ -643,17 +746,15 @@ fn pgp_verifier_new<'a>(errp: Option<&mut *mut crate::error::Error>,
 ///   struct verify_cookie cookie = {
 ///     .key = cert,  /* Move.  */
 ///   };
-///   plaintext = pgp_detached_verifier_new (NULL, policy, signature, source,
-///     get_public_keys_cb, check_cb,
+///   verifier = pgp_detached_verifier_new (NULL, policy, signature,
+///     get_certs_cb, check_cb, NULL,
 ///     &cookie, 1554542219);
-///   assert (source);
+///   assert (verifier);
 ///
-///   nread = pgp_reader_read (NULL, plaintext, buf, sizeof buf);
-///   assert (nread >= 42);
-///   assert (
-///     memcmp (buf, "A Cypherpunk's Manifesto\nby Eric Hughes\n", 40) == 0);
+///   rc = pgp_detached_verifier_verify (NULL, verifier, source);
+///   assert (rc == PGP_STATUS_SUCCESS);
 ///
-///   pgp_reader_free (plaintext);
+///   pgp_detached_verifier_free (verifier);
 ///   pgp_reader_free (source);
 ///   pgp_reader_free (signature);
 ///   pgp_policy_free (policy);
@@ -664,32 +765,46 @@ fn pgp_verifier_new<'a>(errp: Option<&mut *mut crate::error::Error>,
 fn pgp_detached_verifier_new<'a>(errp: Option<&mut *mut crate::error::Error>,
                                  policy: *const Policy,
                                  signature_input: *mut io::Reader,
-                                 input: *mut io::Reader,
-                                 get_public_keys: GetPublicKeysCallback,
+                                 get_certs: GetPublicKeysCallback,
                                  check: CheckCallback,
+                                 inspect: Option<InspectCallback>,
                                  cookie: *mut HelperCookie,
                                  time: time_t)
-                                 -> Maybe<io::Reader>
+                                 -> Maybe<DetachedVerifier>
 {
+    ffi_make_fry_from_errp!(errp);
     let policy = policy.ref_raw().as_ref();
 
-    let helper = VHelper::new(get_public_keys, check, cookie);
+    let helper = VHelper::new(inspect, get_certs, check, cookie);
 
-    DetachedVerifier::from_reader(policy, signature_input.ref_mut_raw(),
-                                  input.ref_mut_raw(), helper, maybe_time(time))
-        .map(|r| io::ReaderKind::Generic(Box::new(r)))
+    ffi_try_or!(openpgp::parse::stream::DetachedVerifierBuilder::from_reader(
+        signature_input.ref_mut_raw()), None)
+        .with_policy(policy, maybe_time(time), helper)
         .move_into_raw(errp)
 }
 
+/// Verifies `data` using `verifier`.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle] pub extern "C"
+fn pgp_detached_verifier_verify(errp: Option<&mut *mut crate::error::Error>,
+                                verifier: *mut DetachedVerifier,
+                                data: *mut io::Reader)
+    -> Status
+{
+    ffi_make_fry_from_errp!(errp);
+    ffi_try_or_status!(
+        verifier.ref_mut_raw().verify_reader(data.ref_mut_raw()));
+    Status::Success
+}
+
+
 
 struct DHelper {
     vhelper: VHelper,
-    inspect_cb: Option<InspectCallback>,
     decrypt_cb: DecryptCallback,
 }
 
 impl DHelper {
-    fn new(get_public_keys: GetPublicKeysCallback,
+    fn new(get_certs: GetPublicKeysCallback,
            decrypt: DecryptCallback,
            check: CheckCallback,
            inspect: Option<InspectCallback>,
@@ -697,45 +812,32 @@ impl DHelper {
        -> Self
     {
         DHelper {
-            vhelper: VHelper::new(get_public_keys, check, cookie),
-            inspect_cb: inspect,
+            vhelper: VHelper::new(inspect, get_certs, check, cookie),
             decrypt_cb: decrypt,
         }
     }
 }
 
 impl VerificationHelper for DHelper {
-    fn get_public_keys(&mut self, ids: &[openpgp::KeyHandle])
-        -> Result<Vec<openpgp::Cert>, failure::Error>
+    fn get_certs(&mut self, ids: &[openpgp::KeyHandle])
+        -> Result<Vec<openpgp::Cert>, anyhow::Error>
     {
-        self.vhelper.get_public_keys(ids)
+        self.vhelper.get_certs(ids)
     }
 
     fn check(&mut self, structure: stream::MessageStructure)
-        -> Result<(), failure::Error>
+        -> Result<(), anyhow::Error>
     {
         self.vhelper.check(structure)
     }
 }
 
 impl DecryptionHelper for DHelper {
-    fn inspect(&mut self, pp: &PacketParser) -> failure::Fallible<()> {
-        if let Some(cb) = self.inspect_cb {
-            match cb(self.vhelper.cookie, pp) {
-                Status::Success => Ok(()),
-                // XXX: Convert the status to an error better.
-                status => Err(failure::format_err!(
-                    "Inspect Callback returned an error: {:?}", status).into()),
-            }
-        } else {
-            Ok(())
-        }
-    }
-
     fn decrypt<D>(&mut self, pkesks: &[PKESK], skesks: &[SKESK],
+                  sym_algo: Option<SymmetricAlgorithm>,
                   mut decrypt: D)
                   -> openpgp::Result<Option<openpgp::Fingerprint>>
-        where D: FnMut(SymmetricAlgorithm, &SessionKey) -> openpgp::Result<()>
+        where D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool
     {
         let mut identity: Maybe<super::super::fingerprint::Fingerprint> = None;
 
@@ -758,17 +860,17 @@ impl DecryptionHelper for DHelper {
 
         extern "C" fn trampoline<D>(data: *mut c_void, algo: u8,
                                          sk: *const crypto::SessionKey)
-                                         -> Status
-            where D: FnMut(SymmetricAlgorithm, &SessionKey)
-                           -> openpgp::Result<()>
+                                         -> bool
+            where D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool
         {
             let closure: &mut D = unsafe { &mut *(data as *mut D) };
-            (*closure)(algo.into(), sk.ref_raw()).into()
+            (*closure)(algo.into(), sk.ref_raw())
         }
 
         let result = (self.decrypt_cb)(
             self.vhelper.cookie,
             pkesks.as_ptr(), pkesks.len(), skesks.as_ptr(), skesks.len(),
+            sym_algo.map(|s| u8::from(s)).unwrap_or(0),
             trampoline::<D>,
             &mut decrypt as *mut _ as *mut c_void,
             &mut identity);
@@ -817,7 +919,7 @@ impl DecryptionHelper for DHelper {
 /// };
 ///
 /// static pgp_status_t
-/// get_public_keys_cb (void *cookie_raw,
+/// get_certs_cb (void *cookie_raw,
 ///                     pgp_keyid_t *keyids, size_t keyids_len,
 ///                     pgp_cert_t **certs, size_t *cert_len,
 ///                     void (**our_free)(void *))
@@ -833,13 +935,12 @@ impl DecryptionHelper for DHelper {
 /// check_cb (void *cookie_opaque, pgp_message_structure_t structure)
 /// {
 ///   pgp_message_structure_iter_t iter =
-///     pgp_message_structure_iter (structure);
+///     pgp_message_structure_into_iter (structure);
 ///   pgp_message_layer_t layer = pgp_message_structure_iter_next (iter);
 ///   assert (layer);
 ///   assert (pgp_message_layer_encryption (layer, NULL, NULL));
 ///   pgp_message_layer_free (layer);
 ///   pgp_message_structure_iter_free (iter);
-///   pgp_message_structure_free (structure);
 ///   return PGP_STATUS_SUCCESS;
 /// }
 ///
@@ -847,11 +948,11 @@ impl DecryptionHelper for DHelper {
 /// decrypt_cb (void *cookie_opaque,
 ///             pgp_pkesk_t *pkesks, size_t pkesk_count,
 ///             pgp_skesk_t *skesks, size_t skesk_count,
+///             uint8_t sym_algo_hint,
 ///             pgp_decryptor_do_decrypt_cb_t *decrypt,
 ///             void *decrypt_cookie,
 ///             pgp_fingerprint_t *identity_out)
 /// {
-///   pgp_status_t rc;
 ///   pgp_error_t err;
 ///   struct decrypt_cookie *cookie = cookie_opaque;
 ///
@@ -863,18 +964,21 @@ impl DecryptionHelper for DHelper {
 ///     pgp_keyid_t keyid = pgp_pkesk_recipient (pkesk);
 ///
 ///     pgp_cert_key_iter_t key_iter = pgp_cert_key_iter (cookie->key);
-///     pgp_key_t key;
-///     while ((key = pgp_cert_key_iter_next (key_iter))) {
+///     pgp_key_amalgamation_t ka;
+///     pgp_key_t key = NULL;
+///     while ((ka = pgp_cert_key_iter_next (key_iter))) {
+///       key = pgp_key_amalgamation_key (ka);
 ///       pgp_keyid_t this_keyid = pgp_key_keyid (key);
 ///       int match = pgp_keyid_equal (this_keyid, keyid);
 ///       pgp_keyid_free (this_keyid);
 ///       if (match)
 ///         break;
 ///       pgp_key_free (key);
+///       pgp_key_amalgamation_free (ka);
 ///     }
 ///     pgp_cert_key_iter_free (key_iter);
 ///     pgp_keyid_free (keyid);
-///     if (! key)
+///     if (! ka)
 ///       continue;
 ///
 ///     uint8_t algo;
@@ -883,17 +987,24 @@ impl DecryptionHelper for DHelper {
 ///     if (pgp_pkesk_decrypt (&err,
 ///                            pkesk, key, &algo,
 ///                            session_key, &session_key_len)) {
-///       error (1, 0, "pgp_pkesk_decrypt: %s", pgp_error_to_string (err));
+///       error (0, 0, "pgp_pkesk_decrypt: %s", pgp_error_to_string (err));
+///       pgp_key_free (key);
+///       pgp_key_amalgamation_free (ka);
+///       continue;
 ///     }
 ///     pgp_key_free (key);
+///     pgp_key_amalgamation_free (ka);
 ///
 ///     pgp_session_key_t sk = pgp_session_key_from_bytes (session_key,
 ///                                                        session_key_len);
-///     rc = decrypt (decrypt_cookie, algo, sk);
-///     pgp_session_key_free (sk);
+///     if (! decrypt (decrypt_cookie, algo, sk)) {
+///       pgp_session_key_free (sk);
+///       continue;
+///     }
 ///
+///     pgp_session_key_free (sk);
 ///     *identity_out = pgp_cert_fingerprint (cookie->key);
-///     return rc;
+///     return PGP_STATUS_SUCCESS;
 ///   }
 ///
 ///   return PGP_STATUS_UNKNOWN_ERROR;
@@ -922,7 +1033,7 @@ impl DecryptionHelper for DHelper {
 ///     .decrypt_called = 0,
 ///   };
 ///   plaintext = pgp_decryptor_new (NULL, policy, source,
-///                                  get_public_keys_cb, decrypt_cb,
+///                                  get_certs_cb, decrypt_cb,
 ///                                  check_cb, NULL, &cookie, 1554542219);
 ///   assert (plaintext);
 ///
@@ -942,7 +1053,7 @@ impl DecryptionHelper for DHelper {
 fn pgp_decryptor_new<'a>(errp: Option<&mut *mut crate::error::Error>,
                          policy: *const Policy,
                          input: *mut io::Reader,
-                         get_public_keys: GetPublicKeysCallback,
+                         get_certs: GetPublicKeysCallback,
                          decrypt: DecryptCallback,
                          check: CheckCallback,
                          inspect: Option<InspectCallback>,
@@ -950,11 +1061,13 @@ fn pgp_decryptor_new<'a>(errp: Option<&mut *mut crate::error::Error>,
                          time: time_t)
                          -> Maybe<io::Reader>
 {
+    ffi_make_fry_from_errp!(errp);
     let policy = policy.ref_raw().as_ref();
     let helper = DHelper::new(
-        get_public_keys, decrypt, check, inspect, cookie);
+        get_certs, decrypt, check, inspect, cookie);
 
-    Decryptor::from_reader(policy, input.ref_mut_raw(), helper, maybe_time(time))
+    ffi_try_or!(DecryptorBuilder::from_reader(input.ref_mut_raw()), None)
+        .with_policy(policy, maybe_time(time), helper)
         .map(|r| io::ReaderKind::Generic(Box::new(r)))
         .move_into_raw(errp)
 }
