@@ -24,7 +24,13 @@ pub struct Policy(Box<dyn policy::Policy>);
 #[crate::ffi_wrapper_type(
     prefix = "pgp_",
     derive = "Clone, Debug")]
-pub struct StandardPolicy(policy::StandardPolicy);
+pub struct StandardPolicy<'a>(policy::StandardPolicy<'a>);
+
+/// A NullPolicy object.
+#[crate::ffi_wrapper_type(
+    prefix = "pgp_",
+    derive = "Clone, Debug")]
+pub struct NullPolicy(policy::NullPolicy);
 
 /// Returns a new standard policy.
 #[::sequoia_ffi_macros::extern_fn] #[no_mangle]
@@ -34,3 +40,12 @@ pub extern "C" fn pgp_standard_policy()
     let p : Box<dyn policy::Policy> = Box::new(policy::StandardPolicy::new());
     p.move_into_raw()
 }
+
+/// Returns a new null policy.
+#[::sequoia_ffi_macros::extern_fn] #[no_mangle]
+pub extern "C" fn pgp_null_policy()
+    -> *mut Policy
+{
+    let p : Box<dyn policy::Policy> = Box::new(policy::NullPolicy::new());
+    p.move_into_raw()
+}
