@@ -1,8 +1,9 @@
 //! ASCII Armor.
 //!
-//! This module deals with ASCII Armored data (see [RFC 4880, section 6]).
+//! This module deals with ASCII Armored data (see [Section 6 of RFC
+//! 4880]).
 //!
-//! [RFC 4880, section 6]: https://tools.ietf.org/html/rfc4880#section-6
+//!   [Section 6 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-6
 //!
 //! # Scope
 //!
@@ -33,12 +34,14 @@ use std::path::Path;
 use std::cmp;
 use std::str;
 use std::borrow::Cow;
+
+#[cfg(any(test, feature = "quickcheck"))]
 use quickcheck::{Arbitrary, Gen};
 
 use crate::vec_truncate;
 use crate::packet::prelude::*;
 use crate::packet::header::{BodyLength, CTBNew, CTBOld};
-use crate::serialize::SerializeInto;
+use crate::serialize::MarshalInto;
 
 /// The encoded output stream must be represented in lines of no more
 /// than 76 characters each (see (see [RFC 4880, section
@@ -67,6 +70,7 @@ pub enum Kind {
     File,
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for Kind {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         use self::Kind::*;
@@ -129,7 +133,7 @@ impl Kind {
     /// This does not include any trailing newline.  It is simply the
     /// length of:
     ///
-    /// ```norun
+    /// ```text
     /// -----BEGIN PGP BLUB -----
     /// ```
     fn header_len(&self) -> usize {
@@ -140,7 +144,7 @@ impl Kind {
 
 /// A filter that applies ASCII Armor to the data written to it.
 pub struct Writer<W: Write> {
-    sink: Option<W>,
+    sink: W,
     kind: Kind,
     stash: Vec<u8>,
     column: usize,
@@ -152,12 +156,33 @@ pub struct Writer<W: Write> {
 impl<W: Write> Writer<W> {
     /// Constructs a new filter for the given type of data.
     ///
-    /// Note: To ensure that we can handle errors during writing of
-    /// the armor footer, this object must be consumed by calling
-    /// [`Writer::finalize()`].  If the object is dropped without
-    /// being finalized, we will panic in debug builds.
+    /// # Example
     ///
-    ///   [`Writer::finalize()`]: #method.finalize
+    /// ```
+    /// use std::io::{Read, Write, Cursor};
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::armor::{Writer, Kind};
+    /// # fn main() { f().unwrap(); }
+    /// # fn f() -> std::io::Result<()> {
+    /// let mut writer = Writer::new(Vec::new(), Kind::File)?;
+    /// writer.write_all(b"Hello world!")?;
+    /// let buffer = writer.finalize()?;
+    /// assert_eq!(
+    ///     String::from_utf8_lossy(&buffer),
+    ///     "-----BEGIN PGP ARMORED FILE-----
+    ///
+    /// SGVsbG8gd29ybGQh
+    /// =s4Gu
+    /// -----END PGP ARMORED FILE-----
+    /// ");
+    /// # Ok(())
+    /// # }
+    /// ```
+    pub fn new(inner: W, kind: Kind) -> Result<Self> {
+        Self::with_headers(inner, kind, Option::<(&str, &str)>::None)
+    }
+
+    /// Constructs a new filter for the given type of data.
     ///
     /// # Example
     ///
@@ -167,8 +192,8 @@ impl<W: Write> Writer<W> {
     /// use openpgp::armor::{Writer, Kind};
     /// # fn main() { f().unwrap(); }
     /// # fn f() -> std::io::Result<()> {
-    /// let mut writer = Writer::new(Vec::new(), Kind::File,
-    ///     &[ ("Key", "Value") ][..])?;
+    /// let mut writer = Writer::with_headers(Vec::new(), Kind::File,
+    ///     vec![ ("Key", "Value") ])?;
     /// writer.write_all(b"Hello world!")?;
     /// let buffer = writer.finalize()?;
     /// assert_eq!(
@@ -183,10 +208,15 @@ impl<W: Write> Writer<W> {
     /// # Ok(())
     /// # }
     /// ```
-    pub fn new(inner: W, kind: Kind, headers: &[(&str, &str)]) -> Result<Self> {
+    pub fn with_headers<I, K, V>(inner: W, kind: Kind, headers: I)
+                                 -> Result<Self>
+        where I: IntoIterator<Item = (K, V)>,
+              K: AsRef<str>,
+              V: AsRef<str>,
+    {
         let mut w = Writer {
-            sink: Some(inner),
-            kind: kind,
+            sink: inner,
+            kind,
             stash: Vec::<u8>::with_capacity(2),
             column: 0,
             crc: CRC::new(),
@@ -199,7 +229,8 @@ impl<W: Write> Writer<W> {
             write!(&mut cur, "{}{}", kind.begin(), LINE_ENDING)?;
 
             for h in headers {
-                write!(&mut cur, "{}: {}{}", h.0, h.1, LINE_ENDING)?;
+                write!(&mut cur, "{}: {}{}", h.0.as_ref(), h.1.as_ref(),
+                       LINE_ENDING)?;
             }
 
             // A blank line separates the headers from the body.
@@ -209,15 +240,20 @@ impl<W: Write> Writer<W> {
         Ok(w)
     }
 
-    fn e_finalized() -> Error {
-        Error::new(ErrorKind::BrokenPipe, "Writer is finalized.")
+    /// Returns a reference to the inner writer.
+    pub fn get_ref(&self) -> &W {
+        &self.sink
+    }
+
+    /// Returns a mutable reference to the inner writer.
+    pub fn get_mut(&mut self) -> &mut W {
+        &mut self.sink
     }
 
     fn finalize_headers(&mut self) -> Result<()> {
         if ! self.dirty {
             self.dirty = true;
-            self.sink.as_mut().ok_or_else(Self::e_finalized)?
-                .write_all(&self.header)?;
+            self.sink.write_all(&self.header)?;
             // Release memory.
             crate::vec_truncate(&mut self.header, 0);
             self.header.shrink_to_fit();
@@ -233,20 +269,10 @@ impl<W: Write> Writer<W> {
     pub fn finalize(mut self) -> Result<W> {
         if ! self.dirty {
             // No data was written to us, don't emit anything.
-            return Ok(self.sink.take().ok_or_else(Self::e_finalized)?);
-        }
-
-        // We need to clear self.sink if this succeeds or not, so be
-        // careful with the result here.  Otherwise, self.sink may
-        // still be present when the object is really dropped,
-        // invoking finalize_armor again.
-        let r = self.finalize_armor();
-        let sink = self.sink.take();
-        if let Err(e) = r {
-            Err(e)
-        } else {
-            sink.ok_or(Self::e_finalized())
+            return Ok(self.sink);
         }
+        self.finalize_armor()?;
+        Ok(self.sink)
     }
 
     /// Writes the footer.
@@ -256,55 +282,48 @@ impl<W: Write> Writer<W> {
             return Ok(());
         }
         self.finalize_headers()?;
-        if let Some(sink) = self.sink.as_mut() {
-            // Write any stashed bytes and pad.
-            if self.stash.len() > 0 {
-                sink.write_all(base64::encode_config(
-                    &self.stash, base64::STANDARD).as_bytes())?;
-                self.column += 4;
-            }
 
-            // Inserts a line break if necessary.
-            //
-            // Unfortunately, we cannot use
-            //self.linebreak()?;
-            //
-            // Therefore, we inline it here.  This is a bit sad.
-            assert!(self.column <= LINE_LENGTH);
-            if self.column == LINE_LENGTH {
-                write!(sink, "{}", LINE_ENDING)?;
-                self.column = 0;
-            }
-
-            if self.column > 0 {
-                write!(sink, "{}", LINE_ENDING)?;
-            }
-
-            let crc = self.crc.finalize();
-            let bytes: [u8; 3] = [
-                (crc >> 16) as u8,
-                (crc >>  8) as u8,
-                (crc >>  0) as u8,
-            ];
-
-            // CRC and footer.
-            write!(sink, "={}{}{}{}",
-                   base64::encode_config(&bytes, base64::STANDARD_NO_PAD),
-                   LINE_ENDING, self.kind.end(), LINE_ENDING)?;
-
-            self.dirty = false;
-            Ok(())
-        } else {
-            Err(Self::e_finalized())
+        // Write any stashed bytes and pad.
+        if self.stash.len() > 0 {
+            self.sink.write_all(base64::encode_config(
+                &self.stash, base64::STANDARD).as_bytes())?;
+            self.column += 4;
         }
+
+        // Inserts a line break if necessary.
+        //
+        // Unfortunately, we cannot use
+        //self.linebreak()?;
+        //
+        // Therefore, we inline it here.  This is a bit sad.
+        assert!(self.column <= LINE_LENGTH);
+        if self.column == LINE_LENGTH {
+            write!(self.sink, "{}", LINE_ENDING)?;
+            self.column = 0;
+        }
+
+        if self.column > 0 {
+            write!(self.sink, "{}", LINE_ENDING)?;
+        }
+
+        // 24-bit CRC
+        let crc = self.crc.finalize();
+        let bytes = &crc.to_be_bytes()[1..4];
+
+        // CRC and footer.
+        write!(self.sink, "={}{}{}{}",
+               base64::encode_config(&bytes, base64::STANDARD_NO_PAD),
+               LINE_ENDING, self.kind.end(), LINE_ENDING)?;
+
+        self.dirty = false;
+        Ok(())
     }
 
     /// Inserts a line break if necessary.
     fn linebreak(&mut self) -> Result<()> {
         assert!(self.column <= LINE_LENGTH);
         if self.column == LINE_LENGTH {
-            write!(self.sink.as_mut().ok_or_else(Self::e_finalized)?,
-                   "{}", LINE_ENDING)?;
+            write!(self.sink, "{}", LINE_ENDING)?;
             self.column = 0;
         }
         Ok(())
@@ -341,7 +360,7 @@ impl<W: Write> Write for Writer<W> {
 
             // If this fails for some reason, and the caller retries
             // the write, we might end up with a stash of size 3.
-            self.sink.as_mut().ok_or_else(Self::e_finalized)?
+            self.sink
                 .write_all(base64::encode_config(
                     &self.stash, base64::STANDARD_NO_PAD).as_bytes())?;
             self.column += 4;
@@ -367,7 +386,7 @@ impl<W: Write> Write for Writer<W> {
         let mut enc = encoded.as_bytes();
         while enc.len() > 0 {
             let n = cmp::min(LINE_LENGTH - self.column, enc.len());
-            self.sink.as_mut().ok_or_else(Self::e_finalized)?
+            self.sink
                 .write_all(&enc[..n])?;
             enc = &enc[n..];
             self.column += n;
@@ -379,16 +398,7 @@ impl<W: Write> Write for Writer<W> {
     }
 
     fn flush(&mut self) -> Result<()> {
-        self.sink.as_mut().ok_or_else(Self::e_finalized)?.flush()
-    }
-}
-
-impl<W: Write> Drop for Writer<W> {
-    fn drop(&mut self) {
-        debug_assert!(self.sink.is_none(),
-                      "armor writer dropped without being finalized, \
-                       use armor::Writer::finalize()");
-        let _ = self.finalize_armor();
+        self.sink.flush()
     }
 }
 
@@ -562,7 +572,7 @@ impl<'a> Reader<'a> {
         Reader {
             source: Box::new(buffered_reader::Generic::new(inner, None)),
             kind: None,
-            mode: mode,
+            mode,
             buffer: Vec::<u8>::with_capacity(1024),
             crc: CRC::new(),
             expect_crc: None,
@@ -759,7 +769,7 @@ impl<'a> Reader<'a> {
                 // Nope, we have actually failed to read this properly
                 return Err(
                     Error::new(ErrorKind::InvalidInput,
-                               "Reached EOF looking for Armor Header Line"));
+                               "Inconsistent quoting of armored data"));
             }
         }
 
@@ -767,11 +777,17 @@ impl<'a> Reader<'a> {
         let mut n = 0;
         let mut lines = 0;
         loop {
-            // Skip any known prefix on lines
+            // Skip any known prefix on lines.
+            //
+            // IMPORTANT: We need to buffer the prefix so that we can
+            // consume it here.  So at every point in this loop where
+            // the control flow wraps around, we need to make sure
+            // that we buffer the prefix in addition to the line.
             self.source.consume(prefix.len());
 
             self.source.consume(n);
 
+            // Buffer the next line.
             let line = self.source.read_to('\n' as u8)?;
             n = line.len();
             lines += 1;
@@ -779,6 +795,15 @@ impl<'a> Reader<'a> {
             let line = str::from_utf8(line);
             // Ignore---don't error out---lines that are not valid UTF8.
             if line.is_err() {
+                // Buffer the next line and the prefix that is going
+                // to be consumed in the next iteration.
+                let next_prefix =
+                    &self.source.data_hard(n + prefix.len())?[n..n + prefix.len()];
+                if prefix != next_prefix {
+                    return Err(
+                        Error::new(ErrorKind::InvalidInput,
+                                   "Inconsistent quoting of armored data"));
+                }
                 continue;
             }
 
@@ -791,7 +816,7 @@ impl<'a> Reader<'a> {
             let line = if line.ends_with(&"\r\n"[..]) {
                 // \r\n.
                 &line[..line.len() - 2]
-            } else if line.len() > 0 {
+            } else if line.ends_with("\n") {
                 // \n.
                 &line[..line.len() - 1]
             } else {
@@ -819,6 +844,16 @@ impl<'a> Reader<'a> {
 
                 self.headers.push((key.into(), value.into()));
             }
+
+            // Buffer the next line and the prefix that is going to be
+            // consumed in the next iteration.
+            let next_prefix =
+                &self.source.data_hard(n + prefix.len())?[n..n + prefix.len()];
+            if prefix != next_prefix {
+                return Err(
+                    Error::new(ErrorKind::InvalidInput,
+                               "Inconsistent quoting of armored data"));
+            }
         }
         self.source.consume(n);
 
@@ -1191,7 +1226,7 @@ impl<'a> Read for Reader<'a> {
             self.source.consume(consumed);
 
             // Skip any expected prefix
-            self.source.consume(self.prefix_len);
+            self.source.data_consume_hard(self.prefix_len)?;
             // Look for a footer.
             let consumed = {
                 // Skip whitespace.
@@ -1234,8 +1269,6 @@ impl<'a> Read for Reader<'a> {
     }
 }
 
-// XXX: impl BufferedReader for Reader
-
 const CRC24_INIT: u32 = 0xB704CE;
 const CRC24_POLY: u32 = 0x1864CFB;
 
@@ -1331,7 +1364,7 @@ mod test {
     fn enarmor() {
         for (bin, asc) in TEST_BIN.iter().zip(TEST_ASC.iter()) {
             let mut w =
-                Writer::new(Vec::new(), Kind::File, &[]).unwrap();
+                Writer::new(Vec::new(), Kind::File).unwrap();
             w.write(&[]).unwrap();  // Avoid zero-length optimization.
             w.write_all(bin).unwrap();
             let buf = w.finalize().unwrap();
@@ -1343,7 +1376,7 @@ mod test {
     #[test]
     fn enarmor_bytewise() {
         for (bin, asc) in TEST_BIN.iter().zip(TEST_ASC.iter()) {
-            let mut w = Writer::new(Vec::new(), Kind::File, &[]).unwrap();
+            let mut w = Writer::new(Vec::new(), Kind::File).unwrap();
             w.write(&[]).unwrap();  // Avoid zero-length optimization.
             for b in bin.iter() {
                 w.write(&[*b]).unwrap();
@@ -1358,12 +1391,12 @@ mod test {
     fn drop_writer() {
         // No ASCII frame shall be emitted if the writer is dropped
         // unused.
-        assert!(Writer::new(Vec::new(), Kind::File, &[]).unwrap()
+        assert!(Writer::new(Vec::new(), Kind::File).unwrap()
                 .finalize().unwrap().is_empty());
 
         // However, if the user insists, we will encode a zero-byte
         // string.
-        let mut w = Writer::new(Vec::new(), Kind::File, &[]).unwrap();
+        let mut w = Writer::new(Vec::new(), Kind::File).unwrap();
         w.write(&[]).unwrap();
         let buf = w.finalize().unwrap();
         assert_eq!(
@@ -1604,7 +1637,7 @@ mod test {
                 return true;
             }
 
-            let mut w = Writer::new(Vec::new(), kind, &[]).unwrap();
+            let mut w = Writer::new(Vec::new(), kind).unwrap();
             w.write_all(&payload).unwrap();
             let encoded = w.finalize().unwrap();
 
@@ -1632,4 +1665,61 @@ mod test {
         r.read(&mut buf).unwrap();
         r.read(&mut buf).unwrap();
     }
+
+    /// Crash in armor parser due to indexing not aligned with UTF-8
+    /// characters.
+    #[test]
+    fn issue_515() {
+        let data = [63, 9, 45, 10, 45, 10, 45, 45, 45, 45, 45, 66, 69,
+                    71, 73, 78, 32, 80, 71, 80, 32, 77, 69, 83, 83,
+                    65, 71, 69, 45, 45, 45, 45, 45, 45, 152, 152, 152,
+                    152, 152, 152, 255, 29, 152, 152, 152, 152, 152,
+                    152, 152, 152, 152, 152, 10, 91, 45, 10, 45, 14,
+                    0, 36, 0, 0, 30, 122, 4, 2, 204, 152];
+
+        let mut reader = Reader::from_bytes(&data[..], None);
+        let mut buf = Vec::new();
+        // `data` is malformed, expect an error.
+        reader.read_to_end(&mut buf).unwrap_err();
+    }
+
+    /// Crash in armor parser due to improper use of the buffered
+    /// reader protocol when consuming quoting prefix.
+    #[test]
+    fn issue_516() {
+        let data = [
+            144, 32, 19, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 125, 13, 125,
+            125, 93, 125, 125, 93, 125, 13, 13, 125, 125, 45, 45, 45,
+            45, 45, 66, 69, 71, 73, 78, 32, 80, 71, 80, 32, 77, 69,
+            83, 83, 65, 71, 69, 45, 45, 45, 45, 45, 125, 13, 125,
+            125, 93, 125, 125, 93, 125, 13, 13, 125, 125, 45, 0, 0,
+            0, 0, 0, 0, 0, 0, 125, 205, 21, 1, 21, 21, 21, 1, 1, 1,
+            1, 21, 149, 21, 21, 21, 21, 32, 4, 141, 141, 141, 141,
+            202, 74, 11, 125, 8, 21, 50, 50, 194, 48, 147, 93, 174,
+            23, 23, 23, 23, 23, 23, 147, 147, 147, 23, 23, 23, 23,
+            23, 23, 48, 125, 125, 93, 125, 13, 125, 125, 125, 93,
+            125, 125, 13, 13, 125, 125, 13, 13, 93, 125, 13, 125, 45,
+            125, 125, 45, 45, 66, 69, 71, 73, 78, 32, 80, 71, 45, 45,
+            125, 10, 45, 45, 0, 0, 10, 45, 45, 210, 10, 0, 0, 87, 0,
+            0, 0, 150, 10, 0, 0, 241, 87, 45, 0, 0, 121, 121, 10, 10,
+            21, 58];
+        let mut reader = Reader::from_bytes(&data[..], None);
+        let mut buf = Vec::new();
+        // `data` is malformed, expect an error.
+        reader.read_to_end(&mut buf).unwrap_err();
+    }
+
+    /// Crash in armor parser due to improper use of the buffered
+    /// reader protocol when consuming quoting prefix.
+    #[test]
+    fn issue_517() {
+        let data = [13, 45, 45, 45, 45, 45, 66, 69, 71, 73, 78, 32, 80,
+                    71, 80, 32, 77, 69, 83, 83, 65, 71, 69, 45, 45, 45,
+                    45, 45, 10, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
+                    13, 13, 139];
+        let mut reader = Reader::from_bytes(&data[..], None);
+        let mut buf = Vec::new();
+        // `data` is malformed, expect an error.
+        reader.read_to_end(&mut buf).unwrap_err();
+    }
 }
