@@ -14,7 +14,9 @@ use crate::crypto::SessionKey;
 
 use std::fmt;
 
+#[cfg(any(test, feature = "quickcheck"))]
 use quickcheck::{Arbitrary, Gen};
+#[cfg(any(test, feature = "quickcheck"))]
 use rand::Rng;
 
 /// String-to-Key (S2K) specifiers.
@@ -64,7 +66,7 @@ impl Default for S2K {
             // SHA2-512.  Furthermore, the digest size is large enough
             // for every cipher algorithm currently in use.
             hash: HashAlgorithm::SHA256,
-            salt: salt,
+            salt,
             // This is the largest count that OpenPGP can represent.
             // On moderate machines, like my Intel(R) Core(TM) i5-2400
             // CPU @ 3.10GHz, it takes ~354ms to derive a key.
@@ -262,6 +264,7 @@ impl fmt::Display for S2K {
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for S2K {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         match g.gen_range(0, 5) {
@@ -290,7 +293,6 @@ mod tests {
     use crate::SymmetricAlgorithm;
     use crate::Packet;
     use crate::parse::{Parse, PacketParser};
-    use crate::serialize::Serialize;
 
     #[test]
     fn s2k_parser_test() {
@@ -423,7 +425,8 @@ mod tests {
 
     quickcheck! {
         fn s2k_roundtrip(s2k: S2K) -> bool {
-            use crate::serialize::SerializeInto;
+            use crate::serialize::Marshal;
+            use crate::serialize::MarshalInto;
 
             eprintln!("in {:?}", s2k);
             use std::io::Cursor;
