@@ -1,3 +1,4 @@
+#[cfg(any(test, feature = "quickcheck"))]
 use quickcheck::{Arbitrary, Gen};
 
 use crate::packet;
@@ -8,25 +9,14 @@ use crate::Packet;
 /// See [Section 5.8 of RFC 4880] for details.
 ///
 ///   [Section 5.8 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-5.8
-#[derive(Clone, Debug)]
+// IMPORTANT: If you add fields to this struct, you need to explicitly
+// IMPORTANT: implement PartialEq, Eq, and Hash.
+#[derive(Clone, Debug, PartialEq, Eq, Hash)]
 pub struct Marker {
     /// CTB packet header fields.
     pub(crate) common: packet::Common,
 }
 
-impl PartialEq for Marker {
-    fn eq(&self, _other: &Marker) -> bool {
-        true
-    }
-}
-
-impl Eq for Marker {}
-
-impl std::hash::Hash for Marker {
-    fn hash<H: std::hash::Hasher>(&self, _state: &mut H) {
-    }
-}
-
 impl Marker {
     pub(crate) const BODY: &'static [u8] = &[0x50, 0x47, 0x50];
 }
@@ -45,6 +35,7 @@ impl From<Marker> for Packet {
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for Marker {
     fn arbitrary<G: Gen>(_: &mut G) -> Self {
         Self::default()
@@ -55,7 +46,7 @@ impl Arbitrary for Marker {
 mod tests {
     use super::*;
     use crate::parse::Parse;
-    use crate::serialize::SerializeInto;
+    use crate::serialize::MarshalInto;
 
     #[test]
     fn roundtrip() {
