@@ -3,10 +3,13 @@
 use std::fmt;
 use std::ops::{Deref, DerefMut};
 
+#[cfg(any(test, feature = "quickcheck"))]
+use quickcheck::{Arbitrary, Gen};
+
 use crate::Error;
 use crate::Result;
 use crate::crypto::{
-    mpis,
+    mpi,
     hash::{self, Hash},
     Signer,
 };
@@ -27,6 +30,31 @@ use crate::packet::signature::subpacket::{
     SubpacketAreas,
 };
 
+#[cfg(any(test, feature = "quickcheck"))]
+/// Like quickcheck::Arbitrary, but bounded.
+trait ArbitraryBounded {
+    /// Generates an arbitrary value, but only recurses if `depth >
+    /// 0`.
+    fn arbitrary_bounded<G: Gen>(g: &mut G, depth: usize) -> Self;
+}
+
+#[cfg(any(test, feature = "quickcheck"))]
+/// Default depth when implementing Arbitrary using ArbitraryBounded.
+const DEFAULT_ARBITRARY_DEPTH: usize = 2;
+
+#[cfg(any(test, feature = "quickcheck"))]
+macro_rules! impl_arbitrary_with_bound {
+    ($typ:path) => {
+        impl Arbitrary for $typ {
+            fn arbitrary<G: Gen>(g: &mut G) -> Self {
+                Self::arbitrary_bounded(
+                    g,
+                    crate::packet::signature::DEFAULT_ARBITRARY_DEPTH)
+            }
+        }
+    }
+}
+
 pub mod subpacket;
 
 /// Builds a signature packet.
@@ -48,14 +76,24 @@ pub mod subpacket;
 ///   [`sign_standalone`]: #method.sign_standalone
 ///   [`sign_timestamp`]: #method.sign_timestamp
 ///
+/// By default, these functions add references to the signing key by adding
+/// Issuer and Issuer Fingerprint subpackets in the unhashed subpacket area.
+/// To override, use [`set_issuer`] and [`set_issuer_fingerprint`].
+/// Caution: this likely makes the signature unverifiable.
+///
+///   [`set_issuer`]: #method.set_issuer
+///   [`set_issuer_fingerprint`]: #method.set_issuer_fingerprint
+///
 /// Signatures must always include a creation time.  We automatically
 /// insert a creation time subpacket with the current time into the
 /// hashed subpacket area.  To override the creation time, use
 /// [`set_signature_creation_time`].
 ///
 ///   [`set_signature_creation_time`]: #method.set_signature_creation_time
+// IMPORTANT: If you add fields to this struct, you need to explicitly
+// IMPORTANT: implement PartialEq, Eq, and Hash.
 #[derive(Clone, Hash, PartialEq, Eq)]
-pub struct Builder {
+pub struct SignatureBuilder {
     /// Version of the signature packet. Must be 4.
     version: u8,
     /// Type of signature.
@@ -68,7 +106,25 @@ pub struct Builder {
     subpackets: SubpacketAreas,
 }
 
-impl Deref for Builder {
+#[cfg(any(test, feature = "quickcheck"))]
+impl ArbitraryBounded for SignatureBuilder {
+    fn arbitrary_bounded<G: Gen>(g: &mut G, depth: usize) -> Self {
+        SignatureBuilder {
+            // XXX: Make this more interesting once we dig other
+            // versions.
+            version: 4,
+            typ: Arbitrary::arbitrary(g),
+            pk_algo: PublicKeyAlgorithm::arbitrary_for_signing(g),
+            hash_algo: Arbitrary::arbitrary(g),
+            subpackets: ArbitraryBounded::arbitrary_bounded(g, depth),
+        }
+    }
+}
+
+#[cfg(any(test, feature = "quickcheck"))]
+impl_arbitrary_with_bound!(SignatureBuilder);
+
+impl Deref for SignatureBuilder {
     type Target = SubpacketAreas;
 
     fn deref(&self) -> &Self::Target {
@@ -76,18 +132,18 @@ impl Deref for Builder {
     }
 }
 
-impl DerefMut for Builder {
+impl DerefMut for SignatureBuilder {
     fn deref_mut(&mut self) -> &mut Self::Target {
         &mut self.subpackets
     }
 }
 
-impl Builder {
-    /// Returns a new `Builder` object.
+impl SignatureBuilder {
+    /// Returns a new `SignatureBuilder` object.
     pub fn new(typ: SignatureType) ->  Self {
-        Builder {
+        SignatureBuilder {
             version: 4,
-            typ: typ,
+            typ,
             pk_algo: PublicKeyAlgorithm::Unknown(0),
             hash_algo: HashAlgorithm::default(),
             subpackets: SubpacketAreas::default(),
@@ -134,9 +190,12 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_standalone(mut self, signer: &mut dyn Signer)
                            -> Result<Signature>
     {
+        self.sort();
         self.pk_algo = signer.public().pk_algo();
         let digest = Signature::hash_standalone(&self)?;
         self.sign(signer, digest)
@@ -146,9 +205,12 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_timestamp(mut self, signer: &mut dyn Signer)
                           -> Result<Signature>
     {
+        self.sort();
         self.pk_algo = signer.public().pk_algo();
         let digest = Signature::hash_timestamp(&self)?;
         self.sign(signer, digest)
@@ -158,14 +220,17 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_direct_key(mut self, signer: &mut dyn Signer)
         -> Result<Signature>
     {
+        self.sort();
         self.pk_algo = signer.public().pk_algo();
         let digest =
             Signature::hash_direct_key(&self,
                                        signer.public()
-                                       .mark_role_primary_ref())?;
+                                       .role_as_primary())?;
 
         self.sign(signer, digest)
     }
@@ -174,12 +239,15 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_userid_binding<P>(mut self, signer: &mut dyn Signer,
                                   key: &Key<P, key::PrimaryRole>,
                                   userid: &UserID)
         -> Result<Signature>
         where P: key::KeyParts,
     {
+        self.sort();
         self.pk_algo = signer.public().pk_algo();
         let digest = Signature::hash_userid_binding(&self, key, userid)?;
 
@@ -190,6 +258,8 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_subkey_binding<P, Q>(mut self, signer: &mut dyn Signer,
                                      primary: &Key<P, key::PrimaryRole>,
                                      subkey: &Key<Q, key::SubordinateRole>)
@@ -197,6 +267,7 @@ impl Builder {
         where P: key:: KeyParts,
               Q: key:: KeyParts,
     {
+        self.sort();
         self.pk_algo = signer.public().pk_algo();
         let digest = Signature::hash_subkey_binding(&self, primary, subkey)?;
 
@@ -208,6 +279,8 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `subkey_signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `subkey_signer`.
     pub fn sign_primary_key_binding<P, Q>(mut self,
                                           subkey_signer: &mut dyn Signer,
                                           primary: &Key<P, key::PrimaryRole>,
@@ -216,6 +289,7 @@ impl Builder {
         where P: key:: KeyParts,
               Q: key:: KeyParts,
     {
+        self.sort();
         self.pk_algo = subkey_signer.public().pk_algo();
         let digest =
             Signature::hash_primary_key_binding(&self, primary, subkey)?;
@@ -227,12 +301,15 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_user_attribute_binding<P>(mut self, signer: &mut dyn Signer,
                                           key: &Key<P, key::PrimaryRole>,
                                           ua: &UserAttribute)
         -> Result<Signature>
         where P: key::KeyParts,
     {
+        self.sort();
         self.pk_algo = signer.public().pk_algo();
         let digest =
             Signature::hash_user_attribute_binding(&self, key, ua)?;
@@ -244,10 +321,13 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_hash(mut self, signer: &mut dyn Signer,
                      mut hash: hash::Context)
         -> Result<Signature>
     {
+        self.sort();
         // Fill out some fields, then hash the packet.
         self.pk_algo = signer.public().pk_algo();
         self.hash_algo = hash.algo();
@@ -264,10 +344,13 @@ impl Builder {
     ///
     /// The Signature's public-key algorithm field is set to the
     /// algorithm used by `signer`.
+    /// If not set before, Issuer and Issuer Fingerprint subpackets are added
+    /// pointing to `signer`.
     pub fn sign_message<M>(mut self, signer: &mut dyn Signer, msg: M)
         -> Result<Signature>
         where M: AsRef<[u8]>
     {
+        self.sort();
         // Hash the message
         let mut hash = self.hash_algo.context()?;
         hash.update(msg.as_ref());
@@ -289,18 +372,25 @@ impl Builder {
         let algo = self.hash_algo;
         let mpis = signer.sign(algo, &digest)?;
 
+        let mut fields = self;
+
+        if fields.issuer().is_none() && fields.issuer_fingerprint().is_none() {
+            fields = fields.set_issuer(signer.public().keyid())?
+                .set_issuer_fingerprint(signer.public().fingerprint())?;
+        }
+
         Ok(Signature4 {
             common: Default::default(),
-            fields: self,
+            fields,
             digest_prefix: [digest[0], digest[1]],
-            mpis: mpis,
+            mpis,
             computed_digest: Some(digest),
             level: 0,
         }.into())
     }
 }
 
-impl From<Signature> for Builder {
+impl From<Signature> for SignatureBuilder {
     fn from(sig: Signature) -> Self {
         match sig {
             Signature::V4(sig) => sig.into(),
@@ -309,13 +399,13 @@ impl From<Signature> for Builder {
     }
 }
 
-impl From<Signature4> for Builder {
+impl From<Signature4> for SignatureBuilder {
     fn from(sig: Signature4) -> Self {
         sig.fields
     }
 }
 
-impl<'a> From<&'a Signature> for &'a Builder {
+impl<'a> From<&'a Signature> for &'a SignatureBuilder {
     fn from(sig: &'a Signature) -> Self {
         match sig {
             Signature::V4(ref sig) => sig.into(),
@@ -324,7 +414,7 @@ impl<'a> From<&'a Signature> for &'a Builder {
     }
 }
 
-impl<'a> From<&'a Signature4> for &'a Builder {
+impl<'a> From<&'a Signature4> for &'a SignatureBuilder {
     fn from(sig: &'a Signature4) -> Self {
         &sig.fields
     }
@@ -345,13 +435,13 @@ pub struct Signature4 {
     /// CTB packet header fields.
     pub(crate) common: packet::Common,
 
-    /// Fields as configured using the builder.
-    pub(crate) fields: Builder,
+    /// Fields as configured using the SignatureBuilder.
+    pub(crate) fields: SignatureBuilder,
 
     /// Lower 16 bits of the signed hash value.
     digest_prefix: [u8; 2],
     /// Signature MPIs.
-    mpis: mpis::Signature,
+    mpis: mpi::Signature,
 
     /// When used in conjunction with a one-pass signature, this is the
     /// hash computed over the enclosed message.
@@ -413,11 +503,8 @@ impl PartialEq for Signature4 {
     /// signatures using this predicate.
     fn eq(&self, other: &Signature4) -> bool {
         self.mpis == other.mpis
-            && self.fields.version == other.fields.version
-            && self.fields.typ == other.fields.typ
-            && self.fields.pk_algo == other.fields.pk_algo
-            && self.fields.hash_algo == other.fields.hash_algo
-            && self.fields.hashed_area() == other.fields.hashed_area()
+            && self.fields == other.fields
+            && self.digest_prefix == other.digest_prefix
     }
 }
 
@@ -426,38 +513,35 @@ impl Eq for Signature4 {}
 impl std::hash::Hash for Signature4 {
     fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
         use std::hash::Hash as StdHash;
-        self.fields.version.hash(state);
-        self.fields.typ.hash(state);
-        self.fields.pk_algo.hash(state);
-        self.fields.hash_algo.hash(state);
-        self.fields.hashed_area().hash(state);
         StdHash::hash(&self.mpis, state);
+        StdHash::hash(&self.fields, state);
+        self.digest_prefix.hash(state);
     }
 }
 
 impl Signature4 {
     /// Creates a new signature packet.
     ///
-    /// If you want to sign something, consider using the [`Builder`]
+    /// If you want to sign something, consider using the [`SignatureBuilder`]
     /// interface.
     ///
-    /// [`Builder`]: struct.Builder.html
+    /// [`SignatureBuilder`]: struct.SignatureBuilder.html
     pub fn new(typ: SignatureType, pk_algo: PublicKeyAlgorithm,
                hash_algo: HashAlgorithm, hashed_area: SubpacketArea,
                unhashed_area: SubpacketArea,
                digest_prefix: [u8; 2],
-               mpis: mpis::Signature) -> Self {
+               mpis: mpi::Signature) -> Self {
         Signature4 {
             common: Default::default(),
-            fields: Builder {
+            fields: SignatureBuilder {
                 version: 4,
-                typ: typ,
-                pk_algo: pk_algo,
-                hash_algo: hash_algo,
+                typ,
+                pk_algo,
+                hash_algo,
                 subpackets: SubpacketAreas::new(hashed_area, unhashed_area),
             },
-            digest_prefix: digest_prefix,
-            mpis: mpis,
+            digest_prefix,
+            mpis,
             computed_digest: None,
             level: 0,
         }
@@ -475,13 +559,13 @@ impl Signature4 {
     }
 
     /// Gets the signature packet's MPIs.
-    pub fn mpis(&self) -> &mpis::Signature {
+    pub fn mpis(&self) -> &mpi::Signature {
         &self.mpis
     }
 
     /// Sets the signature packet's MPIs.
     #[allow(dead_code)]
-    pub(crate) fn set_mpis(&mut self, mpis: mpis::Signature) -> mpis::Signature
+    pub(crate) fn set_mpis(&mut self, mpis: mpi::Signature) -> mpi::Signature
     {
         ::std::mem::replace(&mut self.mpis, mpis)
     }
@@ -515,6 +599,22 @@ impl Signature4 {
     pub(crate) fn set_level(&mut self, level: usize) -> usize {
         ::std::mem::replace(&mut self.level, level)
     }
+
+    /// Tests whether or not this signature is exportable.
+    pub fn exportable(&self) -> Result<()> {
+        if ! self.exportable_certification().unwrap_or(true) {
+            return Err(Error::InvalidOperation(
+                "Cannot export non-exportable certification".into()).into());
+        }
+
+        if self.revocation_keys().any(|r| r.sensitive()) {
+            return Err(Error::InvalidOperation(
+                "Cannot export signature with sensitive designated revoker"
+                    .into()).into());
+        }
+
+        Ok(())
+    }
 }
 
 impl crate::packet::Signature {
@@ -525,7 +625,8 @@ impl crate::packet::Signature {
     pub fn get_issuers(&self) -> Vec<crate::KeyHandle> {
         use crate::packet::signature::subpacket:: SubpacketValue;
 
-        self.hashed_area().iter()
+        let mut issuers: Vec<_> =
+            self.hashed_area().iter()
             .chain(self.unhashed_area().iter())
             .filter_map(|subpacket| {
                 match subpacket.value() {
@@ -534,52 +635,65 @@ impl crate::packet::Signature {
                     _ => None,
                 }
             })
-            .collect()
+            .collect();
+
+        // Sort the issuers so that the fingerprints come first.
+        issuers.sort_by(|a, b| {
+            use crate::KeyHandle::*;
+            use std::cmp::Ordering::*;
+            match (a, b) {
+                (Fingerprint(_), Fingerprint(_)) => Equal,
+                (KeyID(_), Fingerprint(_)) => Greater,
+                (Fingerprint(_), KeyID(_)) => Less,
+                (KeyID(_), KeyID(_)) => Equal,
+            }
+        });
+        issuers
+    }
+
+    /// Compares Signatures ignoring the unhashed subpacket area.
+    ///
+    /// We ignore the unhashed subpacket area when comparing
+    /// signatures.  This prevents a malicious party to take valid
+    /// signatures, add subpackets to the unhashed area, yielding
+    /// valid but distinct signatures.
+    ///
+    /// The problem we are trying to avoid here is signature spamming.
+    /// Ignoring the unhashed subpackets means that we can deduplicate
+    /// signatures using this predicate.
+    pub fn normalized_eq(&self, other: &Signature) -> bool {
+        self.mpis() == other.mpis()
+            && self.version() == other.version()
+            && self.typ() == other.typ()
+            && self.pk_algo() == other.pk_algo()
+            && self.hash_algo() == other.hash_algo()
+            && self.hashed_area() == other.hashed_area()
+            && self.digest_prefix() == other.digest_prefix()
     }
 
     /// Normalizes the signature.
     ///
     /// This function normalizes the *unhashed* signature subpackets.
-    /// All but the following subpackets are removed:
+    /// It removes all but the following self-authenticating
+    /// subpackets:
     ///
-    ///   - `SubpacketValue::Issuer` is left in place, is added, or
-    ///     updated from the *hashed* signature subpackets, and
-    ///   - the first `SubpacketValue::EmbeddedSignature` is left in
-    ///     place.
+    ///   - `SubpacketValue::Issuer`
+    ///   - `SubpacketValue::IssuerFingerprint`
+    ///   - `SubpacketValue::EmbeddedSignature`
     pub fn normalize(&self) -> Self {
-        use crate::packet::signature::subpacket::{Subpacket, SubpacketTag,
-                                           SubpacketValue};
+        use crate::packet::signature::subpacket::SubpacketTag::*;
         let mut sig = self.clone();
         {
             let area = sig.unhashed_area_mut();
             area.clear();
 
-            // First, add an Issuer subpacket derived from information
-            // from the hashed area.
-            if let Some(issuer) = self.issuer_fingerprint() {
-                // Prefer the IssuerFingerprint.
-                area.add(Subpacket::new(
-                    SubpacketValue::Issuer(issuer.into()), false).unwrap())
-                    .unwrap();
-            } else if let Some(issuer) = self.issuer() {
-                // Fall back to the Issuer, which we will also get
-                // from the unhashed area if necessary.
-                area.add(Subpacket::new(
-                    SubpacketValue::Issuer(issuer.clone()), false).unwrap())
-                    .unwrap();
-            }
-
-            // Second, re-add the EmbeddedSignature, if present.
-            if let Some(embedded_sig) =
-                self.unhashed_area().iter().find_map(|v| {
-                    if v.tag() == SubpacketTag::EmbeddedSignature {
-                        Some(v)
-                    } else {
-                        None
-                    }
-                })
+            for spkt in self.unhashed_area().iter()
+                .filter(|s| s.tag() == Issuer
+                        || s.tag() == IssuerFingerprint
+                        || s.tag() == EmbeddedSignature)
             {
-                area.add(embedded_sig.clone()).unwrap();
+                area.add(spkt.clone())
+                    .expect("it did fit into the old area");
             }
         }
         sig
@@ -1105,14 +1219,69 @@ impl From<Signature4> for super::Signature {
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
+impl ArbitraryBounded for super::Signature {
+    fn arbitrary_bounded<G: Gen>(g: &mut G, depth: usize) -> Self {
+        Signature4::arbitrary_bounded(g, depth).into()
+    }
+}
+
+#[cfg(any(test, feature = "quickcheck"))]
+impl_arbitrary_with_bound!(super::Signature);
+
+#[cfg(any(test, feature = "quickcheck"))]
+impl ArbitraryBounded for Signature4 {
+    fn arbitrary_bounded<G: Gen>(g: &mut G, depth: usize) -> Self {
+        use mpi::MPI;
+        use PublicKeyAlgorithm::*;
+
+        let fields = SignatureBuilder::arbitrary_bounded(g, depth);
+        #[allow(deprecated)]
+        let mpis = match fields.pk_algo() {
+            RSAEncryptSign | RSASign => mpi::Signature::RSA  {
+                s: MPI::arbitrary(g),
+            },
+
+            DSA => mpi::Signature::DSA {
+                r: MPI::arbitrary(g),
+                s: MPI::arbitrary(g),
+            },
+
+            EdDSA => mpi::Signature::EdDSA  {
+                r: MPI::arbitrary(g),
+                s: MPI::arbitrary(g),
+            },
+
+            ECDSA => mpi::Signature::ECDSA  {
+                r: MPI::arbitrary(g),
+                s: MPI::arbitrary(g),
+            },
+
+            _ => unreachable!(),
+        };
+
+        Signature4 {
+            common: Arbitrary::arbitrary(g),
+            fields,
+            digest_prefix: [Arbitrary::arbitrary(g),
+                            Arbitrary::arbitrary(g)],
+            mpis,
+            computed_digest: None,
+            level: 0,
+        }
+    }
+}
+
+#[cfg(any(test, feature = "quickcheck"))]
+impl_arbitrary_with_bound!(Signature4);
 
 #[cfg(test)]
 mod test {
     use super::*;
     use crate::KeyID;
+    use crate::cert::prelude::*;
     use crate::crypto;
-    use crate::crypto::mpis::MPI;
-    use crate::Cert;
+    use crate::crypto::mpi::MPI;
     use crate::parse::Parse;
     use crate::packet::Key;
     use crate::packet::key::Key4;
@@ -1276,11 +1445,11 @@ mod test {
         ] {
             let cert = Cert::from_bytes(crate::tests::key(key)).unwrap();
             let mut pair = cert.primary_key().key().clone()
-                .mark_parts_secret().unwrap()
+                .parts_into_secret().unwrap()
                 .into_keypair()
                 .expect("secret key is encrypted/missing");
 
-            let sig = Builder::new(SignatureType::Binary);
+            let sig = SignatureBuilder::new(SignatureType::Binary);
             let hash = hash_algo.context().unwrap();
 
             // Make signature.
@@ -1308,7 +1477,7 @@ mod test {
             .unwrap().into();
         let msg = b"Hello, World";
         let mut pair = key.into_keypair().unwrap();
-        let sig = Builder::new(SignatureType::Binary)
+        let sig = SignatureBuilder::new(SignatureType::Binary)
             .set_signature_creation_time(
                 std::time::SystemTime::now()).unwrap()
             .set_issuer_fingerprint(pair.public().fingerprint()).unwrap()
@@ -1347,11 +1516,11 @@ mod test {
         let mut pnt = [0x40u8; nettle::ed25519::ED25519_KEY_SIZE + 1];
         nettle::ed25519::public_key(&mut pnt[1..], &sec[..]).unwrap();
 
-        let public_mpis = mpis::PublicKey::EdDSA {
+        let public_mpis = mpi::PublicKey::EdDSA {
             curve: Curve::Ed25519,
             q: MPI::new(&pnt[..]),
         };
-        let private_mpis = mpis::SecretKeyMaterial::EdDSA {
+        let private_mpis = mpi::SecretKeyMaterial::EdDSA {
             scalar: MPI::new(&sec[..]).into(),
         };
         let key : key::SecretKey
@@ -1366,7 +1535,7 @@ mod test {
 
         hash.update(&msg[..]);
 
-        Builder::new(SignatureType::Text)
+        SignatureBuilder::new(SignatureType::Text)
             .sign_hash(&mut pair, hash).unwrap();
     }
 
@@ -1385,13 +1554,12 @@ mod test {
             .unwrap();
         let test2 = Cert::from_bytes(
             crate::tests::key("test2-signed-by-test1.pgp")).unwrap();
-        let uid_binding =
-            test2.userids().with_policy(p, None).nth(0).unwrap().bundle();
-        let cert = &uid_binding.certifications()[0];
+        let uid = test2.userids().with_policy(p, None).nth(0).unwrap();
+        let cert = &uid.certifications()[0];
 
         cert.verify_userid_binding(cert_key1,
                                    test2.primary_key().key(),
-                                   uid_binding.userid()).unwrap();
+                                   uid.userid()).unwrap();
     }
 
     #[test]
@@ -1410,41 +1578,37 @@ mod test {
         let keyid = KeyID::from(&fp);
 
         // First, make sure any superfluous subpackets are removed,
-        // yet the Issuer and EmbeddedSignature ones are kept.
-        let mut builder = Builder::new(SignatureType::Text);
-        // This subpacket does not belong there, and should be
-        // removed.
+        // yet the Issuer, IssuerFingerprint and EmbeddedSignature
+        // ones are kept.
+        let mut builder = SignatureBuilder::new(SignatureType::Text);
         builder.unhashed_area_mut().add(Subpacket::new(
             SubpacketValue::IssuerFingerprint(fp.clone()), false).unwrap())
             .unwrap();
         builder.unhashed_area_mut().add(Subpacket::new(
             SubpacketValue::Issuer(keyid.clone()), false).unwrap())
             .unwrap();
+        // This subpacket does not belong there, and should be
+        // removed.
+        builder.unhashed_area_mut().add(Subpacket::new(
+            SubpacketValue::PreferredSymmetricAlgorithms(Vec::new()),
+            false).unwrap()).unwrap();
 
         // Build and add an embedded sig.
-        let embedded_sig = Builder::new(SignatureType::PrimaryKeyBinding)
+        let embedded_sig = SignatureBuilder::new(SignatureType::PrimaryKeyBinding)
             .sign_hash(&mut pair, hash.clone()).unwrap();
         builder.unhashed_area_mut().add(Subpacket::new(
             SubpacketValue::EmbeddedSignature(embedded_sig.into()), false)
                                         .unwrap()).unwrap();
         let sig = builder.sign_hash(&mut pair,
                                     hash.clone()).unwrap().normalize();
-        assert_eq!(sig.unhashed_area().iter().count(), 2);
+        assert_eq!(sig.unhashed_area().iter().count(), 3);
         assert_eq!(*sig.unhashed_area().iter().nth(0).unwrap(),
                    Subpacket::new(SubpacketValue::Issuer(keyid.clone()),
                                   false).unwrap());
         assert_eq!(sig.unhashed_area().iter().nth(1).unwrap().tag(),
                    SubpacketTag::EmbeddedSignature);
-
-        // Now, make sure that an Issuer subpacket is synthesized from
-        // the hashed area for compatibility.
-        let sig = Builder::new(SignatureType::Text)
-            .set_issuer_fingerprint(fp).unwrap()
-            .sign_hash(&mut pair,
-                       hash.clone()).unwrap().normalize();
-        assert_eq!(sig.unhashed_area().iter().count(), 1);
-        assert_eq!(*sig.unhashed_area().iter().nth(0).unwrap(),
-                   Subpacket::new(SubpacketValue::Issuer(keyid.clone()),
+        assert_eq!(*sig.unhashed_area().iter().nth(2).unwrap(),
+                   Subpacket::new(SubpacketValue::IssuerFingerprint(fp.clone()),
                                   false).unwrap());
     }
 
@@ -1454,7 +1618,7 @@ mod test {
             = Key4::generate_ecc(true, Curve::Ed25519).unwrap().into();
         let mut pair = key.into_keypair().unwrap();
 
-        let sig = Builder::new(SignatureType::Standalone)
+        let sig = SignatureBuilder::new(SignatureType::Standalone)
             .set_signature_creation_time(
                 std::time::SystemTime::now()).unwrap()
             .set_issuer_fingerprint(pair.public().fingerprint()).unwrap()
@@ -1486,7 +1650,7 @@ mod test {
             = Key4::generate_ecc(true, Curve::Ed25519).unwrap().into();
         let mut pair = key.into_keypair().unwrap();
 
-        let sig = Builder::new(SignatureType::Timestamp)
+        let sig = SignatureBuilder::new(SignatureType::Timestamp)
             .set_signature_creation_time(
                 std::time::SystemTime::now()).unwrap()
             .set_issuer_fingerprint(pair.public().fingerprint()).unwrap()
@@ -1496,4 +1660,26 @@ mod test {
 
         sig.verify_timestamp(pair.public()).unwrap();
     }
+
+    #[test]
+    fn get_issuers_prefers_fingerprints() -> Result<()> {
+        use crate::KeyHandle;
+        for f in [
+            // This has Fingerprint in the hashed, Issuer in the
+            // unhashed area.
+            "messages/sig.gpg",
+            // This has [Issuer, Fingerprint] in the hashed area.
+            "contrib/gnupg/timestamp-signature-by-alice.asc",
+        ].iter() {
+            let p = Packet::from_bytes(crate::tests::file(f))?;
+            if let Packet::Signature(sig) = p {
+                let issuers = sig.get_issuers();
+                assert_match!(KeyHandle::Fingerprint(_) = &issuers[0]);
+                assert_match!(KeyHandle::KeyID(_) = &issuers[1]);
+            } else {
+                panic!("expected a signature packet");
+            }
+        }
+        Ok(())
+    }
 }
