@@ -10,13 +10,13 @@ use crate::{
     HashAlgorithm,
 };
 use crate::types::Curve;
-use crate::crypto::mpis::{self, MPI};
+use crate::crypto::mpi::{self, MPI};
 use crate::parse::{
     PacketHeaderParser,
     Cookie,
 };
 
-impl mpis::PublicKey {
+impl mpi::PublicKey {
     /// Parses a set of OpenPGP MPIs representing a public key.
     ///
     /// See [Section 3.2 of RFC 4880] for details.
@@ -53,7 +53,7 @@ impl mpis::PublicKey {
                 let n = MPI::parse("rsa_public_n_len", "rsa_public_n", php)?;
                 let e = MPI::parse("rsa_public_e_len", "rsa_public_e", php)?;
 
-                Ok(mpis::PublicKey::RSA { e: e, n: n })
+                Ok(mpi::PublicKey::RSA { e: e, n: n })
             }
 
             DSA => {
@@ -62,7 +62,7 @@ impl mpis::PublicKey {
                 let g = MPI::parse("dsa_public_g_len", "dsa_public_g", php)?;
                 let y = MPI::parse("dsa_public_y_len", "dsa_public_y", php)?;
 
-                Ok(mpis::PublicKey::DSA {
+                Ok(mpi::PublicKey::DSA {
                     p: p,
                     q: q,
                     g: g,
@@ -70,7 +70,7 @@ impl mpis::PublicKey {
                 })
             }
 
-            ElgamalEncrypt | ElgamalEncryptSign => {
+            ElGamalEncrypt | ElGamalEncryptSign => {
                 let p = MPI::parse("elgamal_public_p_len", "elgamal_public_p",
                                    php)?;
                 let g = MPI::parse("elgamal_public_g_len", "elgamal_public_g",
@@ -78,7 +78,7 @@ impl mpis::PublicKey {
                 let y = MPI::parse("elgamal_public_y_len", "elgamal_public_y",
                                    php)?;
 
-                Ok(mpis::PublicKey::Elgamal {
+                Ok(mpi::PublicKey::ElGamal {
                     p: p,
                     g: g,
                     y: y,
@@ -90,7 +90,7 @@ impl mpis::PublicKey {
                 let curve = php.parse_bytes("curve", curve_len)?;
                 let q = MPI::parse("eddsa_public_len", "eddsa_public", php)?;
 
-                Ok(mpis::PublicKey::EdDSA {
+                Ok(mpi::PublicKey::EdDSA {
                     curve: Curve::from_oid(&curve),
                     q: q
                 })
@@ -101,7 +101,7 @@ impl mpis::PublicKey {
                 let curve = php.parse_bytes("curve", curve_len)?;
                 let q = MPI::parse("ecdsa_public_len", "ecdsa_public", php)?;
 
-                Ok(mpis::PublicKey::ECDSA {
+                Ok(mpi::PublicKey::ECDSA {
                     curve: Curve::from_oid(&curve),
                     q: q
                 })
@@ -122,11 +122,11 @@ impl mpis::PublicKey {
                 let hash: HashAlgorithm = php.parse_u8("kdf_hash")?.into();
                 let sym: SymmetricAlgorithm = php.parse_u8("kek_symm")?.into();
 
-                Ok(mpis::PublicKey::ECDH {
+                Ok(mpi::PublicKey::ECDH {
                     curve: Curve::from_oid(&curve),
-                    q: q,
-                    hash: hash,
-                    sym: sym
+                    q,
+                    hash,
+                    sym
                 })
             }
 
@@ -138,22 +138,24 @@ impl mpis::PublicKey {
                 }
                 let rest = php.parse_bytes_eof("rest")?;
 
-                Ok(mpis::PublicKey::Unknown {
+                Ok(mpi::PublicKey::Unknown {
                     mpis: mpis.into_boxed_slice(),
                     rest: rest.into_boxed_slice(),
                 })
             }
+
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
 
-impl mpis::SecretKeyMaterial {
+impl mpi::SecretKeyMaterial {
     /// Parses secret key MPIs for `algo` plus their SHA1 checksum. Fails if the
     /// checksum is wrong.
     pub fn parse_chksumd<T: Read>(algo: PublicKeyAlgorithm, cur: T)
                                   -> Result<Self> {
         use std::io::Cursor;
-        use crate::serialize::Serialize;
+        use crate::serialize::Marshal;
 
         // read mpis
         let bio = buffered_reader::Generic::with_cookie(
@@ -218,7 +220,7 @@ impl mpis::SecretKeyMaterial {
                 let q = MPI::parse("rsa_secret_q_len", "rsa_secret_q", php)?;
                 let u = MPI::parse("rsa_secret_u_len", "rsa_secret_u", php)?;
 
-                Ok(mpis::SecretKeyMaterial::RSA {
+                Ok(mpi::SecretKeyMaterial::RSA {
                     d: d.into(),
                     p: p.into(),
                     q: q.into(),
@@ -229,36 +231,36 @@ impl mpis::SecretKeyMaterial {
             DSA => {
                 let x = MPI::parse("dsa_secret_len", "dsa_secret", php)?;
 
-                Ok(mpis::SecretKeyMaterial::DSA {
+                Ok(mpi::SecretKeyMaterial::DSA {
                     x: x.into(),
                 })
             }
 
-            ElgamalEncrypt | ElgamalEncryptSign => {
+            ElGamalEncrypt | ElGamalEncryptSign => {
                 let x = MPI::parse("elgamal_secret_len", "elgamal_secret",
                                    php)?;
 
-                Ok(mpis::SecretKeyMaterial::Elgamal {
+                Ok(mpi::SecretKeyMaterial::ElGamal {
                     x: x.into(),
                 })
             }
 
             EdDSA => {
-                Ok(mpis::SecretKeyMaterial::EdDSA {
+                Ok(mpi::SecretKeyMaterial::EdDSA {
                     scalar: MPI::parse("eddsa_secret_len", "eddsa_secret", php)?
                                 .into()
                 })
             }
 
             ECDSA => {
-                Ok(mpis::SecretKeyMaterial::ECDSA {
+                Ok(mpi::SecretKeyMaterial::ECDSA {
                     scalar: MPI::parse("ecdsa_secret_len", "ecdsa_secret", php)?
                                 .into()
                 })
             }
 
             ECDH => {
-                Ok(mpis::SecretKeyMaterial::ECDH {
+                Ok(mpi::SecretKeyMaterial::ECDH {
                     scalar: MPI::parse("ecdh_secret_len", "ecdh_secret", php)?
                                 .into()
                 })
@@ -272,16 +274,18 @@ impl mpis::SecretKeyMaterial {
                 }
                 let rest = php.parse_bytes_eof("rest")?;
 
-                Ok(mpis::SecretKeyMaterial::Unknown {
+                Ok(mpi::SecretKeyMaterial::Unknown {
                     mpis: mpis.into_boxed_slice(),
                     rest: rest.into(),
                 })
             }
+
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
 
-impl mpis::Ciphertext {
+impl mpi::Ciphertext {
     /// Parses a set of OpenPGP MPIs representing a ciphertext.
     ///
     /// Expects MPIs for a public key algorithm `algo`s ciphertext.
@@ -317,16 +321,16 @@ impl mpis::Ciphertext {
                 let c = MPI::parse("rsa_ciphertxt_len", "rsa_ciphertxt",
                                    php)?;
 
-                Ok(mpis::Ciphertext::RSA {
+                Ok(mpi::Ciphertext::RSA {
                     c: c,
                 })
             }
 
-            ElgamalEncrypt | ElgamalEncryptSign => {
+            ElGamalEncrypt | ElGamalEncryptSign => {
                 let e = MPI::parse("elgamal_e_len", "elgamal_e", php)?;
                 let c = MPI::parse("elgamal_c_len", "elgamal_c", php)?;
 
-                Ok(mpis::Ciphertext::Elgamal {
+                Ok(mpi::Ciphertext::ElGamal {
                     e: e,
                     c: c,
                 })
@@ -338,7 +342,7 @@ impl mpis::Ciphertext {
                 let key = Vec::from(&php.parse_bytes("ecdh_esk", key_len)?
                                     [..key_len]);
 
-                Ok(mpis::Ciphertext::ECDH {
+                Ok(mpi::Ciphertext::ECDH {
                     e: e, key: key.into_boxed_slice()
                 })
             }
@@ -351,7 +355,7 @@ impl mpis::Ciphertext {
                 }
                 let rest = php.parse_bytes_eof("rest")?;
 
-                Ok(mpis::Ciphertext::Unknown {
+                Ok(mpi::Ciphertext::Unknown {
                     mpis: mpis.into_boxed_slice(),
                     rest: rest.into_boxed_slice(),
                 })
@@ -359,11 +363,13 @@ impl mpis::Ciphertext {
 
             RSASign | DSA | EdDSA | ECDSA => Err(Error::InvalidArgument(
                 format!("not an encryption algorithm: {:?}", algo)).into()),
+
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
 
-impl mpis::Signature {
+impl mpi::Signature {
     /// Parses a set of OpenPGP MPIs representing a signature.
     ///
     /// Expects MPIs for a public key algorithm `algo`s signature.
@@ -398,7 +404,7 @@ impl mpis::Signature {
             RSAEncryptSign | RSASign => {
                 let s = MPI::parse("rsa_signature_len", "rsa_signature", php)?;
 
-                Ok(mpis::Signature::RSA {
+                Ok(mpi::Signature::RSA {
                     s: s,
                 })
             }
@@ -409,19 +415,19 @@ impl mpis::Signature {
                 let s = MPI::parse("dsa_sig_s_len", "dsa_sig_s",
                                    php)?;
 
-                Ok(mpis::Signature::DSA {
+                Ok(mpi::Signature::DSA {
                     r: r,
                     s: s,
                 })
             }
 
-            ElgamalEncryptSign => {
+            ElGamalEncryptSign => {
                 let r = MPI::parse("elgamal_sig_r_len",
                                    "elgamal_sig_r", php)?;
                 let s = MPI::parse("elgamal_sig_s_len",
                                    "elgamal_sig_s", php)?;
 
-                Ok(mpis::Signature::Elgamal {
+                Ok(mpi::Signature::ElGamal {
                     r: r,
                     s: s,
                 })
@@ -433,7 +439,7 @@ impl mpis::Signature {
                 let s = MPI::parse("eddsa_sig_s_len", "eddsa_sig_s",
                                    php)?;
 
-                Ok(mpis::Signature::EdDSA {
+                Ok(mpi::Signature::EdDSA {
                     r: r,
                     s: s,
                 })
@@ -445,7 +451,7 @@ impl mpis::Signature {
                 let s = MPI::parse("ecdsa_sig_s_len", "ecdsa_sig_s",
                                    php)?;
 
-                Ok(mpis::Signature::ECDSA {
+                Ok(mpi::Signature::ECDSA {
                     r: r,
                     s: s,
                 })
@@ -459,14 +465,16 @@ impl mpis::Signature {
                 }
                 let rest = php.parse_bytes_eof("rest")?;
 
-                Ok(mpis::Signature::Unknown {
+                Ok(mpi::Signature::Unknown {
                     mpis: mpis.into_boxed_slice(),
                     rest: rest.into_boxed_slice(),
                 })
             }
 
-            RSAEncrypt | ElgamalEncrypt | ECDH => Err(Error::InvalidArgument(
+            RSAEncrypt | ElGamalEncrypt | ECDH => Err(Error::InvalidArgument(
                 format!("not a signature algorithm: {:?}", algo)).into()),
+
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -475,15 +483,16 @@ impl mpis::Signature {
 fn mpis_parse_test() {
     use super::Parse;
     use crate::PublicKeyAlgorithm::*;
+    use crate::serialize::MarshalInto;
 
     // Dummy RSA public key.
     {
         let buf = b"\x00\x01\x01\x00\x02\x02".to_vec();
-        let mpis = mpis::PublicKey::parse(RSAEncryptSign, buf).unwrap();
+        let mpis = mpi::PublicKey::parse(RSAEncryptSign, buf).unwrap();
 
         //assert_eq!(mpis.serialized_len(), 6);
         match &mpis {
-            &mpis::PublicKey::RSA{ ref n, ref e } => {
+            &mpi::PublicKey::RSA{ ref n, ref e } => {
                 assert_eq!(n.bits(), 1);
                 assert_eq!(n.value()[0], 1);
                 assert_eq!(n.value().len(), 1);
@@ -499,7 +508,7 @@ fn mpis_parse_test() {
     // The number 2.
     {
         let buf = b"\x00\x02\x02".to_vec();
-        let mpis = mpis::Ciphertext::parse(RSAEncryptSign, buf).unwrap();
+        let mpis = mpi::Ciphertext::parse(RSAEncryptSign, buf).unwrap();
 
         assert_eq!(mpis.serialized_len(), 3);
     }
