@@ -1,20 +1,12 @@
 use crate::Result;
-use crate::Cert;
+use crate::cert::prelude::*;
 use crate::packet::{key, Signature, Tag};
 use crate::serialize::{
-    PacketRef, Serialize, SerializeInto,
+    PacketRef,
+    Marshal, MarshalInto,
     generic_serialize_into, generic_export_into,
 };
 
-impl Serialize for Cert {
-    fn serialize(&self, o: &mut dyn std::io::Write) -> Result<()> {
-        self.serialize_common(o, false)
-    }
-
-    fn export(&self, o: &mut dyn std::io::Write) -> Result<()> {
-        self.serialize_common(o, true)
-    }
-}
 
 impl Cert {
     /// Serializes or exports the Cert.
@@ -38,8 +30,8 @@ impl Cert {
     fn serialize_common(&self, o: &mut dyn std::io::Write, export: bool)
                         -> Result<()>
     {
-        let primary = self.primary_key().bundle();
-        PacketRef::PublicKey(primary.key().mark_role_primary_ref())
+        let primary = self.primary_key();
+        PacketRef::PublicKey(primary.key())
             .serialize(o)?;
 
         // Writes a signature if it is exportable or `! export`.
@@ -47,7 +39,7 @@ impl Cert {
             |o: &mut dyn std::io::Write, sig: &Signature| -> Result<()>
         {
             if export {
-                if sig.exportable_certification().unwrap_or(true) {
+                if sig.exportable().is_ok() {
                     PacketRef::Signature(sig).export(o)?;
                 }
             } else {
@@ -69,9 +61,9 @@ impl Cert {
             serialize_sig(o, s)?;
         }
 
-        for u in self.userids().bundles() {
+        for u in self.userids() {
             if export && ! u.self_signatures().iter().chain(u.self_revocations()).any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -92,9 +84,9 @@ impl Cert {
             }
         }
 
-        for u in self.user_attributes().bundles() {
+        for u in self.user_attributes() {
             if export && ! u.self_signatures().iter().chain(u.self_revocations()).any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -117,7 +109,7 @@ impl Cert {
 
         for k in self.subkeys() {
             if export && ! k.self_signatures().iter().chain(k.self_revocations()).any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -140,7 +132,7 @@ impl Cert {
 
         for u in self.unknowns() {
             if export && ! u.certifications().iter().any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -170,12 +162,25 @@ impl Cert {
     }
 }
 
-impl SerializeInto for Cert {
+impl crate::serialize::Serialize for Cert {}
+
+impl Marshal for Cert {
+    fn serialize(&self, o: &mut dyn std::io::Write) -> Result<()> {
+        self.serialize_common(o, false)
+    }
+
+    fn export(&self, o: &mut dyn std::io::Write) -> Result<()> {
+        self.serialize_common(o, true)
+    }
+}
+
+impl crate::serialize::SerializeInto for Cert {}
+
+impl MarshalInto for Cert {
     fn serialized_len(&self) -> usize {
         let mut l = 0;
-        let primary = self.primary_key().bundle();
-        l += PacketRef::PublicKey(primary.key().mark_role_primary_ref())
-            .serialized_len();
+        let primary = self.primary_key();
+        l += PacketRef::PublicKey(primary.key()).serialized_len();
 
         for s in primary.self_revocations() {
             l += PacketRef::Signature(s).serialized_len();
@@ -190,7 +195,7 @@ impl SerializeInto for Cert {
             l += PacketRef::Signature(s).serialized_len();
         }
 
-        for u in self.userids().bundles() {
+        for u in self.userids() {
             l += PacketRef::UserID(u.userid()).serialized_len();
 
             for s in u.self_revocations() {
@@ -207,7 +212,7 @@ impl SerializeInto for Cert {
             }
         }
 
-        for u in self.user_attributes().bundles() {
+        for u in self.user_attributes() {
             l += PacketRef::UserAttribute(u.user_attribute()).serialized_len();
 
             for s in u.self_revocations() {
@@ -287,11 +292,12 @@ impl Cert {
 
 /// A reference to a Cert that allows serialization of secret keys.
 ///
-/// To avoid accidental leakage `Cert::serialize()` skips secret keys.
-/// To serialize `Cert`s with secret keys, use [`Cert::as_tsk()`] to
-/// create a `TSK`, which is a shim on top of the `Cert`, and serialize
-/// this.
+/// To avoid accidental leakage, secret keys are not serialized when a
+/// serializing a [`Cert`].  To serialize [`Cert`]s with secret keys,
+/// use [`Cert::as_tsk()`] to create a `TSK`, which is a shim on top
+/// of the `Cert`, and serialize this.
 ///
+/// [`Cert`]: ../cert/struct.Cert.html
 /// [`Cert::as_tsk()`]: ../cert/struct.Cert.html#method.as_tsk
 ///
 /// # Example
@@ -318,7 +324,7 @@ impl<'a> TSK<'a> {
     /// Creates a new view for the given `Cert`.
     fn new(cert: &'a Cert) -> Self {
         Self {
-            cert: cert,
+            cert,
             filter: None,
         }
     }
@@ -347,7 +353,7 @@ impl<'a> TSK<'a> {
     ///
     /// let cert_ = Cert::from_bytes(&buf)?;
     /// assert_eq!(cert_.keys().with_policy(p, None).alive().revoked(false).secret().count(), 1);
-    /// assert!(cert_.primary_key().secret().is_some());
+    /// assert!(cert_.primary_key().has_secret());
     /// # Ok(()) }
     pub fn set_filter<P>(mut self, predicate: P) -> Self
         where P: 'a + Fn(&'a key::UnspecifiedSecret) -> bool
@@ -369,7 +375,7 @@ impl<'a> TSK<'a> {
             |o: &mut dyn std::io::Write, sig: &Signature| -> Result<()>
         {
             if export {
-                if sig.exportable_certification().unwrap_or(true) {
+                if sig.exportable().is_ok() {
                     PacketRef::Signature(sig).export(o)?;
                 }
             } else {
@@ -383,7 +389,7 @@ impl<'a> TSK<'a> {
             |o: &mut dyn std::io::Write, key: &'a key::UnspecifiedSecret,
              tag_public, tag_secret|
         {
-            let tag = if key.secret().is_some()
+            let tag = if key.has_secret()
                 && self.filter.as_ref().map(|f| f(key)).unwrap_or(true) {
                 tag_secret
             } else {
@@ -403,8 +409,8 @@ impl<'a> TSK<'a> {
             }
         };
 
-        let primary = self.cert.primary_key().bundle();
-        serialize_key(o, primary.key().mark_role_primary_ref().into(),
+        let primary = self.cert.primary_key();
+        serialize_key(o, primary.key().into(),
                       Tag::PublicKey, Tag::SecretKey)?;
 
         for s in primary.self_signatures() {
@@ -420,9 +426,9 @@ impl<'a> TSK<'a> {
             serialize_sig(o, s)?;
         }
 
-        for u in self.cert.userids().bundles() {
+        for u in self.cert.userids() {
             if export && ! u.self_signatures().iter().chain(u.self_revocations()).any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -443,9 +449,9 @@ impl<'a> TSK<'a> {
             }
         }
 
-        for u in self.cert.user_attributes().bundles() {
+        for u in self.cert.user_attributes() {
             if export && ! u.self_signatures().iter().chain(u.self_revocations()).any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -468,7 +474,7 @@ impl<'a> TSK<'a> {
 
         for k in self.cert.subkeys() {
             if export && ! k.self_signatures().iter().chain(k.self_revocations()).any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -492,7 +498,7 @@ impl<'a> TSK<'a> {
 
         for u in self.cert.unknowns() {
             if export && ! u.certifications().iter().any(
-                |s| s.exportable_certification().unwrap_or(true))
+                |s| s.exportable().is_ok())
             {
                 // No exportable selfsig on this component, skip it.
                 continue;
@@ -522,7 +528,9 @@ impl<'a> TSK<'a> {
     }
 }
 
-impl<'a> Serialize for TSK<'a> {
+impl<'a> crate::serialize::Serialize for TSK<'a> {}
+
+impl<'a> Marshal for TSK<'a> {
     fn serialize(&self, o: &mut dyn std::io::Write) -> Result<()> {
         self.serialize_common(o, false)
     }
@@ -532,7 +540,9 @@ impl<'a> Serialize for TSK<'a> {
     }
 }
 
-impl<'a> SerializeInto for TSK<'a> {
+impl<'a> crate::serialize::SerializeInto for TSK<'a> {}
+
+impl<'a> MarshalInto for TSK<'a> {
     fn serialized_len(&self) -> usize {
         let mut l = 0;
 
@@ -540,7 +550,7 @@ impl<'a> SerializeInto for TSK<'a> {
         let serialized_len_key
             = |key: &'a key::UnspecifiedSecret, tag_public, tag_secret|
         {
-            let tag = if key.secret().is_some()
+            let tag = if key.has_secret()
                 && self.filter.as_ref().map(|f| f(key)).unwrap_or(true) {
                 tag_secret
             } else {
@@ -558,8 +568,8 @@ impl<'a> SerializeInto for TSK<'a> {
             packet.serialized_len()
         };
 
-        let primary = self.cert.primary_key().bundle();
-        l += serialized_len_key(primary.key().mark_role_primary_ref().into(),
+        let primary = self.cert.primary_key();
+        l += serialized_len_key(primary.key().into(),
                                 Tag::PublicKey, Tag::SecretKey);
 
         for s in primary.self_signatures() {
@@ -575,7 +585,7 @@ impl<'a> SerializeInto for TSK<'a> {
             l += PacketRef::Signature(s).serialized_len();
         }
 
-        for u in self.cert.userids().bundles() {
+        for u in self.cert.userids() {
             l += PacketRef::UserID(u.userid()).serialized_len();
 
             for s in u.self_revocations() {
@@ -592,7 +602,7 @@ impl<'a> SerializeInto for TSK<'a> {
             }
         }
 
-        for u in self.cert.user_attributes().bundles() {
+        for u in self.cert.user_attributes() {
             l += PacketRef::UserAttribute(u.user_attribute()).serialized_len();
 
             for s in u.self_revocations() {
@@ -664,9 +674,7 @@ impl<'a> SerializeInto for TSK<'a> {
 mod test {
     use super::*;
     use crate::vec_truncate;
-    use crate::cert::components::Amalgamation;
     use crate::parse::Parse;
-    use crate::serialize::Serialize;
     use crate::packet::key;
     use crate::policy::StandardPolicy as P;
 
@@ -744,7 +752,7 @@ mod test {
     #[test]
     fn export() {
         use crate::Packet;
-        use crate::cert::CertBuilder;
+        use crate::cert::prelude::*;
         use crate::types::{Curve, KeyFlags, SignatureType};
         use crate::packet::{
             signature, UserID, user_attribute::{UserAttribute, Subpacket},
@@ -754,14 +762,14 @@ mod test {
         let p = &P::new();
 
         let (cert, _) = CertBuilder::new().generate().unwrap();
-        let mut keypair = cert.primary_key().key().clone().mark_parts_secret()
+        let mut keypair = cert.primary_key().key().clone().parts_into_secret()
             .unwrap().into_keypair().unwrap();
 
         let key: key::SecretSubkey =
             Key4::generate_ecc(false, Curve::Cv25519).unwrap().into();
-        let key_binding = key.mark_parts_public_ref().bind(
+        let key_binding = key.bind(
             &mut keypair, &cert,
-            signature::Builder::new(SignatureType::SubkeyBinding)
+            signature::SignatureBuilder::new(SignatureType::SubkeyBinding)
                 .set_key_flags(
                     &KeyFlags::default().set_transport_encryption(true))
                 .unwrap()
@@ -770,7 +778,7 @@ mod test {
         let uid = UserID::from("foo");
         let uid_binding = uid.bind(
             &mut keypair, &cert,
-            signature::Builder::from(
+            signature::SignatureBuilder::from(
                 cert.primary_key().with_policy(p, None).unwrap()
                     .direct_key_signature().unwrap().clone())
                     .set_type(SignatureType::PositiveCertification)
@@ -781,7 +789,7 @@ mod test {
         ]).unwrap();
         let ua_binding = ua.bind(
             &mut keypair, &cert,
-            signature::Builder::from(
+            signature::SignatureBuilder::from(
                 cert.primary_key().with_policy(p, None).unwrap()
                     .direct_key_signature().unwrap().clone())
                 .set_type(SignatureType::PositiveCertification)
@@ -794,8 +802,7 @@ mod test {
         ]).unwrap();
 
         assert_eq!(cert.subkeys().count(), 1);
-        assert!(cert.subkeys().nth(0).unwrap().binding_signature(p, None)
-                .is_some());
+        cert.subkeys().nth(0).unwrap().binding_signature(p, None).unwrap();
         assert_eq!(cert.userids().count(), 1);
         assert!(cert.userids().with_policy(p, None).nth(0).is_some());
         assert_eq!(cert.user_attributes().count(), 1);
