@@ -1,25 +1,73 @@
-//! Primitive OpenPGP types.
+//! Primitive types.
 //!
 //! This module provides types used in OpenPGP, like enumerations
 //! describing algorithms.
+//!
+//! # Common Operations
+//!
+//!  - *Rounding the creation time of signatures*: See the [`Timestamp::round_down`] method.
+//!  - *Checking key usage flags*: See the [`KeyFlags`] data structure.
+//!  - *Setting key validity ranges*: See the [`Timestamp`] and [`Duration`] data structures.
+//!
+//! # Data structures
+//!
+//! ## `CompressionLevel`
+//!
+//! Allows adjusting the amount of effort spent on compressing encoded data.
+//! This structure additionally has several helper methods for commonly used
+//! compression strategies.
+//!
+//! ## `Features`
+//!
+//! Describes particular features supported by the given OpenPGP implementation.
+//!
+//! ## `KeyFlags`
+//!
+//! Holds imformation about a key in particular how the given key can be used.
+//!
+//! ## `RevocationKey`
+//!
+//! Describes a key that has been designated to issue revocation signatures.
+//!
+//! # `ServerPreferences`
+//!
+//! Describes preferences regarding to key servers.
+//!
+//! ## `Timestamp` and `Duration`
+//!
+//! In OpenPGP time is represented as the number of seconds since the UNIX epoch stored
+//! as an `u32`. These two data structures allow manipulating OpenPGP time ensuring
+//! that adding or subtracting durations will never overflow or underflow without
+//! notice.
+//!
+//! [`Timestamp::round_down`]: struct.Timestamp.html#method.round_down
+//! [`KeyFlags`]: struct.KeyFlags.html
+//! [`Timestamp`]: struct.Timestamp.html
+//! [`Duration`]: struct.Duration.html
 
 use std::fmt;
 use std::str::FromStr;
 use std::result;
 
+#[cfg(any(test, feature = "quickcheck"))]
 use quickcheck::{Arbitrary, Gen};
 
 use crate::Error;
 use crate::Result;
 
+mod compression_level;
+pub use compression_level::CompressionLevel;
 mod features;
 pub use self::features::Features;
 mod key_flags;
 pub use self::key_flags::KeyFlags;
+mod revocation_key;
+pub use revocation_key::RevocationKey;
 mod server_preferences;
 pub use self::server_preferences::KeyServerPreferences;
 mod timestamp;
 pub use timestamp::{Timestamp, Duration};
+pub(crate) use timestamp::normalize_systemtime;
 
 /// Removes padding bytes from bitfields.
 ///
@@ -36,6 +84,20 @@ pub(crate) fn bitfield_remove_padding(b: &mut Vec<u8>) -> usize {
 /// The OpenPGP public key algorithms as defined in [Section 9.1 of
 /// RFC 4880], and [Section 5 of RFC 6637].
 ///
+/// # Examples
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use openpgp::cert::prelude::*;
+/// use openpgp::types::PublicKeyAlgorithm;
+///
+/// let (cert, _) = CertBuilder::new()
+///     .set_cipher_suite(CipherSuite::Cv25519)
+///     .generate().unwrap();
+///
+/// assert_eq!(cert.primary_key().pk_algo(), PublicKeyAlgorithm::EdDSA);
+/// ```
+///
 ///   [Section 9.1 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-9.1
 ///   [Section 5 of RFC 6637]: https://tools.ietf.org/html/rfc6637
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug, PartialOrd, Ord)]
@@ -50,59 +112,100 @@ pub enum PublicKeyAlgorithm {
     #[deprecated(since = "rfc4880",
                  note = "Use `PublicKeyAlgorithm::RSAEncryptSign`.")]
     RSASign,
-    /// Elgamal (Encrypt-Only)
-    ElgamalEncrypt,
+    /// ElGamal (Encrypt-Only)
+    ElGamalEncrypt,
     /// DSA (Digital Signature Algorithm)
     DSA,
     /// Elliptic curve DH
     ECDH,
     /// Elliptic curve DSA
     ECDSA,
-    /// Elgamal (Encrypt or Sign)
+    /// ElGamal (Encrypt or Sign)
     #[deprecated(since = "rfc4880",
                  note = "If you really must, use \
-                         `PublicKeyAlgorithm::ElgamalEncrypt`.")]
-    ElgamalEncryptSign,
+                         `PublicKeyAlgorithm::ElGamalEncrypt`.")]
+    ElGamalEncryptSign,
     /// "Twisted" Edwards curve DSA
     EdDSA,
     /// Private algorithm identifier.
     Private(u8),
     /// Unknown algorithm identifier.
     Unknown(u8),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl PublicKeyAlgorithm {
     /// Returns true if the algorithm can sign data.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::PublicKeyAlgorithm;
+    ///
+    /// assert!(PublicKeyAlgorithm::EdDSA.for_signing());
+    /// assert!(PublicKeyAlgorithm::RSAEncryptSign.for_signing());
+    /// assert!(!PublicKeyAlgorithm::ElGamalEncrypt.for_signing());
+    /// ```
     pub fn for_signing(&self) -> bool {
         use self::PublicKeyAlgorithm::*;
         #[allow(deprecated)]
         match &self {
-            RSAEncryptSign | RSASign | DSA | ECDSA | ElgamalEncryptSign
+            RSAEncryptSign | RSASign | DSA | ECDSA | ElGamalEncryptSign
                 | EdDSA => true,
+            __Nonexhaustive => unreachable!(),
             _ => false,
         }
     }
 
     /// Returns true if the algorithm can encrypt data.
-    pub fn can_encrypt(&self) -> bool {
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::PublicKeyAlgorithm;
+    ///
+    /// assert!(!PublicKeyAlgorithm::EdDSA.for_encryption());
+    /// assert!(PublicKeyAlgorithm::RSAEncryptSign.for_encryption());
+    /// assert!(PublicKeyAlgorithm::ElGamalEncrypt.for_encryption());
+    /// ```
+    pub fn for_encryption(&self) -> bool {
         use self::PublicKeyAlgorithm::*;
         #[allow(deprecated)]
         match &self {
-            RSAEncryptSign | RSAEncrypt | ElgamalEncrypt | ECDH
-                | ElgamalEncryptSign => true,
+            RSAEncryptSign | RSAEncrypt | ElGamalEncrypt | ECDH
+                | ElGamalEncryptSign => true,
+            __Nonexhaustive => unreachable!(),
             _ => false,
         }
     }
 
     /// Returns whether this algorithm is supported.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::PublicKeyAlgorithm;
+    ///
+    /// assert!(PublicKeyAlgorithm::EdDSA.is_supported());
+    /// assert!(PublicKeyAlgorithm::RSAEncryptSign.is_supported());
+    /// assert!(!PublicKeyAlgorithm::ElGamalEncrypt.is_supported());
+    /// assert!(!PublicKeyAlgorithm::Private(101).is_supported());
+    /// ```
     pub fn is_supported(&self) -> bool {
         use self::PublicKeyAlgorithm::*;
         #[allow(deprecated)]
         match &self {
             RSAEncryptSign | RSAEncrypt | RSASign | DSA | ECDH | ECDSA | EdDSA
                 => true,
-            ElgamalEncrypt | ElgamalEncryptSign | Private(_) | Unknown(_)
+            ElGamalEncrypt | ElGamalEncryptSign | Private(_) | Unknown(_)
                 => false,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -115,11 +218,11 @@ impl From<u8> for PublicKeyAlgorithm {
             1 => RSAEncryptSign,
             2 => RSAEncrypt,
             3 => RSASign,
-            16 => ElgamalEncrypt,
+            16 => ElGamalEncrypt,
             17 => DSA,
             18 => ECDH,
             19 => ECDSA,
-            20 => ElgamalEncryptSign,
+            20 => ElGamalEncryptSign,
             22 => EdDSA,
             100..=110 => Private(u),
             u => Unknown(u),
@@ -135,14 +238,15 @@ impl From<PublicKeyAlgorithm> for u8 {
             RSAEncryptSign => 1,
             RSAEncrypt => 2,
             RSASign => 3,
-            ElgamalEncrypt => 16,
+            ElGamalEncrypt => 16,
             DSA => 17,
             ECDH => 18,
             ECDSA => 19,
-            ElgamalEncryptSign => 20,
+            ElGamalEncryptSign => 20,
             EdDSA => 22,
             Private(u) => u,
             Unknown(u) => u,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -155,26 +259,47 @@ impl fmt::Display for PublicKeyAlgorithm {
             RSAEncryptSign => f.write_str("RSA (Encrypt or Sign)"),
             RSAEncrypt => f.write_str("RSA Encrypt-Only"),
             RSASign => f.write_str("RSA Sign-Only"),
-            ElgamalEncrypt => f.write_str("Elgamal (Encrypt-Only)"),
+            ElGamalEncrypt => f.write_str("ElGamal (Encrypt-Only)"),
             DSA => f.write_str("DSA (Digital Signature Algorithm)"),
             ECDSA => f.write_str("ECDSA public key algorithm"),
-            ElgamalEncryptSign => f.write_str("Elgamal (Encrypt or Sign)"),
+            ElGamalEncryptSign => f.write_str("ElGamal (Encrypt or Sign)"),
             ECDH => f.write_str("ECDH public key algorithm"),
             EdDSA => f.write_str("EdDSA Edwards-curve Digital Signature Algorithm"),
             Private(u) =>
                 f.write_fmt(format_args!("Private/Experimental public key algorithm {}", u)),
             Unknown(u) =>
                 f.write_fmt(format_args!("Unknown public key algorithm {}", u)),
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for PublicKeyAlgorithm {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
+impl PublicKeyAlgorithm {
+    pub(crate) fn arbitrary_for_signing<G: Gen>(g: &mut G) -> Self {
+        use rand::Rng;
+        use self::PublicKeyAlgorithm::*;
+        #[allow(deprecated)]
+        let a = match g.gen_range(0, 5) {
+            0 => RSAEncryptSign,
+            1 => RSASign,
+            2 => DSA,
+            3 => ECDSA,
+            4 => EdDSA,
+            _ => unreachable!(),
+        };
+        assert!(a.for_signing());
+        a
+    }
+}
+
 /// Elliptic curves used in OpenPGP.
 ///
 /// `PublicKeyAlgorithm` does not differentiate between elliptic
@@ -199,6 +324,10 @@ pub enum Curve {
     Cv25519,
     /// Unknown curve.
     Unknown(Box<[u8]>),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl Curve {
@@ -212,6 +341,18 @@ impl Curve {
     /// only because some legacy PGP application like HKP need it.
     ///
     /// Returns `None` for unknown curves.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::Curve;
+    ///
+    /// assert_eq!(Curve::NistP256.bits(), Some(256));
+    /// assert_eq!(Curve::NistP384.bits(), Some(384));
+    /// assert_eq!(Curve::Ed25519.bits(), Some(256));
+    /// assert_eq!(Curve::Unknown(Box::new([0x2B, 0x11])).bits(), None);
+    /// ```
     pub fn bits(&self) -> Option<usize> {
         use self::Curve::*;
 
@@ -224,6 +365,7 @@ impl Curve {
             Ed25519 => Some(256),
             Cv25519 => Some(256),
             Unknown(_) => None,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -243,6 +385,7 @@ impl fmt::Display for Curve {
                 => f.write_str("Elliptic curve Diffie-Hellman using D.J. Bernstein's Curve25519"),
             Unknown(ref oid)
              => write!(f, "Unknown curve (OID: {:?})", oid),
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -261,6 +404,16 @@ const CV25519_OID: &[u8] =
 
 impl Curve {
     /// Parses the given OID.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::Curve;
+    ///
+    /// assert_eq!(Curve::from_oid(&[0x2B, 0x81, 0x04, 0x00, 0x22]), Curve::NistP384);
+    /// assert_eq!(Curve::from_oid(&[0x2B, 0x11]), Curve::Unknown(Box::new([0x2B, 0x11])));
+    /// ```
     pub fn from_oid(oid: &[u8]) -> Curve {
         // Match on OIDs, see section 11 of RFC6637.
         match oid {
@@ -276,6 +429,16 @@ impl Curve {
     }
 
     /// Returns this curve's OID.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::Curve;
+    ///
+    /// assert_eq!(Curve::NistP384.oid(), &[0x2B, 0x81, 0x04, 0x00, 0x22]);
+    /// assert_eq!(Curve::Unknown(Box::new([0x2B, 0x11])).oid(), &[0x2B, 0x11]);
+    /// ```
     pub fn oid(&self) -> &[u8] {
         match self {
             &Curve::NistP256 => NIST_P256_OID,
@@ -286,11 +449,24 @@ impl Curve {
             &Curve::Ed25519 => ED25519_OID,
             &Curve::Cv25519 => CV25519_OID,
             &Curve::Unknown(ref oid) => oid,
+            Curve::__Nonexhaustive => unreachable!(),
         }
     }
 
     /// Returns the length of a coordinate in bits.
     ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::Curve;
+    ///
+    /// assert!(if let Ok(256) = Curve::NistP256.len() { true } else { false });
+    /// assert!(if let Ok(384) = Curve::NistP384.len() { true } else { false });
+    /// assert!(if let Ok(256) = Curve::Ed25519.len() { true } else { false });
+    /// assert!(if let Err(_) = Curve::Unknown(Box::new([0x2B, 0x11])).len() { true } else { false });
+    /// ```
+    ///
     /// # Errors
     ///
     /// Returns `Error::UnsupportedEllipticCurve` if the curve is not
@@ -307,10 +483,23 @@ impl Curve {
             &Curve::Unknown(_) =>
                 Err(Error::UnsupportedEllipticCurve(self.clone())
                     .into()),
+            Curve::__Nonexhaustive => unreachable!(),
         }
     }
 
     /// Returns whether this algorithm is supported.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::Curve;
+    ///
+    /// assert!(Curve::NistP256.is_supported());
+    /// assert!(Curve::NistP384.is_supported());
+    /// assert!(Curve::Ed25519.is_supported());
+    /// assert!(!Curve::Unknown(Box::new([0x2B, 0x11])).is_supported());
+    /// ```
     pub fn is_supported(&self) -> bool {
         use self::Curve::*;
         match &self {
@@ -318,10 +507,12 @@ impl Curve {
                 => true,
             BrainpoolP256 | BrainpoolP512 | Unknown(_)
                 => false,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for Curve {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         match u8::arbitrary(g) % 8 {
@@ -348,6 +539,24 @@ impl Arbitrary for Curve {
 /// symbolic one.
 ///
 ///   [`SymmetricAlgorithm::from`]: https://doc.rust-lang.org/std/convert/trait.From.html
+///
+/// # Examples
+///
+/// Use `SymmetricAlgorithm` to set the preferred symmetric algorithms on a signature:
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use openpgp::packet::signature::SignatureBuilder;
+/// use openpgp::types::{HashAlgorithm, SymmetricAlgorithm, SignatureType};
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let mut builder = SignatureBuilder::new(SignatureType::DirectKey)
+///     .set_hash_algo(HashAlgorithm::SHA512)
+///     .set_preferred_symmetric_algorithms(vec![
+///         SymmetricAlgorithm::AES256,
+///     ])?;
+/// # Ok(()) }
+/// ```
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug, PartialOrd, Ord)]
 pub enum SymmetricAlgorithm {
     /// Null encryption.
@@ -378,6 +587,10 @@ pub enum SymmetricAlgorithm {
     Private(u8),
     /// Unknown algorithm identifier.
     Unknown(u8),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl Default for SymmetricAlgorithm {
@@ -388,6 +601,20 @@ impl Default for SymmetricAlgorithm {
 
 impl SymmetricAlgorithm {
     /// Returns whether this algorithm is supported.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::SymmetricAlgorithm;
+    ///
+    /// assert!(SymmetricAlgorithm::AES256.is_supported());
+    /// assert!(SymmetricAlgorithm::TripleDES.is_supported());
+    ///
+    /// assert!(!SymmetricAlgorithm::IDEA.is_supported());
+    /// assert!(!SymmetricAlgorithm::Unencrypted.is_supported());
+    /// assert!(!SymmetricAlgorithm::Private(101).is_supported());
+    /// ```
     pub fn is_supported(&self) -> bool {
         use self::SymmetricAlgorithm::*;
         match &self {
@@ -396,6 +623,7 @@ impl SymmetricAlgorithm {
                 => true,
             Unencrypted | IDEA | Private(_) | Unknown(_)
                 => false,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -438,6 +666,7 @@ impl From<SymmetricAlgorithm> for u8 {
             SymmetricAlgorithm::Camellia256 => 13,
             SymmetricAlgorithm::Private(u) => u,
             SymmetricAlgorithm::Unknown(u) => u,
+            SymmetricAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -473,10 +702,12 @@ impl fmt::Display for SymmetricAlgorithm {
                 f.write_fmt(format_args!("Private/Experimental symmetric key algorithm {}", u)),
             SymmetricAlgorithm::Unknown(u) =>
                 f.write_fmt(format_args!("Unknown symmetric key algorithm {}", u)),
+            SymmetricAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for SymmetricAlgorithm {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
@@ -493,6 +724,26 @@ impl Arbitrary for SymmetricAlgorithm {
 /// symbolic one.
 ///
 ///   [`AEADAlgorithm::from`]: https://doc.rust-lang.org/std/convert/trait.From.html
+///
+/// This feature is [experimental](../index.html#experimental-features).
+///
+/// # Examples
+///
+/// Use `AEADAlgorithm` to set the preferred AEAD algorithms on a signature:
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use openpgp::packet::signature::SignatureBuilder;
+/// use openpgp::types::{Features, HashAlgorithm, AEADAlgorithm, SignatureType};
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let features = Features::default().set_aead(true);
+/// let mut builder = SignatureBuilder::new(SignatureType::DirectKey)
+///     .set_features(&features)?
+///     .set_preferred_aead_algorithms(vec![
+///         AEADAlgorithm::EAX,
+///     ])?;
+/// # Ok(()) }
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug, PartialOrd, Ord)]
 pub enum AEADAlgorithm {
     /// EAX mode.
@@ -503,10 +754,25 @@ pub enum AEADAlgorithm {
     Private(u8),
     /// Unknown algorithm identifier.
     Unknown(u8),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl AEADAlgorithm {
     /// Returns whether this algorithm is supported.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::AEADAlgorithm;
+    ///
+    /// assert!(AEADAlgorithm::EAX.is_supported());
+    ///
+    /// assert!(!AEADAlgorithm::OCB.is_supported());
+    /// ```
     pub fn is_supported(&self) -> bool {
         use self::AEADAlgorithm::*;
         match &self {
@@ -514,6 +780,7 @@ impl AEADAlgorithm {
                 => true,
             OCB | Private(_) | Unknown(_)
                 => false,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -536,6 +803,7 @@ impl From<AEADAlgorithm> for u8 {
             AEADAlgorithm::OCB => 2,
             AEADAlgorithm::Private(u) => u,
             AEADAlgorithm::Unknown(u) => u,
+            AEADAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -551,10 +819,12 @@ impl fmt::Display for AEADAlgorithm {
                 f.write_fmt(format_args!("Private/Experimental AEAD algorithm {}", u)),
             AEADAlgorithm::Unknown(u) =>
                 f.write_fmt(format_args!("Unknown AEAD algorithm {}", u)),
+            AEADAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for AEADAlgorithm {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
@@ -564,6 +834,25 @@ impl Arbitrary for AEADAlgorithm {
 /// The OpenPGP compression algorithms as defined in [Section 9.3 of RFC 4880].
 ///
 ///   [Section 9.3 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-9.3
+///
+/// # Examples
+///
+/// Use `CompressionAlgorithm` to set the preferred compressions algorithms on
+/// a signature:
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use openpgp::packet::signature::SignatureBuilder;
+/// use openpgp::types::{HashAlgorithm, CompressionAlgorithm, SignatureType};
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let mut builder = SignatureBuilder::new(SignatureType::DirectKey)
+///     .set_hash_algo(HashAlgorithm::SHA512)
+///     .set_preferred_compression_algorithms(vec![
+///         CompressionAlgorithm::Zlib,
+///         CompressionAlgorithm::BZip2,
+///     ])?;
+/// # Ok(()) }
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug, PartialOrd, Ord)]
 pub enum CompressionAlgorithm {
     /// Null compression.
@@ -588,6 +877,10 @@ pub enum CompressionAlgorithm {
     Private(u8),
     /// Unknown compression algorithm identifier.
     Unknown(u8),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl Default for CompressionAlgorithm {
@@ -606,6 +899,18 @@ impl Default for CompressionAlgorithm {
 
 impl CompressionAlgorithm {
     /// Returns whether this algorithm is supported.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::CompressionAlgorithm;
+    ///
+    /// assert!(CompressionAlgorithm::Uncompressed.is_supported());
+    /// assert!(CompressionAlgorithm::Zip.is_supported());
+    ///
+    /// assert!(!CompressionAlgorithm::Private(101).is_supported());
+    /// ```
     pub fn is_supported(&self) -> bool {
         use self::CompressionAlgorithm::*;
         match &self {
@@ -614,6 +919,7 @@ impl CompressionAlgorithm {
             Zip | Zlib => true,
             #[cfg(feature = "compression-bzip2")]
             BZip2 => true,
+            __Nonexhaustive => unreachable!(),
             _ => false,
         }
     }
@@ -641,6 +947,7 @@ impl From<CompressionAlgorithm> for u8 {
             CompressionAlgorithm::BZip2 => 3,
             CompressionAlgorithm::Private(u) => u,
             CompressionAlgorithm::Unknown(u) => u,
+            CompressionAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -656,10 +963,12 @@ impl fmt::Display for CompressionAlgorithm {
                 f.write_fmt(format_args!("Private/Experimental compression algorithm {}", u)),
             CompressionAlgorithm::Unknown(u) =>
                 f.write_fmt(format_args!("Unknown comppression algorithm {}", u)),
+            CompressionAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for CompressionAlgorithm {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
@@ -668,7 +977,22 @@ impl Arbitrary for CompressionAlgorithm {
 
 /// The OpenPGP hash algorithms as defined in [Section 9.4 of RFC 4880].
 ///
-///   [Section 9.4 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-9.4
+/// # Examples
+///
+/// Use `HashAlgorithm` to set the preferred hash algorithms on a signature:
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use openpgp::packet::signature::SignatureBuilder;
+/// use openpgp::types::{HashAlgorithm, SignatureType};
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let mut builder = SignatureBuilder::new(SignatureType::DirectKey)
+///     .set_hash_algo(HashAlgorithm::SHA512);
+/// # Ok(()) }
+/// ```
+///
+/// [Section 9.4 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-9.4
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug, PartialOrd, Ord)]
 pub enum HashAlgorithm {
     /// Rivest et.al. message digest 5.
@@ -689,11 +1013,17 @@ pub enum HashAlgorithm {
     Private(u8),
     /// Unknown hash algorithm identifier.
     Unknown(u8),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl Default for HashAlgorithm {
     fn default() -> Self {
-        HashAlgorithm::SHA256
+        // SHA512 is almost twice as fast as SHA256 on 64-bit
+        // architectures because it operates on 64-bit words.
+        HashAlgorithm::SHA512
     }
 }
 
@@ -725,6 +1055,7 @@ impl From<HashAlgorithm> for u8 {
             HashAlgorithm::SHA224 => 11,
             HashAlgorithm::Private(u) => u,
             HashAlgorithm::Unknown(u) => u,
+            HashAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -767,10 +1098,12 @@ impl fmt::Display for HashAlgorithm {
                 f.write_fmt(format_args!("Private/Experimental hash algorithm {}", u)),
             HashAlgorithm::Unknown(u) =>
                 f.write_fmt(format_args!("Unknown hash algorithm {}", u)),
+            HashAlgorithm::__Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for HashAlgorithm {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
@@ -780,6 +1113,22 @@ impl Arbitrary for HashAlgorithm {
 /// Signature type as defined in [Section 5.2.1 of RFC 4880].
 ///
 ///   [Section 5.2.1 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-5.2.1
+///
+/// # Examples
+///
+/// Use `SignatureType` to create a timestamp signature:
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use std::time::SystemTime;
+/// use openpgp::packet::signature::SignatureBuilder;
+/// use openpgp::types::SignatureType;
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let mut builder = SignatureBuilder::new(SignatureType::Timestamp)
+///     .set_signature_creation_time(SystemTime::now())?;
+/// # Ok(()) }
+/// ```
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug)]
 pub enum SignatureType {
     /// Signature over a binary document.
@@ -819,6 +1168,10 @@ pub enum SignatureType {
 
     /// Catchall.
     Unknown(u8),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl From<u8> for SignatureType {
@@ -863,6 +1216,7 @@ impl From<SignatureType> for u8 {
             SignatureType::Timestamp => 0x40,
             SignatureType::Confirmation => 0x50,
             SignatureType::Unknown(u) => u,
+            SignatureType::__Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -902,10 +1256,12 @@ impl fmt::Display for SignatureType {
                 f.write_str("Confirmation"),
             SignatureType::Unknown(u) =>
                 f.write_fmt(format_args!("Unknown signature type 0x{:x}", u)),
+            SignatureType::__Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for SignatureType {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
@@ -917,6 +1273,55 @@ impl Arbitrary for SignatureType {
 /// See the description of revocation subpackets [Section 5.2.3.23 of RFC 4880].
 ///
 ///   [Section 5.2.3.23 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-5.2.3.23
+///
+/// # Examples
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use openpgp::cert::prelude::*;
+/// use openpgp::policy::StandardPolicy;
+/// use openpgp::types::{RevocationStatus, ReasonForRevocation, SignatureType};
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let p = &StandardPolicy::new();
+///
+/// // A certificate with a User ID.
+/// let (cert, _) = CertBuilder::new()
+///     .add_userid("Alice <alice@example.org>")
+///     .generate()?;
+///
+/// let mut keypair = cert.primary_key().key().clone()
+///     .parts_into_secret()?.into_keypair()?;
+/// let ca = cert.userids().nth(0).unwrap();
+///
+/// // Generate the revocation for the first and only UserID.
+/// let revocation =
+///     UserIDRevocationBuilder::new()
+///     .set_reason_for_revocation(
+///         ReasonForRevocation::UIDRetired,
+///         b"Left example.org.")?
+///     .build(&mut keypair, &cert, ca.userid(), None)?;
+/// assert_eq!(revocation.typ(), SignatureType::CertificationRevocation);
+///
+/// // Now merge the revocation signature into the Cert.
+/// let cert = cert.merge_packets(revocation.clone())?;
+///
+/// // Check that it is revoked.
+/// let ca = cert.userids().nth(0).unwrap();
+/// let status = ca.with_policy(p, None)?.revocation_status();
+/// if let RevocationStatus::Revoked(revs) = status {
+///     assert_eq!(revs.len(), 1);
+///     let rev = revs[0];
+///
+///     assert_eq!(rev.typ(), SignatureType::CertificationRevocation);
+///     assert_eq!(rev.reason_for_revocation(),
+///                Some((ReasonForRevocation::UIDRetired,
+///                      "Left example.org.".as_bytes())));
+///    // User ID has been revoked.
+/// }
+/// # else { unreachable!(); }
+/// # Ok(()) }
+/// ```
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug, PartialOrd, Ord)]
 pub enum ReasonForRevocation {
     /// No reason specified (key revocations or cert revocations)
@@ -939,6 +1344,10 @@ pub enum ReasonForRevocation {
 
     /// Unknown reason identifier.
     Unknown(u8),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl From<u8> for ReasonForRevocation {
@@ -967,6 +1376,7 @@ impl From<ReasonForRevocation> for u8 {
             UIDRetired => 32,
             Private(u) => u,
             Unknown(u) => u,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -991,10 +1401,12 @@ impl fmt::Display for ReasonForRevocation {
             Unknown(u) =>
                 f.write_fmt(format_args!(
                     "Unknown revocation reason {}", u)),
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for ReasonForRevocation {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
@@ -1012,7 +1424,64 @@ impl Arbitrary for ReasonForRevocation {
 /// should be considered invalid *after* the revocation signature's
 /// creation time.  `KeySuperseded`, `KeyRetired`, and `UIDRetired`
 /// are considered soft revocations.
-#[derive(Clone, Copy, PartialEq, Eq)]
+///
+/// # Examples
+///
+/// A certificate is considered to be revoked when a hard revocation is present
+/// even if it is not live at the specified time.
+///
+/// Here, a certificate is generated at `t0` and then revoked later at `t2`.
+/// At `t1` (`t0` < `t1` < `t2`) depending on the revocation type it will be
+/// either considered revoked (hard revocation) or not revoked (soft revocation):
+///
+/// ```rust
+/// # use sequoia_openpgp as openpgp;
+/// use std::time::{Duration, SystemTime};
+/// use openpgp::cert::prelude::*;
+/// use openpgp::types::{RevocationStatus, ReasonForRevocation};
+/// use openpgp::policy::StandardPolicy;
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let p = &StandardPolicy::new();
+///
+/// let t0 = SystemTime::now();
+/// let (cert, _) =
+///     CertBuilder::general_purpose(None, Some("alice@example.org"))
+///     .set_creation_time(t0)
+///     .generate()?;
+///
+/// let t2 = t0 + Duration::from_secs(3600);
+///
+/// let mut signer = cert.primary_key().key().clone()
+///     .parts_into_secret()?.into_keypair()?;
+///
+/// // Create a hard revocation (KeyCompromised):
+/// let sig = CertRevocationBuilder::new()
+///     .set_reason_for_revocation(ReasonForRevocation::KeyCompromised,
+///                                b"The butler did it :/")?
+///     .set_signature_creation_time(t2)?
+///     .build(&mut signer, &cert, None)?;
+///
+/// let t1 = t0 + Duration::from_secs(1200);
+/// let cert1 = cert.clone().merge_packets(sig.clone())?;
+/// assert_eq!(cert1.revocation_status(p, Some(t1)),
+///            RevocationStatus::Revoked(vec![ &sig.into() ]));
+///
+/// // Create a soft revocation (KeySuperseded):
+/// let sig = CertRevocationBuilder::new()
+///     .set_reason_for_revocation(ReasonForRevocation::KeySuperseded,
+///                                b"Migrated to key XYZ")?
+///     .set_signature_creation_time(t2)?
+///     .build(&mut signer, &cert, None)?;
+///
+/// let t1 = t0 + Duration::from_secs(1200);
+/// let cert2 = cert.clone().merge_packets(sig.clone())?;
+/// assert_eq!(cert2.revocation_status(p, Some(t1)),
+///            RevocationStatus::NotAsFarAsWeKnow);
+/// #     Ok(())
+/// # }
+/// ```
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
 pub enum RevocationType {
     /// A hard revocation.
     ///
@@ -1034,6 +1503,18 @@ pub enum RevocationType {
 
 impl ReasonForRevocation {
     /// Returns the revocation's `RevocationType`.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::types::{ReasonForRevocation, RevocationType};
+    ///
+    /// assert_eq!(ReasonForRevocation::KeyCompromised.revocation_type(), RevocationType::Hard);
+    /// assert_eq!(ReasonForRevocation::Private(101).revocation_type(), RevocationType::Hard);
+    ///
+    /// assert_eq!(ReasonForRevocation::KeyRetired.revocation_type(), RevocationType::Soft);
+    /// ```
     pub fn revocation_type(&self) -> RevocationType {
         match self {
             ReasonForRevocation::Unspecified => RevocationType::Hard,
@@ -1043,6 +1524,7 @@ impl ReasonForRevocation {
             ReasonForRevocation::UIDRetired => RevocationType::Soft,
             ReasonForRevocation::Private(_) => RevocationType::Hard,
             ReasonForRevocation::Unknown(_) => RevocationType::Hard,
+            ReasonForRevocation::__Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -1052,6 +1534,28 @@ impl ReasonForRevocation {
 /// See the description of literal data packets [Section 5.9 of RFC 4880].
 ///
 ///   [Section 5.9 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-5.9
+///
+/// # Examples
+///
+/// Construct a new [`Message`] containing one text literal packet:
+///
+/// [`Message`]: struct.Message.html
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use std::convert::TryFrom;
+/// use openpgp::packet::prelude::*;
+/// use openpgp::types::DataFormat;
+/// use openpgp::message::Message;
+///
+/// let mut packets = Vec::new();
+/// let mut lit = Literal::new(DataFormat::Text);
+/// lit.set_body(b"data".to_vec());
+/// packets.push(lit.into());
+///
+/// let message = Message::try_from(packets);
+/// assert!(message.is_ok(), "{:?}", message);
+/// ```
 #[derive(Clone, Copy, Hash, PartialEq, Eq, Debug, PartialOrd, Ord)]
 pub enum DataFormat {
     /// Binary data.
@@ -1079,6 +1583,10 @@ pub enum DataFormat {
 
     /// Unknown format specifier.
     Unknown(char),
+
+    /// This marks this enum as non-exhaustive.  Do not use this
+    /// variant.
+    #[doc(hidden)] __Nonexhaustive,
 }
 
 impl Default for DataFormat {
@@ -1121,6 +1629,7 @@ impl From<DataFormat> for char {
             Unicode => 'u',
             MIME => 'm',
             Unknown(c) => c,
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
@@ -1140,16 +1649,77 @@ impl fmt::Display for DataFormat {
             Unknown(c) =>
                 f.write_fmt(format_args!(
                     "Unknown data format identifier {:?}", c)),
+            __Nonexhaustive => unreachable!(),
         }
     }
 }
 
+#[cfg(any(test, feature = "quickcheck"))]
 impl Arbitrary for DataFormat {
     fn arbitrary<G: Gen>(g: &mut G) -> Self {
         u8::arbitrary(g).into()
     }
 }
 
+/// The revocation status.
+///
+/// # Examples
+///
+/// Generates a new certificate then checks if the User ID is revoked or not under
+/// the given policy using [`ValidUserIDAmalgamation`]:
+///
+/// [`ValidUserIDAmalgamation`]: ../cert/amalgamation/type.ValidUserIDAmalgamation.html
+///
+/// ```rust
+/// use sequoia_openpgp as openpgp;
+/// use openpgp::cert::prelude::*;
+/// use openpgp::policy::StandardPolicy;
+/// use openpgp::types::RevocationStatus;
+///
+/// # fn main() -> openpgp::Result<()> {
+/// let p = &StandardPolicy::new();
+///
+/// let (cert, _) =
+///     CertBuilder::general_purpose(None, Some("alice@example.org"))
+///     .generate()?;
+/// let cert = cert.with_policy(p, None)?;
+/// let ua = cert.userids().nth(0).expect("User IDs");
+///
+/// match ua.revocation_status() {
+///     RevocationStatus::Revoked(revs) => {
+///         // The certificate holder revoked the User ID.
+/// #       unreachable!();
+///     }
+///     RevocationStatus::CouldBe(revs) => {
+///         // There are third-party revocations.  You still need
+///         // to check that they are valid (this is necessary,
+///         // because without the Certificates are not normally
+///         // available to Sequoia).
+/// #       unreachable!();
+///     }
+///     RevocationStatus::NotAsFarAsWeKnow => {
+///         // We have no evidence that the User ID is revoked.
+///     }
+/// }
+/// #     Ok(())
+/// # }
+/// ```
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub enum RevocationStatus<'a> {
+    /// The key is definitely revoked.
+    ///
+    /// The relevant self-revocations are returned.
+    Revoked(Vec<&'a crate::packet::Signature>),
+    /// There is a revocation certificate from a possible designated
+    /// revoker.
+    CouldBe(Vec<&'a crate::packet::Signature>),
+    /// The key does not appear to be revoked.
+    ///
+    /// An attacker could still have performed a DoS, which prevents
+    /// us from seeing the revocation certificate.
+    NotAsFarAsWeKnow,
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
