@@ -53,8 +53,6 @@
 extern crate capnp;
 #[macro_use]
 extern crate capnp_rpc;
-#[macro_use]
-extern crate failure;
 extern crate futures;
 extern crate rand;
 extern crate rusqlite;
@@ -125,11 +123,7 @@ impl Store {
         let core = Core::new()?;
         let handle = core.handle();
 
-        let mut rpc_system
-            = match descriptor.connect(&handle) {
-                Ok(r) => r,
-                Err(e) => return Err(e.into()),
-            };
+        let mut rpc_system = descriptor.connect(&handle)?;
 
         let client: node::Client = rpc_system.bootstrap(Side::Server);
         handle.spawn(rpc_system.map_err(|_e| ()));
@@ -202,7 +196,7 @@ impl Store {
     pub fn lookup(c: &Context, fp: &Fingerprint) -> Result<Key> {
         let (mut core, client) = Self::connect(c)?;
         let mut request = client.lookup_by_fingerprint_request();
-        let fp = fp.to_hex();
+        let fp = format!("{:X}", fp);
         request.get().set_fingerprint(&fp);
         let key = make_request!(&mut core, request)?;
         Ok(Key::new(Rc::new(RefCell::new(core)), key))
@@ -387,7 +381,7 @@ impl Mapping {
     pub fn add(&self, label: &str, fingerprint: &Fingerprint) -> Result<Binding> {
         let mut request = self.mapping.add_request();
         request.get().set_label(label);
-        request.get().set_fingerprint(fingerprint.to_hex().as_ref());
+        request.get().set_fingerprint(format!("{:X}", fingerprint).as_ref());
         let binding = make_request!(self.core.borrow_mut(), request)?;
         Ok(Binding::new(self.core.clone(), Some(label), binding))
     }
@@ -421,7 +415,7 @@ impl Mapping {
         let fingerprint = cert.fingerprint();
         let mut request = self.mapping.add_request();
         request.get().set_label(label);
-        request.get().set_fingerprint(fingerprint.to_hex().as_ref());
+        request.get().set_fingerprint(format!("{:X}", fingerprint).as_ref());
         let binding = make_request!(self.core.borrow_mut(), request)?;
         let binding = Binding::new(self.core.clone(), Some(label), binding);
         binding.import(cert)
@@ -893,7 +887,7 @@ impl Key {
     /// # let new = Cert::from_bytes(
     /// #     &include_bytes!("../../openpgp/tests/data/keys/testy-new.pgp")[..]).unwrap();
     /// let mapping = Mapping::open(&ctx, REALM_CONTACTS, "default")?;
-    /// let fp = Fingerprint::from_hex("3E8877C877274692975189F5D03F6F865226FE8B").unwrap();
+    /// let fp = "3E8877C877274692975189F5D03F6F865226FE8B".parse().unwrap();
     /// let binding = mapping.add("Testy McTestface", &fp)?;
     /// let key = binding.key()?;
     /// let r = key.import(&old)?;
@@ -1087,7 +1081,7 @@ impl Iterator for BundleIter {
                 self.core.borrow_mut(), request,
                 |r: node::binding_iter::item::Reader|
                 Ok((String::from(r.get_label()?),
-                    openpgp::Fingerprint::from_hex(r.get_fingerprint()?).unwrap(),
+                    r.get_fingerprint().unwrap().parse()?,
                     Binding::new(self.core.clone(), Some(r.get_label()?),
                                  r.get_binding()?))))
         };
@@ -1110,7 +1104,7 @@ impl Iterator for KeyIter {
             make_request_map!(
                 self.core.borrow_mut(), request,
                 |r: node::key_iter::item::Reader|
-                Ok((openpgp::Fingerprint::from_hex(r.get_fingerprint()?).unwrap(),
+                Ok((r.get_fingerprint().unwrap().parse()?,
                     Key::new(self.core.clone(), r.get_key()?))))
         };
         doit().ok()
@@ -1154,11 +1148,11 @@ impl Iterator for LogIter {
 /* Error handling.  */
 
 /// Results for sequoia-store.
-pub type Result<T> = ::std::result::Result<T, failure::Error>;
+pub type Result<T> = ::std::result::Result<T, anyhow::Error>;
 
 
 // Converts from backend errors.
-impl From<node::Error> for failure::Error {
+impl From<node::Error> for anyhow::Error {
     fn from(error: node::Error) -> Self {
         match error {
             node::Error::Unspecified => Error::StoreError.into(),
@@ -1180,37 +1174,31 @@ impl From<node::Error> for failure::Error {
     }
 }
 
-#[derive(Fail, Debug)]
+#[derive(thiserror::Error, Debug)]
 /// Errors returned from the store.
 pub enum Error {
     /// A requested key was not found.
-    #[fail(display = "Key not found")]
+    #[error("Key not found")]
     NotFound,
     /// The new key is in conflict with the current key.
-    #[fail(display = "New key conflicts with the current key")]
+    #[error("New key conflicts with the current key")]
     Conflict,
     /// This is a catch-all for unspecified backend errors, and should
     /// go away soon.
-    #[fail(display = "Unspecified store error")]
+    #[error("Unspecified store error")]
     StoreError,
     /// A protocol error occurred.
-    #[fail(display = "Unspecified protocol error")]
+    #[error("Unspecified protocol error")]
     ProtocolError,
     /// A Cert is malformed.
-    #[fail(display = "Malformed Cert")]
+    #[error("Malformed Cert")]
     MalformedCert,
     /// A fingerprint is malformed.
-    #[fail(display = "Malformed fingerprint")]
+    #[error("Malformed fingerprint")]
     MalformedFingerprint,
     /// A `capnp::Error` occurred.
-    #[fail(display = "Internal RPC error")]
-    RpcError(capnp::Error),
-}
-
-impl From<capnp::Error> for Error {
-    fn from(error: capnp::Error) -> Self {
-        Error::RpcError(error)
-    }
+    #[error("Internal RPC error")]
+    RpcError(#[from] capnp::Error),
 }
 
 impl From<capnp::NotInSchema> for Error {
