@@ -3,7 +3,7 @@ use std::io::{self, Read};
 extern crate sequoia_openpgp as openpgp;
 use self::openpgp::types::{Duration, Timestamp, SymmetricAlgorithm};
 use self::openpgp::fmt::hex;
-use self::openpgp::crypto::mpis;
+use self::openpgp::crypto::mpi;
 use self::openpgp::{Packet, Result};
 use self::openpgp::packet::prelude::*;
 use self::openpgp::packet::header::CTB;
@@ -60,7 +60,7 @@ pub fn dump<W>(input: &mut dyn io::Read, output: &mut dyn io::Write,
 {
     let mut ppr
         = self::openpgp::parse::PacketParserBuilder::from_reader(input)?
-        .map(hex).finalize()?;
+        .map(hex).build()?;
     let mut message_encrypted = false;
     let width = width.into().unwrap_or(80);
     let mut dumper = PacketDumper::new(width, mpis);
@@ -124,10 +124,10 @@ pub fn dump<W>(input: &mut dyn io::Read, output: &mut dyn io::Write,
 
                 let mut fields = Vec::new();
                 fields.push(format!("Session key: {}", hex::encode(sk)));
-                if pp.decrypted() {
-                    fields.push("Decryption successful".into());
-                } else {
+                if pp.encrypted() {
                     fields.push("Decryption failed".into());
+                } else {
+                    fields.push("Decryption successful".into());
                 }
                 Some(fields)
             },
@@ -137,9 +137,9 @@ pub fn dump<W>(input: &mut dyn io::Read, output: &mut dyn io::Write,
         let header = pp.header().clone();
         let map = pp.take_map();
 
+        let recursion_depth = pp.recursion_depth();
         let (packet, ppr_) = pp.recurse()?;
         ppr = ppr_;
-        let recursion_depth = ppr.last_recursion_depth().unwrap();
 
         dumper.packet(output, recursion_depth as usize,
                       header, packet, map, additional_fields)?;
@@ -307,35 +307,35 @@ impl PacketDumper {
 
                 let ii = format!("{}    ", i);
                 match k.mpis() {
-                    mpis::PublicKey::RSA { e, n } =>
+                    mpi::PublicKey::RSA { e, n } =>
                         pd.dump_mpis(output, &ii,
                                        &[e.value(), n.value()],
                                        &["e", "n"])?,
-                    mpis::PublicKey::DSA { p, q, g, y } =>
+                    mpi::PublicKey::DSA { p, q, g, y } =>
                         pd.dump_mpis(output, &ii,
                                        &[p.value(), q.value(), g.value(),
                                          y.value()],
                                        &["p", "q", "g", "y"])?,
-                    mpis::PublicKey::Elgamal { p, g, y } =>
+                    mpi::PublicKey::ElGamal { p, g, y } =>
                         pd.dump_mpis(output, &ii,
                                        &[p.value(), g.value(), y.value()],
                                        &["p", "g", "y"])?,
-                    mpis::PublicKey::EdDSA { curve, q } => {
+                    mpi::PublicKey::EdDSA { curve, q } => {
                         writeln!(output, "{}  Curve: {}", ii, curve)?;
                         pd.dump_mpis(output, &ii, &[q.value()], &["q"])?;
                     },
-                    mpis::PublicKey::ECDSA { curve, q } => {
+                    mpi::PublicKey::ECDSA { curve, q } => {
                         writeln!(output, "{}  Curve: {}", ii, curve)?;
                         pd.dump_mpis(output, &ii, &[q.value()], &["q"])?;
                     },
-                    mpis::PublicKey::ECDH { curve, q, hash, sym } => {
+                    mpi::PublicKey::ECDH { curve, q, hash, sym } => {
                         writeln!(output, "{}  Curve: {}", ii, curve)?;
                         writeln!(output, "{}  Hash algo: {}", ii, hash)?;
                         writeln!(output, "{}  Symmetric algo: {}", ii,
                                  sym)?;
                         pd.dump_mpis(output, &ii, &[q.value()], &["q"])?;
                     },
-                    mpis::PublicKey::Unknown { mpis, rest } => {
+                    mpi::PublicKey::Unknown { mpis, rest } => {
                         let keys: Vec<String> =
                             (0..mpis.len()).map(
                                 |i| format!("mpi{}", i)).collect();
@@ -350,9 +350,10 @@ impl PacketDumper {
 
                         pd.dump_mpis(output, &ii, &[&rest[..]], &["rest"])?;
                     },
+                    mpi::PublicKey::__Nonexhaustive => unreachable!(),
                 }
 
-                if let Some(secrets) = k.secret() {
+                if let Some(secrets) = k.optional_secret() {
                     use self::openpgp::packet::key::SecretKeyMaterial;
                     writeln!(output, "{}", i)?;
                     writeln!(output, "{}  Secret Key:", i)?;
@@ -363,30 +364,30 @@ impl PacketDumper {
                             |mpis| -> Result<()> {
                                 match mpis
                                 {
-                                    mpis::SecretKeyMaterial::RSA { d, p, q, u } =>
+                                    mpi::SecretKeyMaterial::RSA { d, p, q, u } =>
                                         pd.dump_mpis(output, &ii,
                                                        &[d.value(), p.value(),
                                                          q.value(), u.value()],
                                                        &["d", "p", "q", "u"])?,
-                                    mpis::SecretKeyMaterial::DSA { x } =>
+                                    mpi::SecretKeyMaterial::DSA { x } =>
                                         pd.dump_mpis(output, &ii, &[x.value()],
                                                        &["x"])?,
-                                    mpis::SecretKeyMaterial::Elgamal { x } =>
+                                    mpi::SecretKeyMaterial::ElGamal { x } =>
                                         pd.dump_mpis(output, &ii, &[x.value()],
                                                        &["x"])?,
-                                    mpis::SecretKeyMaterial::EdDSA { scalar } =>
+                                    mpi::SecretKeyMaterial::EdDSA { scalar } =>
                                         pd.dump_mpis(output, &ii,
                                                        &[scalar.value()],
                                                        &["scalar"])?,
-                                    mpis::SecretKeyMaterial::ECDSA { scalar } =>
+                                    mpi::SecretKeyMaterial::ECDSA { scalar } =>
                                         pd.dump_mpis(output, &ii,
                                                        &[scalar.value()],
                                                        &["scalar"])?,
-                                    mpis::SecretKeyMaterial::ECDH { scalar } =>
+                                    mpi::SecretKeyMaterial::ECDH { scalar } =>
                                         pd.dump_mpis(output, &ii,
                                                        &[scalar.value()],
                                                        &["scalar"])?,
-                                    mpis::SecretKeyMaterial::Unknown { mpis, rest } => {
+                                    mpi::SecretKeyMaterial::Unknown { mpis, rest } => {
                                         let keys: Vec<String> =
                                             (0..mpis.len()).map(
                                                 |i| format!("mpi{}", i)).collect();
@@ -402,6 +403,8 @@ impl PacketDumper {
                                         pd.dump_mpis(output, &ii, &[rest],
                                                        &["rest"])?;
                                     },
+                                    mpi::SecretKeyMaterial::__Nonexhaustive =>
+                                        unreachable!(),
                                 } Ok(()) })?,
                         SecretKeyMaterial::Encrypted(ref e) => {
                             writeln!(output, "{}", i)?;
@@ -463,27 +466,27 @@ impl PacketDumper {
 
                     let ii = format!("{}    ", i);
                     match s.mpis() {
-                        mpis::Signature::RSA { s } =>
+                        mpi::Signature::RSA { s } =>
                             self.dump_mpis(output, &ii,
                                            &[s.value()],
                                            &["s"])?,
-                        mpis::Signature::DSA { r, s } =>
+                        mpi::Signature::DSA { r, s } =>
                             self.dump_mpis(output, &ii,
                                            &[r.value(), s.value()],
                                            &["r", "s"])?,
-                        mpis::Signature::Elgamal { r, s } =>
+                        mpi::Signature::ElGamal { r, s } =>
                             self.dump_mpis(output, &ii,
                                            &[r.value(), s.value()],
                                            &["r", "s"])?,
-                        mpis::Signature::EdDSA { r, s } =>
+                        mpi::Signature::EdDSA { r, s } =>
                             self.dump_mpis(output, &ii,
                                            &[r.value(), s.value()],
                                            &["r", "s"])?,
-                        mpis::Signature::ECDSA { r, s } =>
+                        mpi::Signature::ECDSA { r, s } =>
                             self.dump_mpis(output, &ii,
                                            &[r.value(), s.value()],
                                            &["r", "s"])?,
-                        mpis::Signature::Unknown { mpis, rest } => {
+                        mpi::Signature::Unknown { mpis, rest } => {
                             let keys: Vec<String> =
                                 (0..mpis.len()).map(
                                     |i| format!("mpi{}", i)).collect();
@@ -498,6 +501,8 @@ impl PacketDumper {
 
                             self.dump_mpis(output, &ii, &[&rest[..]], &["rest"])?;
                         },
+                        mpi::Signature::__Nonexhaustive => unreachable!(),
+
                     }
                 }
             },
@@ -579,19 +584,19 @@ impl PacketDumper {
 
                     let ii = format!("{}    ", i);
                     match p.esk() {
-                        mpis::Ciphertext::RSA { c } =>
+                        mpi::Ciphertext::RSA { c } =>
                             self.dump_mpis(output, &ii,
                                            &[c.value()],
                                            &["c"])?,
-                        mpis::Ciphertext::Elgamal { e, c } =>
+                        mpi::Ciphertext::ElGamal { e, c } =>
                             self.dump_mpis(output, &ii,
                                            &[e.value(), c.value()],
                                            &["e", "c"])?,
-                        mpis::Ciphertext::ECDH { e, key } =>
+                        mpi::Ciphertext::ECDH { e, key } =>
                             self.dump_mpis(output, &ii,
                                            &[e.value(), key],
                                            &["e", "key"])?,
-                        mpis::Ciphertext::Unknown { mpis, rest } => {
+                        mpi::Ciphertext::Unknown { mpis, rest } => {
                             let keys: Vec<String> =
                                 (0..mpis.len()).map(
                                     |i| format!("mpi{}", i)).collect();
@@ -606,6 +611,7 @@ impl PacketDumper {
 
                             self.dump_mpis(output, &ii, &[rest], &["rest"])?;
                         },
+                        mpi::Ciphertext::__Nonexhaustive => unreachable!(),
                     }
                 }
             },
@@ -684,24 +690,9 @@ impl PacketDumper {
 
             for field in map.iter() {
                 if field.name() == "body" {
-                    hd.write_labeled(field.data(), |offset, data| {
-                        let mut l = String::new();
-                        for _ in 0..offset {
-                            l.push(' ');
-                        }
-                        for &c in data {
-                            l.push(if c < 32 {
-                                '.'
-                            } else if c < 128 {
-                                c.into()
-                            } else {
-                                '.'
-                            })
-                        }
-                        Some(l)
-                    })?;
+                    hd.write_ascii(field.as_bytes())?;
                 } else {
-                    hd.write(field.data(), field.name())?;
+                    hd.write(field.as_bytes(), field.name())?;
                 }
             }
 
@@ -760,10 +751,15 @@ impl PacketDumper {
                 write!(output, "{}    Symmetric algo preferences: {}", i,
                        c.iter().map(|c| format!("{:?}", c))
                        .collect::<Vec<String>>().join(", "))?,
-            RevocationKey{class, pk_algo, ref fp} =>
+            RevocationKey(rk) => {
+                let (pk_algo, fp) = rk.revoker();
                 write!(output,
-                       "{}    Revocation key: class {} algo {} fingerprint {}", i,
-                       class, pk_algo, fp)?,
+                       "{}    Revocation key: {}/{}", i,
+                       fp, pk_algo)?;
+                if rk.sensitive() {
+                    write!(output, ", sensitive")?;
+                }
+            },
             Issuer(ref is) =>
                 write!(output, "{}    Issuer: {}", i, is)?,
             NotationData(ref n) =>
