@@ -1,4 +1,4 @@
-use failure::{self, ResultExt};
+use anyhow::Context as _;
 use std::cmp::Ordering;
 use std::collections::{HashMap, HashSet};
 use std::fs::File;
@@ -11,6 +11,7 @@ use sequoia_core::Context;
 use crate::openpgp::types::{
     CompressionAlgorithm,
 };
+use crate::openpgp::cert::prelude::*;
 use crate::openpgp::crypto;
 use crate::openpgp::{Cert, KeyID, Result};
 use crate::openpgp::packet::prelude::*;
@@ -18,17 +19,14 @@ use crate::openpgp::parse::{
     Parse,
     PacketParserResult,
 };
-use crate::openpgp::parse::stream::{
-    Verifier, DetachedVerifier, VerificationResult, VerificationHelper,
-    MessageStructure, MessageLayer,
-};
+use crate::openpgp::parse::stream::*;
 use crate::openpgp::serialize::stream::{
     Message, Signer, LiteralWriter, Encryptor, Recipient,
     Compressor,
-};
-use crate::openpgp::serialize::padding::{
-    Padder,
-    padme,
+    padding::{
+        Padder,
+        padme,
+    },
 };
 use crate::openpgp::policy::Policy;
 extern crate sequoia_store as store;
@@ -55,7 +53,7 @@ fn get_signing_keys(certs: &[openpgp::Cert], p: &dyn Policy,
             .for_signing()
             .map(|ka| ka.key())
         {
-            if let Some(secret) = key.secret() {
+            if let Some(secret) = key.optional_secret() {
                 let unencrypted = match secret {
                     SecretKeyMaterial::Encrypted(ref e) => {
                         let password = rpassword::read_password_from_tty(Some(
@@ -73,24 +71,20 @@ fn get_signing_keys(certs: &[openpgp::Cert], p: &dyn Policy,
             }
         }
 
-        return Err(failure::err_msg(
+        return Err(anyhow::anyhow!(
             format!("Found no suitable signing key on {}", tsk)));
     }
 
     Ok(keys)
 }
 
-pub fn encrypt(policy: &dyn Policy,
-               mapping: &mut store::Mapping,
-               input: &mut dyn io::Read, output: &mut dyn io::Write,
-               npasswords: usize, recipients: Vec<&str>,
-               mut certs: Vec<openpgp::Cert>, signers: Vec<openpgp::Cert>,
-               mode: openpgp::types::KeyFlags, compression: &str,
-               time: Option<SystemTime>)
-               -> Result<()> {
-    for r in recipients {
-        certs.push(mapping.lookup(r).context("No such key found")?.cert()?);
-    }
+pub fn encrypt<'a>(policy: &'a dyn Policy,
+                   input: &mut dyn io::Read, message: Message<'a>,
+                   npasswords: usize, recipients: &'a [openpgp::Cert],
+                   signers: Vec<openpgp::Cert>,
+                   mode: openpgp::types::KeyFlags, compression: &str,
+                   time: Option<SystemTime>)
+                   -> Result<()> {
     let mut passwords: Vec<crypto::Password> = Vec::with_capacity(npasswords);
     for n in 0..npasswords {
         let nprompt = format!("Enter password {}: ", n + 1);
@@ -102,19 +96,16 @@ pub fn encrypt(policy: &dyn Policy,
             }))?.into());
     }
 
-    if certs.len() + passwords.len() == 0 {
-        return Err(failure::format_err!(
+    if recipients.len() + passwords.len() == 0 {
+        return Err(anyhow::anyhow!(
             "Neither recipient nor password given"));
     }
 
     let mut signers = get_signing_keys(&signers, policy, time)?;
 
-    // Build a vector of references to hand to Signer.
-    let recipients: Vec<&openpgp::Cert> = certs.iter().collect();
-
     // Build a vector of recipients to hand to Encryptor.
     let mut recipient_subkeys: Vec<Recipient> = Vec::new();
-    for cert in certs.iter() {
+    for cert in recipients.iter() {
         let mut count = 0;
         for key in cert.keys().with_policy(policy, None).alive().revoked(false)
             .key_flags(&mode).map(|ka| ka.key())
@@ -123,26 +114,15 @@ pub fn encrypt(policy: &dyn Policy,
             count += 1;
         }
         if count == 0 {
-            return Err(failure::format_err!(
+            return Err(anyhow::anyhow!(
                 "Key {} has no suitable encryption key", cert));
         }
     }
 
-    // Stream an OpenPGP message.
-    let message = Message::new(output);
-
     // We want to encrypt a literal data packet.
-    let mut encryptor = if let Some(p) = passwords.pop() {
-        Encryptor::with_password(message, p)
-    } else {
-        Encryptor::for_recipient(message, recipient_subkeys.pop().unwrap())
-    };
-    for p in passwords {
-        encryptor = encryptor.add_password(p);
-    }
-    for r in recipient_subkeys {
-        encryptor = encryptor.add_recipient(r);
-    }
+    let encryptor =
+        Encryptor::for_recipients(message, recipient_subkeys)
+        .add_passwords(passwords);
 
     let mut sink = encryptor.build()
         .context("Failed to create encryptor")?;
@@ -168,7 +148,7 @@ pub fn encrypt(policy: &dyn Policy,
                 signer = signer.creation_time(time);
             }
         }
-        for r in recipients {
+        for r in recipients.iter() {
             signer = signer.add_intended_recipient(r);
         }
         sink = signer.build()?;
@@ -246,17 +226,17 @@ impl<'a> VHelper<'a> {
     }
 
     fn print_sigs(&mut self, results: &[VerificationResult]) {
-        use self::VerificationResult::*;
+        use self::VerificationError::*;
         for result in results {
             let (issuer, level) = match result {
-                GoodChecksum { sig, ka, .. } =>
+                Ok(GoodChecksum { sig, ka, .. }) =>
                     (ka.key().keyid(), sig.level()),
-                NotAlive { sig, .. } =>
-                    (sig.get_issuers().get(0)
-                         .map(|i| i.into())
-                         .unwrap_or(KeyID::wildcard()),
-                     sig.level()),
-                MissingKey { sig } => {
+                Err(MalformedSignature { error, .. }) => {
+                    eprintln!("Malformed signature: {}", error);
+                    self.broken_signatures += 1;
+                    continue;
+                },
+                Err(MissingKey { sig, .. }) => {
                     let issuer = sig.get_issuers().get(0)
                         .expect("missing key checksum has an issuer")
                         .to_string();
@@ -267,18 +247,28 @@ impl<'a> VHelper<'a> {
                     eprintln!("No key to check {} from {}", what, issuer);
                     self.unknown_checksums += 1;
                     continue;
-                }
-                Error { sig, error } => {
-                    let issuer = sig.get_issuers().get(0)
-                        .expect("key has an issuer")
-                        .to_string();
+                },
+                Err(UnboundKey { cert, error, .. }) => {
+                    eprintln!("Signing key on {} is not bound: {}",
+                              cert.fingerprint(), error);
+                    self.bad_checksums += 1;
+                    continue;
+                },
+                Err(BadKey { ka, error, .. }) => {
+                    eprintln!("Signing key on {} is bad: {}",
+                              ka.cert().fingerprint(), error);
+                    self.bad_checksums += 1;
+                    continue;
+                },
+                Err(BadSignature { sig, ka, error }) => {
+                    let issuer = ka.fingerprint().to_string();
                     let what = match sig.level() {
                         0 => "checksum".into(),
                         n => format!("level {} notarizing checksum", n),
                     };
                     eprintln!("Error verifying {} from {}: {}",
                               what, issuer, error);
-                    self.broken_signatures += 1;
+                    self.bad_checksums += 1;
                     continue;
                 }
             };
@@ -294,33 +284,18 @@ impl<'a> VHelper<'a> {
 
             let issuer_str = issuer.to_string();
             let label = self.labels.get(&issuer).unwrap_or(&issuer_str);
-            match result {
-                GoodChecksum { .. } => {
-                    eprintln!("Good {} from {}", what, label);
-                    if trusted {
-                        self.good_signatures += 1;
-                    } else {
-                        self.good_checksums += 1;
-                    }
-                },
-                NotAlive { .. } => {
-                    eprintln!("Good checksum, but not alive: {} from {}",
-                              what, label);
-                    if trusted {
-                        self.bad_signatures += 1;
-                    } else {
-                        self.bad_checksums += 1;
-                    }
-                },
-                MissingKey { .. } => unreachable!("handled above"),
-                Error { .. } => unreachable!("handled above"),
+            eprintln!("Good {} from {}", what, label);
+            if trusted {
+                self.good_signatures += 1;
+            } else {
+                self.good_checksums += 1;
             }
         }
     }
 }
 
 impl<'a> VerificationHelper for VHelper<'a> {
-    fn get_public_keys(&mut self, ids: &[openpgp::KeyHandle]) -> Result<Vec<Cert>> {
+    fn get_certs(&mut self, ids: &[openpgp::KeyHandle]) -> Result<Vec<Cert>> {
         let mut certs = self.certs.take().unwrap();
         // Get all keys.
         let seen: HashSet<_> = certs.iter()
@@ -373,7 +348,7 @@ impl<'a> VerificationHelper for VHelper<'a> {
     }
 
     fn check(&mut self, structure: MessageStructure) -> Result<()> {
-        for layer in structure.iter() {
+        for layer in structure {
             match layer {
                 MessageLayer::Compression { algo } =>
                     eprintln!("Compressed using {}", algo),
@@ -394,7 +369,7 @@ impl<'a> VerificationHelper for VHelper<'a> {
             Ok(())
         } else {
             self.print_status();
-            Err(failure::err_msg("Verification failed"))
+            Err(anyhow::anyhow!("Verification failed"))
         }
     }
 }
@@ -407,22 +382,19 @@ pub fn verify(ctx: &Context, policy: &dyn Policy,
               signatures: usize, certs: Vec<Cert>)
               -> Result<()> {
     let helper = VHelper::new(ctx, mapping, signatures, certs);
-    let mut verifier = if let Some(dsig) = detached {
-        DetachedVerifier::from_reader(policy, dsig, input, helper, None)?
+    let helper = if let Some(dsig) = detached {
+        let mut v = DetachedVerifierBuilder::from_reader(dsig)?
+            .with_policy(policy, None, helper)?;
+        v.verify_reader(input)?;
+        v.into_helper()
     } else {
-        Verifier::from_reader(policy, input, helper, None)?
+        let mut v = VerifierBuilder::from_reader(input)?
+            .with_policy(policy, None, helper)?;
+        io::copy(&mut v, output)?;
+        v.into_helper()
     };
 
-    io::copy(&mut verifier, output)
-        .map_err(|e| if e.get_ref().is_some() {
-            // Wrapped failure::Error.  Recover it.
-            failure::Error::from_boxed_compat(e.into_inner().unwrap())
-        } else {
-            // Plain io::Error.
-            e.into()
-        })?;
-
-    verifier.into_helper().print_status();
+    helper.print_status();
     Ok(())
 }
 
@@ -432,7 +404,7 @@ pub fn split(input: &mut dyn io::Read, prefix: &str)
     // nested packets.
     let mut ppr =
         openpgp::parse::PacketParserBuilder::from_reader(input)?
-        .map(true).finalize()?;
+        .map(true).build()?;
 
     // This encodes our position in the tree.
     let mut pos = vec![0];
@@ -450,13 +422,13 @@ pub fn split(input: &mut dyn io::Read, prefix: &str)
 
             // Write all the bytes.
             for field in map.iter() {
-                sink.write_all(field.data())?;
+                sink.write_all(field.as_bytes())?;
             }
         }
 
+        let old_depth = Some(pp.recursion_depth());
         ppr = pp.recurse()?.1;
-        let old_depth = ppr.last_recursion_depth();
-        let new_depth = ppr.recursion_depth();
+        let new_depth = ppr.as_ref().map(|pp| pp.recursion_depth());
 
         // Update pos.
         match old_depth.cmp(&new_depth) {
@@ -483,7 +455,7 @@ pub fn join(inputs: Option<clap::Values>, output: &mut dyn io::Write)
             // We (ab)use the mapping feature to create byte-accurate
             // copies.
             for field in pp.map().expect("must be mapped").iter() {
-                output.write_all(field.data())?;
+                output.write_all(field.as_bytes())?;
             }
 
             ppr = pp.next()?.1;
@@ -495,13 +467,13 @@ pub fn join(inputs: Option<clap::Values>, output: &mut dyn io::Write)
         for name in inputs {
             let ppr =
                 openpgp::parse::PacketParserBuilder::from_file(name)?
-                .map(true).finalize()?;
+                .map(true).build()?;
             copy(ppr, output)?;
         }
     } else {
         let ppr =
             openpgp::parse::PacketParserBuilder::from_reader(io::stdin())?
-            .map(true).finalize()?;
+            .map(true).build()?;
         copy(ppr, output)?;
     }
     Ok(())
