@@ -2,8 +2,6 @@
 
 extern crate clap;
 #[macro_use]
-extern crate failure;
-#[macro_use]
 extern crate prettytable;
 extern crate rpassword;
 extern crate tempfile;
@@ -12,7 +10,7 @@ extern crate itertools;
 extern crate tokio_core;
 
 use crossterm::terminal;
-use failure::ResultExt;
+use anyhow::Context as _;
 use prettytable::{Table, Cell, Row};
 use std::fs::{File, OpenOptions};
 use std::io::{self, Write};
@@ -25,13 +23,14 @@ extern crate sequoia_core;
 extern crate sequoia_net;
 extern crate sequoia_store as store;
 
-use crate::openpgp::{armor, Fingerprint, Cert};
+use crate::openpgp::Result;
+use crate::openpgp::{armor, Cert};
 use sequoia_autocrypt as autocrypt;
 use crate::openpgp::fmt::hex;
 use crate::openpgp::types::KeyFlags;
 use crate::openpgp::parse::Parse;
-use crate::openpgp::serialize::Serialize;
-use crate::openpgp::cert::CertParser;
+use crate::openpgp::serialize::{Serialize, stream::{Message, Armorer}};
+use crate::openpgp::cert::prelude::*;
 use crate::openpgp::policy::StandardPolicy as P;
 use sequoia_core::{Context, NetworkPolicy};
 use sequoia_net::{KeyServer, wkd};
@@ -41,7 +40,7 @@ mod sq_cli;
 mod commands;
 use commands::dump::Convert;
 
-fn open_or_stdin(f: Option<&str>) -> Result<Box<dyn io::Read>, failure::Error> {
+fn open_or_stdin(f: Option<&str>) -> Result<Box<dyn io::Read>> {
     match f {
         Some(f) => Ok(Box::new(File::open(f)
                                .context("Failed to open input file")?)),
@@ -50,7 +49,7 @@ fn open_or_stdin(f: Option<&str>) -> Result<Box<dyn io::Read>, failure::Error> {
 }
 
 fn create_or_stdout(f: Option<&str>, force: bool)
-    -> Result<Box<dyn io::Write>, failure::Error> {
+    -> Result<Box<dyn io::Write>> {
     match f {
         None => Ok(Box::new(io::stdout())),
         Some(p) if p == "-" => Ok(Box::new(io::stdout())),
@@ -64,82 +63,23 @@ fn create_or_stdout(f: Option<&str>, force: bool)
                             .open(f)
                             .context("Failed to create output file")?))
             } else {
-                Err(failure::err_msg(
+                Err(anyhow::anyhow!(
                     format!("File {:?} exists, use --force to overwrite", p)))
             }
         }
     }
 }
 
-// XXX: This is a candidate for inclusion in the library.
-enum Writer<T: Write> {
-    Binary {
-        inner: T,
-    },
-    Armored {
-        inner: openpgp::armor::Writer<T>,
-    },
-}
-
-impl<T: Write> From<T> for Writer<T> {
-    fn from(inner: T) -> Self {
-        Writer::Binary { inner }
-    }
-}
-
-impl<T: Write> From<openpgp::armor::Writer<T>> for Writer<T> {
-    fn from(inner: openpgp::armor::Writer<T>) -> Self {
-        Writer::Armored { inner }
-    }
-}
-
-impl<T: Write> Writer<T> {
-    pub fn armor(self, kind: openpgp::armor::Kind, headers: &[(&str, &str)])
-                 -> openpgp::Result<Self>
-    {
-        match self {
-            Writer::Binary { inner } =>
-                Ok(openpgp::armor::Writer::new(inner, kind, headers)?
-                   .into()),
-            Writer::Armored { .. } =>
-                Err(openpgp::Error::InvalidOperation("already armored".into())
-                    .into()),
-        }
-    }
-
-    pub fn finalize(self) -> std::io::Result<T> {
-        match self {
-            Writer::Binary { inner } => Ok(inner),
-            Writer::Armored { inner } => inner.finalize(),
-        }
-    }
-}
-
-impl<T: Write> Write for Writer<T> {
-    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
-        match self {
-            Writer::Binary { inner } => inner.write(buf),
-            Writer::Armored { inner } => inner.write(buf),
-        }
-    }
-    fn flush(&mut self) -> std::io::Result<()> {
-        match self {
-            Writer::Binary { inner } => inner.flush(),
-            Writer::Armored { inner } => inner.flush(),
-        }
-    }
-}
-
-fn create_or_stdout_pgp(f: Option<&str>, force: bool,
-                        binary: bool, kind: armor::Kind)
-    -> Result<Writer<Box<dyn Write>>, failure::Error>
+fn create_or_stdout_pgp<'a>(f: Option<&str>, force: bool,
+                            binary: bool, kind: armor::Kind)
+                            -> Result<Message<'a>>
 {
     let sink = create_or_stdout(f, force)?;
-    let mut sink = Writer::from(sink);
+    let mut message = Message::new(sink);
     if ! binary {
-        sink = sink.armor(kind, &[])?;
+        message = Armorer::new(message).kind(kind).build()?;
     }
-    Ok(sink)
+    Ok(message)
 }
 
 fn load_certs<'a, I>(files: I) -> openpgp::Result<Vec<Cert>>
@@ -179,9 +119,9 @@ fn serialize_keyring(mut output: &mut dyn io::Write, certs: &[Cert], binary: boo
     let headers: Vec<_> = headers.iter()
         .map(|value| ("Comment", value.as_str()))
         .collect();
-    let mut output = armor::Writer::new(&mut output,
-                                        armor::Kind::PublicKey,
-                                        &headers)?;
+    let mut output = armor::Writer::with_headers(&mut output,
+                                                 armor::Kind::PublicKey,
+                                                 headers)?;
     for cert in certs {
         cert.serialize(&mut output)?;
     }
@@ -212,11 +152,16 @@ fn help_warning(arg: &str) {
     }
 }
 
-fn real_main() -> Result<(), failure::Error> {
-    let policy = &P::new();
+fn main() -> Result<()> {
+    let policy = &mut P::new();
 
     let matches = sq_cli::build().get_matches();
 
+    let known_notations: Vec<&str> = matches.values_of("known-notation")
+        .unwrap_or_default()
+        .collect();
+    policy.good_critical_notations(&known_notations);
+
     let network_policy = match matches.value_of("policy") {
         None => NetworkPolicy::Encrypted,
         Some("offline") => NetworkPolicy::Offline,
@@ -266,19 +211,22 @@ fn real_main() -> Result<(), failure::Error> {
                               m.is_present("dump"), m.is_present("hex"))?;
         },
         ("encrypt",  Some(m)) => {
+            let mapping = Mapping::open(&ctx, realm_name, mapping_name)
+                .context("Failed to open the mapping")?;
+            let mut recipients = m.values_of("recipient-key-file")
+                .map(load_certs)
+                .unwrap_or(Ok(vec![]))?;
+            if let Some(r) = m.values_of("recipient") {
+                for recipient in r {
+                    recipients.push(mapping.lookup(recipient)
+                                    .context("No such key found")?.cert()?);
+                }
+            }
             let mut input = open_or_stdin(m.value_of("input"))?;
-            let mut output =
+            let output =
                 create_or_stdout_pgp(m.value_of("output"), force,
                                      m.is_present("binary"),
                                      armor::Kind::Message)?;
-            let mut mapping = Mapping::open(&ctx, realm_name, mapping_name)
-                .context("Failed to open the mapping")?;
-            let recipients = m.values_of("recipient")
-                .map(|r| r.collect())
-                .unwrap_or(vec![]);
-            let additional_certs = m.values_of("recipient-key-file")
-                .map(load_certs)
-                .unwrap_or(Ok(vec![]))?;
             let additional_secrets = m.values_of("signer-key-file")
                 .map(load_certs)
                 .unwrap_or(Ok(vec![]))?;
@@ -299,13 +247,12 @@ fn real_main() -> Result<(), failure::Error> {
             } else {
                 None
             };
-            commands::encrypt(policy, &mut mapping, &mut input, &mut output,
+            commands::encrypt(policy, &mut input, output,
                               m.occurrences_of("symmetric") as usize,
-                              recipients, additional_certs, additional_secrets,
+                              &recipients, additional_secrets,
                               mode,
                               m.value_of("compression").expect("has default"),
                               time.into())?;
-            output.finalize()?;
         },
         ("sign",  Some(m)) => {
             let mut input = open_or_stdin(m.value_of("input"))?;
@@ -385,13 +332,14 @@ fn real_main() -> Result<(), failure::Error> {
                     let cert = Cert::from_reader(input)?;
                     let addr = m.value_of("address").map(|a| a.to_string())
                         .or_else(|| {
-                            cert.primary_userid(policy, None)
+                            cert.with_policy(policy, None)
+                                .and_then(|vcert| vcert.primary_userid()).ok()
                                 .map(|ca| ca.userid().to_string())
                         });
                     let ac = autocrypt::AutocryptHeader::new_sender(
                         policy,
                         &cert,
-                        &addr.ok_or(failure::err_msg(
+                        &addr.ok_or(anyhow::anyhow!(
                             "No well-formed primary userid found, use \
                              --address to specify one"))?,
                         m.value_of("prefer-encrypt").expect("has default"))?;
@@ -480,7 +428,7 @@ fn real_main() -> Result<(), failure::Error> {
             match m.subcommand() {
                 ("get",  Some(m)) => {
                     let keyid = m.value_of("keyid").unwrap();
-                    let id = openpgp::KeyID::from_hex(keyid);
+                    let id = keyid.parse();
                     if id.is_err() {
                         eprintln!("Malformed key ID: {:?}\n\
                                    (Note: only long Key IDs are supported.)",
@@ -518,7 +466,7 @@ fn real_main() -> Result<(), failure::Error> {
                     list_bindings(&mapping, realm_name, mapping_name)?;
                 },
                 ("add",  Some(m)) => {
-                    let fp = Fingerprint::from_hex(m.value_of("fingerprint").unwrap())
+                    let fp = m.value_of("fingerprint").unwrap().parse()
                         .expect("Malformed fingerprint");
                     mapping.add(m.value_of("label").unwrap(), &fp)?;
                 },
@@ -681,7 +629,7 @@ fn real_main() -> Result<(), failure::Error> {
 }
 
 fn list_bindings(mapping: &Mapping, realm: &str, name: &str)
-                 -> Result<(), failure::Error> {
+                 -> Result<()> {
     if mapping.iter()?.count() == 0 {
         println!("No label-key bindings in the \"{}/{}\" mapping.",
                  realm, name);
@@ -725,7 +673,7 @@ fn print_log(iter: LogIter, with_slug: bool) {
 
 /// Parses the given string depicting a ISO 8601 timestamp.
 fn parse_iso8601(s: &str, pad_date_with: chrono::NaiveTime)
-                 -> failure::Fallible<DateTime<Utc>>
+                 -> Result<DateTime<Utc>>
 {
     // If you modify this function this function, synchronize the
     // changes with the copy in sqv.rs!
@@ -766,7 +714,7 @@ fn parse_iso8601(s: &str, pad_date_with: chrono::NaiveTime)
             return Ok(DateTime::from_utc(d.and_time(pad_date_with), Utc));
         }
     }
-    Err(failure::format_err!("Malformed ISO8601 timestamp: {}", s))
+    Err(anyhow::anyhow!("Malformed ISO8601 timestamp: {}", s))
 }
 
 #[test]
@@ -794,16 +742,3 @@ fn test_parse_iso8601() {
     parse_iso8601("2017031", z).unwrap();
     // parse_iso8601("2017", z).unwrap(); // ditto
 }
-
-fn main() {
-    if let Err(e) = real_main() {
-        let mut cause = e.as_fail();
-        eprint!("{}", cause);
-        while let Some(c) = cause.cause() {
-            eprint!(":\n  {}", c);
-            cause = c;
-        }
-        eprintln!();
-        exit(2);
-    }
-}
