@@ -27,6 +27,16 @@ pub fn build() -> App<'static, 'static> {
              .long("force")
              .short("f")
              .help("Overwrite existing files"))
+        .arg(Arg::with_name("known-notation")
+             .long("known-notation")
+             .multiple(true)
+             .takes_value(true)
+             .value_name("NOTATION")
+             .number_of_values(1)
+             .help("The notation name is considered known. \
+               This is used when validating sigantures. \
+               Signatures that have unknown notations with the \
+               critical bit set are considered invalid."))
         .subcommand(SubCommand::with_name("decrypt")
                     .display_order(10)
                     .about("Decrypts an OpenPGP message")
@@ -393,12 +403,21 @@ pub fn build() -> App<'static, 'static> {
                              .long("with-password")
                              .help("Prompt for a password to protect the \
                                     generated key with."))
-                        .arg(Arg::with_name("expiry")
-                             .value_name("EXPIRY")
-                             .long("expiry")
+
+                        .group(ArgGroup::with_name("expiration-group")
+                               .args(&["expires", "expires-in"]))
+
+                        .arg(Arg::with_name("expires")
+                             .value_name("TIME")
+                             .long("expires")
+                             .help("Absolute time When the key should expire, \
+                                    or 'never'."))
+                        .arg(Arg::with_name("expires-in")
+                             .value_name("DURATION")
+                             .long("expires-in")
                              // Catch negative numbers.
                              .allow_hyphen_values(true)
-                             .help("When the key should expire.  \
+                             .help("Relative time when the key should expire.  \
                                     Either 'N[ymwd]', for N years, months, \
                                     weeks, or days, or 'never'."))
 
@@ -559,9 +578,12 @@ pub fn build() -> App<'static, 'static> {
                                         is updated and existing ones \
                                         will be updated.")
                                 .arg(Arg::with_name("base_directory")
-                                     .value_name("BASE-DIRECTORY")
+                                     .value_name("WEB-ROOT")
                                      .required(true)
-                                     .help("The location to write the WKD to"))
+                                     .help("The location to write the WKD to. \
+                                            This must be the directory the \
+                                            webserver is serving the \
+                                            '.well-known' directory from."))
                                 .arg(Arg::with_name("domain")
                                     .value_name("DOMAIN")
                                     .help("The domain for the WKD.")
